--====================================================================--
-- EuGTK - Copyright 2012 by Irv Mullins, released under the LGPL
--====================================================================--
 
-- EuGTK is a wrapper which enables the Euphoria 4.0 programming 
-- language to use the GTK 3.0 libraries, for rapid GUI program 
-- development in a simple, easy-to-understand style.

-- This library is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2 of the License, or (at your option) any later
-- version.
-- 
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- 
-- See the GNU Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Pl, Suite 330, Boston, MA  02111-1307 USA
-- http://www.gnu.org/licenses/gpl.html`

namespace gtk 

-- This file is hopefully up-to-date with Gtk as of version 3.3.18
-- Modify the following line if necessary:

ifdef WINDOWS then
global object GTKPATH={}
				GTKPATH=getenv("GTK_PATH")

global sequence APP_DIR=current_dir()

if atom(GTKPATH) then
	GTKPATH={}
end if
if length(GTKPATH)>2 then
	if not(equal(GTKPATH,"\\")) then
		GTKPATH&="\\"
		chdir(GTKPATH)
	end if
end if

global constant gtk_lib = GTKPATH&"libgtk-3-0.dll"
global constant gobject_lib = GTKPATH&"libgobject-2.0-0.dll"
global constant gdkpixbuf_lib =GTKPATH&"libgdk_pixbuf-2.0-0.dll"
global constant glib_lib = GTKPATH&"libglib-2.0-0.dll"
global constant pango_lib = GTKPATH&"libpango-1.0-0.dll"
global constant gdk_lib = GTKPATH&"libgdk-3-0.dll"
global constant pangocairo_lib = GTKPATH&"libpangocairo-1.0-0.dll"
global constant cairo_lib = GTKPATH&"libcairo-2.dll"



elsedef

global constant gtk_lib = "libgtk-3.so.0"

end ifdef


ifdef WINDOWS then
	export constant 	version = "4.5.9_w32"
elsedef 
	export constant 	version = "4.5.9"
end ifdef

export constant 	
	release_date = "May 5, 2012",
	author = "Irv Mullins",
	copyright = "© 2012 by Irv Mullins"

public include GtkEnums.e
public include GtkRoutines.e
public include std/io.e
public include std/text.e
public include std/console.e
public include std/convert.e
public include std/machine.e
public include std/filesys.e
public include std/sequence.e

ifdef WINDOWS then
	 include std/dll.e
elsedef
	public include std/dll.e
end ifdef

public include std/types.e
public include std/serialize.e
include std/error.e
include std/search.e
include std/sort.e
include std/locale.e

ifdef WINDOWS then
constant LIBS={GTK,GDK,GDKPIXBUF,GLIB,GOBJECT,PANGO,PANGOCAIRO,CAIRO}
constant LIBNAMES={gtk_lib,gdk_lib,gdkpixbuf_lib,glib_lib,gobject_lib,pango_lib,pangocairo_lib,cairo_lib}	
global function call_back(object id)
		id={'+',id}
		return dll:call_back(id)
	   end function

export function define_c_func(object lib,object routine_name,sequence args,atom return_type)
atom result
	result = dll:define_c_func(lib,routine_name,args,return_type)
	if result<0 then
		puts(1,"GtkEngine.e : Debug Windows :"&routine_name&" not found.\n")
	for i=1 to length(LIBS) do
		result = dll:define_c_func(LIBS[i],routine_name,args,return_type)
		if result>-1 then
			puts(1,"Gotcha! found in"&LIBNAMES[i]&"\n")
			exit
		end if
	end for
	end if
	if result<0 then
		puts(1,"GtkEngine.e : Debug Windows :"&routine_name&" not found. Even the hard way\n")
	end if
	return result
end function

export function define_c_proc(object lib,object routine_name,sequence args)
atom result
	result = dll:define_c_proc(lib,routine_name,args)
	if result<0 then
		puts(1,"GtkEngine.e : Debug Windows :"&routine_name&" not found.\n")
	for i=1 to length(LIBS) do
		result = dll:define_c_proc(LIBS[i],routine_name,args)
		if result>-1 then
			puts(1,"Gotcha! found in "&LIBNAMES[i]&"\n")
			exit
		end if
	end for
	end if
	if result <0 then
		puts(1,"GtkEngine.e : Debug Windows :"&routine_name&" not found. Even the hard way\n")
	end if
	return result
end function


end ifdef


constant -- GTK versions;
	major_ = c_func(define_c_func(GTK,"gtk_get_major_version",{},I)),
	minor_ = c_func(define_c_func(GTK,"gtk_get_minor_version",{},I)),
	micro_ = c_func(define_c_func(GTK,"gtk_get_micro_version",{},I))

-- guess the actual GTK lib name in use;

ifdef WINDOWS then

object libs = sort(vslice(dir(sprintf("%s*",{gtk_lib})),1))

elsedef

object libs = sort(vslice(dir(sprintf("/usr/lib/%s*",{gtk_lib})),1))

end ifdef

export constant 
	lib_version = sprintf("%d.%d.%d",{major_,minor_,micro_}),
	lib_name = libs[$]

ifdef VERSION then
	printf(1,"GTK: %s - %s\n",{lib_version,lib_name})
	printf(1,"EuGTK: %s - %s",{version,release_date})
end ifdef

-- validate correct versions;
if major_ != 3 then 
	crash(`
	------------------------------------------------------------
	GtkLib version mismatch, expecting %d, found %d
	------------------------------------------------------------`,
	{3,major_})
end if

if not equal(enums:version,version) then 
	crash(`
	------------------------------------------------------------
	EuGTK version mismatch: GtkEnums.e is %s, expecting %s
	------------------------------------------------------------`,
	{enums:version,version})
end if

if not equal(routines:version,version) then
	crash(`
	------------------------------------------------------------
	EuGTK version mismatch, GtkRoutines.e is %s, expecting %s
	------------------------------------------------------------`,
	{routines:version,version})
end if

------------------------------------------------------------------------
-- [1] Initialize the GTK engine and set program name;
------------------------------------------------------------------------
constant cmd = command_line()
ifdef CMD then display(cmd) end ifdef

if c_func(define_c_func(GTK,"gtk_init_check",{I,I},B),{0,0}) != 1 then
	crash("Cannot initialize GTK engine!")
else
	c_proc(define_c_proc(GLIB,"g_set_prgname",{S}),
		{allocate_string(filename(cmd[2]),1)})
end if

ifdef COMPILE then
	puts(1,"If program runs without error, it should compile successfully!")
end ifdef

------------------------------------------------------------------------
-- [2] Routines to manage 'method' calls;
------------------------------------------------------------------------
enum NAME, PARAMS, RETVAL, VECTOR
enum ANCESTOR = 2

------------------------------------------------------------------------
function build_hierarchy(integer class) -- used by lookup_method() below
------------------------------------------------------------------------
integer x = class
object ancestry = {x}
while ancestry[$] > 0 do
	for i = 1 to length(control[x][ANCESTOR]) do
		ancestry &= control[x][ANCESTOR][i]
	end for
	x = control[x][ANCESTOR][$]
end while
return ancestry
end function

-------------------------------------------------------------------------
function lookup_method(sequence property, integer class, integer fatal=1)
-------------------------------------------------------------------------
sequence err = `

Warning:
	Can't find method '%s'
	Check spelling and docs for %s class`

integer method = 0
object ancestors = build_hierarchy(class)
object name, params, retval
object prop = property
integer a,b

ifdef LOOK then ---------------------------------------------------------------
	printf(1,"Lookup Method: '%s' for %s",{property,control[class][$]})
	ifdef ANCESTRY then
		puts(1,"\n\t")
		display(build_hierarchy(class))
	end ifdef
end ifdef ---------------------------------------------------------------------

if match("set_",property) or match("get_",property) then
	prop = property[5..$]
end if

for i = 1 to length(ancestors)-1 do
	class = ancestors[i] 
	
	a = find(property,vslice(control[class],1))
	if a > 0 then -- 'long form' - includes the set_ or get_ prefix
		method = a
		name = sprintf("%s_%s",{control[class][NAME],property}) 
		goto "success"
	end if
	
	b = find(prop,vslice(control[class],1))
	if b > 0 then -- 'short form' - without the set_ or get_ prefix
		method = b
		name = sprintf("%s_%s",{control[class][NAME],prop}) 
		goto "success"
	end if

end for

if class > 0 then
	name = control[ancestors[1]][$]
else
	name = "UNKNOWN"
end if

if fatal then 
	if Question(0,"Error","Can't find method",
		sprintf("%s\nfor %s\n\n<small>Open property browser?</small>",
			{property,name})) = MB_YES then
			system_exec("eui property_browser.ex",0)
	end if
else 
	ifdef LOOK then puts(1," (not found)\n") end ifdef
	return -1
end if

label "success"--

ifdef LOOK then --------------------------------------------------------
	printf(1,"\t=> %s\n",{control[class][$]}) 
end ifdef --------------------------------------------------------------

control[class][method] = minsize(control[class][method],4,0)

params = control[class][method][PARAMS]
retval = control[class][method][RETVAL]

for i = 1 to length(params) do -- change 'array pointer' into generic 'pointer'
	if params[i] = A then params[i] = P end if
end for

if control[class][method][VECTOR] = 0 then -- not yet 'instantiated'

	if control[class][method][RETVAL] = 0 then -- a GTK procedure
		control[class][method][VECTOR] = define_c_proc(GTK,name,params)
		
	elsif control[class][method][RETVAL] > 0 then -- a GTK function
		control[class][method][VECTOR] = define_c_func(GTK,name,params,retval)
		
	else
		control[class][method][VECTOR] = retval -- an Eu function
	end if

end if

return control[class][method]

end function

------------------------------------------------------------------------
-- [3] functions to create a EuGTK object;
------------------------------------------------------------------------
------------------------------------------------------------------------
function defaults(sequence prefix, integer class)
------------------------------------------------------------------------
-- This routine helps create() select a default method based on type of 
-- parameter passed to the create func;

switch prefix do
	case "new" then
	switch class do
		case GFile then args[1] = canonical_path(args[1]) return "new_for_path"
		case GtkCSSProvider then return "get_default"
		case GtkAppChooserDialog then return "new_for_content_type"
		case GtkLinkButton then return "new_with_label"
		case GtkExpander then return "new_with_mnemonic"
		case GtkImage, GtkStatusIcon, GtkMenuToolButton, GtkToggleToolButton then 
			if string(args[1]) then
				if file_exists(canonical_path(args[1])) then 
					return "new_from_file" 
				end if
				if match("gtk-",args[1]) = 1 then 
					if atom(args[2]) and args[2] = 0 then args[2] = -1 end if
					return "new_from_stock" 
				end if
				return "new"
			else 
				if args[1] = 0 then return "new" end if
				if lookup_class(args[1]) = GdkPixbuf then
					return "new_from_pixbuf" end if
				if lookup_class(args[1]) = GIcon then
					return "new_from_gicon" end if
				return "new"
			end if
		case GtkRecentManager then return "get_default"
		case GtkButton then return "new_from_stock"
		case GtkImageMenuItem then args = insert(args,0,2) return "new_from_stock" 
		case GtkSpinButton then
			if lookup_class(args[1]) = GtkAdjustment then
				return "new"
			else return "new_with_range"
			end if
		case GtkScale then
			if lookup_class(args[2]) = GtkAdjustment then
				return "new"
			else return "new_with_range"
			end if
		case GtkToggleButton, GtkCheckButton, GtkMenuItem, GtkCheckMenuItem then 
			if match("_",args[1]) > 0 then return "new_with_mnemonic"
			else return "new_with_label"
			end if
		case GtkRadioButton then return "new_with_mnemonic_from_widget"
		case GtkRadioMenuItem then 
			if args[1] != 0 then 
				return "new_with_mnemonic_from_widget"
			else return "new_with_label" 
			end if
		case GdkPixbuf then 
			if args[2] = 0 and args[3] = 0 then return "new_from_file"
			else return "new_from_file_at_size" end if
		case CairoSVGSurface then return "create"
		case GdkCairo_t, Cairo_t, CairoSurface, CairoImageSurface then 
			if string(args[1]) and file_exists(canonical_path(args[1])) then
			   return "create_from_png" else return "create" end if
		case CairoLinearGradient then return "create_linear"
		case CairoRadialGradient then return "create_radial"
		case GtkClipboard then return "get"
		case PangoCairoLayout then return "create_layout"
		case GtkComboBox, GtkComboBoxText then 
			switch args[1] do
				case "with_entry","with entry" then
					return "new_with_entry"
			end switch
	    case GtkRadioToolButton then 
			if args[1] = 0 then return "new_from_stock"
			else return "new_with_stock_from_widget"
			end if
		case GtkToolButton then 
			if string(args[1]) then return "new_from_stock" else return "new" end if
		case GtkIconTheme then return "get_default"
		case GtkTreePath then 
			if string(args[1]) then return "new_from_string" else return "new" end if
	end switch --class

------------------------------
-- alt. calls for convenience;
------------------------------
-- Fonts;
		case "set_font" then return "override_font"
-- Colors;
-- following calls default to state 0 (NORMAL), since 99% of the time 
-- that's what you want, so we supply a 0 and shift the remaining arg 
-- to the right;

		case "set_background" then 
			args = insert(args,0,2) 
			return "override_background_color" 
		case "set_foreground","set_color" then 
			args = insert(args,0,2) 
			return "override_color"

-- if you want to set the color for a state other than NORMAL, you can 
-- use the original form: 
-- set(x,"override_background_color,STATE,"color name")
-- set(x,"override_color",STATE,"color name")

	end switch 
return prefix
end function

object args 

--====================================================================--
  export function create(integer class, 
	object p1 = 0, object p2 = 0, object p3 = 0, object p4 = 0,
	object p5 = 0, object p6 = 0, object p7 = 0, object p8 = 0)
--====================================================================--
sequence err = `

Fatal Error:
	Unable to create a new %s;
	Perhaps you need a later version of GTK?`

	args = {p1,p2,p3,p4,p5,p6,p7,p8}
	object method = lookup_method(defaults("new",class),class) 
	if method[VECTOR] = -1 then
		Warn(0,"Error","Cannot create object",control[class][$])
		crash(err,{control[class][$]})
	end if

	ifdef CREATE then --------------------------------------------------
		printf(1,"CREATE %s",{control[class][$]}) 
		ifdef METHOD then
			printf(1,"\n\tExpects: [ %s]",{decode_params(method[PARAMS])})
			ifdef DETAIL then
				display(method)
			end ifdef
		end ifdef
		ifdef ARGS then
			puts(1,"\n\tARGS: ") display("[]",args[1..length(method[PARAMS])])
		end ifdef
	end ifdef ----------------------------------------------------------

	args = args[1..length(method[PARAMS])]
	if string(p1) and file_exists(canonical_path(p1)) then
		args[1] = canonical_path(p1)
	end if

	for i = 1 to length(method[PARAMS]) do
		switch method[PARAMS][i] do
			case S then 
				if string(args[i]) then args[i] = allocate_string(args[i],0) end if
		end switch
	end for

	ifdef ARGS then
		display(args)
	end ifdef
	
	object handle
	if method[VECTOR] > 0 then 
		handle = c_func(method[VECTOR],args) 
	else
		handle = call_func(-method[VECTOR],args)
	end if

	for i = 1 to length(method[PARAMS]) do
		switch method[PARAMS][i] do
			case S then free(args[i])
		end switch
	end for

	register(handle,class)
	connect_signals(class,handle,{p1,p2,p3,p4})

	ifdef CREATE then --------------------------------------------------
		printf(1,"\thandle = %d\n",handle)
	end ifdef ----------------------------------------------------------

	switch class do -- filter out objects which can't be 'destroyed'...
		case GdkCairo_t, CairoLinearGradient, CairoRadialGradient,
			CairoImageSurface, GFile, GtkListStore, GtkTreeStore then -- do nothing
	
		case else -- set up routine to remove object from registry when done with it.
			if not atom(lookup_method("destroy",class,FALSE)) then
				connect(handle,"destroy",deregister,handle)	
			end if
	end switch -- prevents crash when trying the impossible :)

	return handle
end function

------------------------------------------------------------------------
-- [4] function to set a EuGTK object property;
------------------------------------------------------------------------
--====================================================================--
  export function set(atom handle, sequence property,
	object p1 = 0, object p2 = 0, object p3 = 0, object p4 = 0,
	object p5 = 0, object p6 = 0, object p7 = 0, object p8 = 0)
--====================================================================--
sequence err = `

Warning:
	Cannot set property '%s', misspelled or
	unavailable in this version of GTK.`

object result = 1
integer class

	args = {handle,p1,p2,p3,p4,p5,p6,p7,p8}
	class = lookup_class(handle)
	property = lower(join(split(property,' '),'_'))

	object method = lookup_method(defaults("set_" & property,class),class)
	if method[VECTOR] = -1 then
		if Question(0,"Warning","Cannot set property",
			sprintf("%s\nQuit?",{property})) = MB_YES then
		crash(err,{property})
		else
			return 0
		end if
	end if

	args = args[1..length(method[PARAMS])]

	if match("from_file",method[NAME]) then 
		args[2] = canonical_path(p1)
	end if

	switch method[NAME] do
		case "set_sort_column_id","set_tooltip_column" then args[2] -= 1
		case "set_text","set_markup" then 
			if class = PangoCairoLayout then args[3] = length(p1) end if
		case "set_active" then
			if class = GtkComboBox or class = GtkComboBoxText then args[2] -=1 end if
		case "append" then
			if class = GtkMenu then result = p1 end if
	end switch

	ifdef SET then -----------------------------------------------------
		printf(1,"SET: %s property: '%s' ",{control[class][$],property})
		ifdef METHOD then
			puts(1,"\n\tExpects: ") display(decode_params(method[PARAMS]))
			puts(1,"\tReturns: ") display(decode_params(method[RETVAL]))
			puts(1,"\tVector: ") display(method[VECTOR]) 
		end ifdef
		ifdef ARGS then
			puts(1,"\tARGS: ") display(args)
		end ifdef
	end ifdef ----------------------------------------------------------

	for i = 1 to length(method[PARAMS]) do
		switch method[PARAMS][i] do
			case S then 
				if string(args[i]) then 
					args[i] = allocate_string(args[i],1)
				else args[i] = allocate_string(sprintf("%g",args[i]))
				end if
			case X then args[i] = pango_font(args[i])
			case C then args[i] = color_to_rgba(args[i])
			case A then args[i] = allocate_string_pointer_array(args[i])
			case P then 
				if string(args[i]) then 
					args[i] = allocate_string(args[i],0) 
				end if
		end switch
	end for

	if method[RETVAL] = 0 then
		c_proc(method[VECTOR],args)

	elsif method[RETVAL] > 0 then
		result = c_func(method[VECTOR],args)

	else call_func(-method[VECTOR],args)
	end if

ifdef SET then ---------------------------------------------------------
	puts(1,"\n") 
end ifdef --------------------------------------------------------------

return result
end function

------------------------------------------------------------------------
-- [5] function to get a EuGTK object property;
------------------------------------------------------------------------
--====================================================================--
  export function get(atom handle, sequence property,
	object p1 = machine:allocate(8), object p2 = machine:allocate(8),
	object p3 = machine:allocate(16), object p4 = machine:allocate(8),
	object p5 = machine:allocate(8), object p6 = machine:allocate(8))
--====================================================================--
sequence err = `
Error:
	Cannot get property '%s',
	Unavailable in this version of GTK.`

object result
integer class
object fullpath

	args = {handle,p1,p2,p3,p4,p5,p6}
	class = lookup_class(handle)
	property = lower(join(split(property,' '),'_'))

	object method = lookup_method("get_" & property,class)
	if method[VECTOR] = -1 then
		if Question(0,"Error",
			"Invalid method",
			sprintf("get_%s\nfor class %s \nQuit?",
			{property,control[class][$]})) = MB_YES then
			crash(err,{property})
		else 
			return 0
		end if
	end if

	if equal("load_from_path",property) then
		args[2] = canonical_path(args[2])
	end if

	ifdef GET then ---------------------------------------------------------------
		printf(1,"GET: %s #%d property: '%s' ",{control[class][$],handle,property})
		ifdef METHOD then
			puts(1,"\n\tExpects: ") display(decode_params(method[PARAMS]))
			puts(1,"\tReturns: ") display(decode_params(method[RETVAL]))
			puts(1,"\tVector: ") display(method[VECTOR]) 
		end ifdef
		ifdef ARGS then
			puts(1,"\tARGS: ") display(args[1..length(method[PARAMS])])
		end ifdef
	end ifdef --------------------------------------------------------------------

	args = args[1..length(method[PARAMS])]
	for i = 1 to length(method[PARAMS]) do
		switch method[PARAMS][i] do
			case S then 
				if string(args[i]) then args[i] = allocate_string(args[i],1) end if
		end switch
	end for

	if method[RETVAL] > 0 then -- call a GTK function
		result = c_func(method[VECTOR],args)
	    switch method[RETVAL] do 
			case S then 
				if result != 0 then result = peek_string(result) end if
		end switch

	elsif method[RETVAL] = 0 then -- call a GTK procedure
		c_proc(method[VECTOR],args) 
		for i = 2 to length(method[PARAMS]) do
		switch method[PARAMS][i] do
			case I then args[i] = peek4s(args[i])
			case S then if args[i] != 0 then args[i] = peek_string(args[i]) end if
			case F then args[i] = peek4u(args[i])
		end switch
		end for
		result = args[2..$]

	else result = call_func(-method[VECTOR],args) -- call an Eu function

	end if

-- The exceptions below allow the registry to track non-user-created widgets,
-- for example, widgets which are created by GTK automatically, such as the
-- selection object for tree-views, etc.

	fullpath = control[class][NAME] & "_" & property
	--display(fullpath)
	switch fullpath do 
		case "gtk_tree_view_selection" then register(result,GtkTreeSelection)
		case "gtk_entry_buffer" then register(result,GtkEntryBuffer)
		case "gtk_text_view_buffer" then register(result,GtkTextBuffer)
		case "gtk_text_buffer_tag_table" then register(result,GtkTextTagTable)
		case "gdk_cairo_target" then register(result,CairoSurface)
		case "gtk_cell_area_create_context" then register(result,GtkCellAreaContext)
		case "gtk_window_window" then register(result,GdkWindow)
		case "gtk_dialog_action_area" then register(result,GtkBox)
		case "gtk_dialog_content_area" then register(result,GtkBox)
		case "gtk_dialog_style_context" then register(result,GtkStyleContext)
		case "gtk_print_context_cairo_context" then register(result,Cairo_t)
		case "gtk_window_style_context" then register(result,GtkStyleContext)
		case "gtk_color_selection_dialog_color_selection" then register(result,GtkColorSelection)
		case "gtk_color_button_rgba" then result = result[1]
		case "gtk_icon_theme_load_icon" then register(result,GdkPixbuf)
		case "gtk_label_selection_bounds" then result = peek4u(args[2]) & peek4u(args[3])
		case "gtk_print_operation_print_settings" then register(result,GtkPrintSettings)
		case "gtk_print_operation_default_page_setup" then register(result,GtkPageSetup)
		case "gtk_page_setup_unix_dialog_page_setup" then register(result,GtkPageSetup)
		case "gtk_page_setup_unix_dialog_print_settings" then register(result,GtkPrintSettings)
		case "gdk_display_default_screen" then register(result,GdkScreen)
		case "gdk_display_window_at_pointer" then  
			result &= peek4u(args[2]) & peek4u(args[3]) register(result,GdkWindow)
	end switch

	switch class do -- these classes depend upon an 'info' object;

		case GtkAppChooser, GtkAppChooserButton,GtkAppChooserDialog,GtkAppChooserWidget then
			switch property do
				case "app_info" then register(result,GAppInfo)
			end switch

		case GtkRecentChooser, GtkRecentChooserDialog, GtkRecentChooserMenu, GtkRecentChooserWidget then
			switch property do
				case "current_item" then register(result,GtkRecentInfo)
			end switch

	 -- patch for zero-based return value;
		case GtkComboBox, GtkComboBoxText then
			switch property do
				case "active","get_active" then result += 1 
			end switch

	end switch

	switch property do
			case "path" then register(result,GtkWidgetPath)
			case "pointer" then if length(result) > 3 then return result[2..3] end if
			case "gicon" then register(result,GIcon)
	end switch

	ifdef GET then ----------------------------------------------------- 
		puts(1,"\n") 
	end ifdef ----------------------------------------------------------

	return result
end function

------------------------------------------------------------------------
-- [6] Signal handling;
------------------------------------------------------------------------
------------------------------------------------------------------------
procedure connect_signals(integer class, atom handle, object p)
------------------------------------------------------------------------
-- this routine automatically connects these objects' default signal to
-- a Eu global function name or call_back(routine_id()) supplied on the 
-- create() line. This makes for cleaner code, less typing.

switch class do
	case GtkButton then connect(handle,"clicked",p[2],p[3])
	case GtkToolButton then connect(handle,"clicked",p[2],p[3]) 
	case GtkToggleButton, GtkCheckButton, GtkCheckMenuItem then 
		connect(handle,"toggled",p[2],p[3])
	case GtkRadioButton, GtkRadioMenuItem, GtkRadioToolButton then 
		connect(handle,"toggled",p[3],p[4])
	case GtkSwitch then connect(handle,"notify::active",p[1],p[2])
	case GtkColorButton then connect(handle,"color-set",p[1],p[2])
	case GtkFontButton then connect(handle,"font-set",p[2],p[3])
	case GtkFileChooserButton then connect(handle,"file-set",p[3],p[4])
	case GtkRecentChooserMenu then connect(handle,"item-activated",p[1],p[2])
	case GtkMenuItem, GtkImageMenuItem then 
		if string(p[3]) then p[3] = allocate_string(p[3]) end if
		connect(handle,"activate",p[2],p[3])
end switch
end procedure

constant fn1 = define_c_func(GOBJECT,"g_signal_connect_data",{P,P,P,P,I,I},I)
------------------------------------------------------------------------
export function connect(object handle, object signal, object fn = 0, 
	object data = 0)
------------------------------------------------------------------------
-- this function is for manually connecting signals to Eu routines
-- when it is not appropriate to connect within the create() function
-- it returns a unique sigid (which you usually can ignore, but will
-- sometimes want to store for later use to disconnect the signal)

sequence err = `

Connect error:
	Can't find Eu function %s()
	either not written, or not in scope!`

	object x

if atom(fn) and fn = 0 then return 1 end if -- skip if null func

	if string(signal) then
		signal = allocate_string(signal,1)
	end if
	
	if not atom(data) then 
		data = allocate_string(data,0)  
	end if
	
	if string(fn) then	
		ifdef COMPILE then --------------------------------------------------
			if not equal("Quit",fn) then
				printf(1,"Function %s not visible\n",{fn})
				if Question(0,"Error",
					sprintf("Function %s",{fn}),
					"will not be linked when compiled!\nQuit?") = MB_YES then
					crash("Error in linking function %s",{fn})
				end if
			end if
		end ifdef -----------------------------------------------------------
		x = routine_id(fn) 
		if x < 1 then 
			crash(err,{fn})
		else 
			fn = call_back(x)	
		end if
	end if

	return c_func(fn1,{handle,signal,fn,data,0,0})

end function

constant fn2 = define_c_proc(GOBJECT,"g_signal_handler_disconnect",{P,I})
-------------------------------------------------------------------------------
export procedure disconnect(atom handle, integer sigid) -- sigid from connect() 
-------------------------------------------------------------------------------
c_proc(fn2,{handle,sigid})
end procedure

------------------------------------------------------------------------
-- [7] Convenience functions for clearer code, less typing;
------------------------------------------------------------------------

constant fn3 = define_c_proc(GTK,"gtk_container_add",{P,P})
------------------------------------------------------------------------
export procedure add(atom parent, object child) -- add one or a list 
------------------------------------------------------------------------
if atom(child) then
	c_proc(fn3,{parent,child})
else
	for i =  1 to length(child) do
		add(parent,child[i])
	end for
end if
end procedure

constant fn4 = define_c_proc(GTK,"gtk_box_pack_start",{P,P,B,B,I})
constant fn5 = define_c_proc(GTK,"gtk_box_pack_end",{P,P,B,B,I})
------------------------------------------------------------------------
export procedure pack(atom parent, object child, 
	boolean expand = FALSE, boolean fill = FALSE, integer padding = 0)
------------------------------------------------------------------------
-- normal pack() statement packs a vertical container top-to-bottom,
-- or a horizontal container left-to-right,
-- if child is negative, then packs bottom-to-top/right-to-left

if atom(child) then
	if child > 0 then c_proc(fn4,{parent,child,expand,fill,padding})
	else
		c_proc(fn5,{parent,-child,expand,fill,padding})
	end if
else
	for i = 1 to length(child) do
		pack(parent,child[i],expand,fill,padding)
	end for
end if
end procedure

constant fn6 = define_c_proc(GTK,"gtk_widget_show",{P})
------------------------------------------------------------------------
export procedure show(object x) -- show a widget or list of widgets
------------------------------------------------------------------------
if atom(x) then
	c_proc(fn6,{x})
else
	for i = 1 to length(x) do
		show(x[i])
	end for
end if
end procedure

constant fn7 = define_c_proc(GTK,"gtk_widget_show_all",{P})
------------------------------------------------------------------------
export procedure show_all(object x) -- show container & all its children
------------------------------------------------------------------------
	c_proc(fn7,{x})
end procedure

constant fn8 = define_c_proc(GTK,"gtk_widget_hide",{P})
------------------------------------------------------------------------
export procedure hide(object x) -- hide a widget or list of widgets
------------------------------------------------------------------------
if atom(x) then 
	c_proc(fn8,{x})
else
	for i = 1 to length(x) do
		hide(x[i])
	end for
end if
end procedure

------------------------------------------------------------------------
export function run(atom dlg) -- alias
------------------------------------------------------------------------
object result = get(dlg,"run")
hide(dlg)
return result
end function

------------------------------------------------------------------------
export procedure destroy(atom ctl) -- alias
------------------------------------------------------------------------
set(ctl,"destroy")
end procedure

------------------------------------------------------------------------
-- [8] GTK Engine control
------------------------------------------------------------------------

constant fn9 = define_c_proc(GTK,"gtk_main",{})
------------------------------------------------------------------------
export procedure main() -- starts GTK main processing loop;
------------------------------------------------------------------------
	c_proc(fn9,{})
end procedure

constant fn10 = define_c_func(GTK,"gtk_main_level",{},I)
------------------------------------------------------------------------
export function main_level()
------------------------------------------------------------------------
	return c_func(fn10,{})
end function

constant fn11 = define_c_func(GTK,"gtk_main_iteration",{},B)
------------------------------------------------------------------------
export function main_iteration()-- do one time thru gtk main loop;
------------------------------------------------------------------------
	return c_func(fn11,{})
end function

constant fn12 = define_c_func(GTK,"gtk_main_iteration_do",{B},B)
------------------------------------------------------------------------
export function main_iteration_do(integer blocking)
------------------------------------------------------------------------
	return c_func(fn12,{blocking})
end function

constant fn13 = define_c_func(GTK,"gtk_events_pending",{},B)
------------------------------------------------------------------------
export function events_pending()
------------------------------------------------------------------------
return c_func(fn13,{})
end function

constant fn14 = define_c_func(GTK,"gtk_main_quit",{},I)
------------------------------------------------------------------------
export function Quit() -- kills the gtk main loop and frees memory;
------------------------------------------------------------------------
	return c_func(fn14,{})
end function

------------------------------------------------------------------------
-- [9] Color functions
------------------------------------------------------------------------
atom rgba = allocate(32) -- gtk's color structure

constant fn15 = define_c_func(GDK,"gdk_rgba_parse",{P,P},B)
------------------------------------------------------------------------
export function color_to_rgba(object color)
------------------------------------------------------------------------
object tmp = color
	if atom(color) then color = sprintf("#%06x",color) end if
	color = allocate_string(color,1) 
	if c_func(fn15,{rgba,color}) = 0 then 
		Error(0,"GtkEngine Error","Can't parse color:",tmp)
		crash("Error in string_to_rgba, can't parse color '%s'\n",{tmp})
	end if
	return rgba
end function

constant fn16 = define_c_func(GDK,"gdk_rgba_to_string",{P},S)
------------------------------------------------------------------------
export function rgba_to_string(object color)
------------------------------------------------------------------------
	if atom(color) and color = 0 then 
		crash("Error in rgba_to_string: rgba is null")
	end if
	if sequence(color) and atom(color[1]) then color = color[1] end if
	return peek_string(c_func(fn16,{color}))
end function

------------------------------------------------------------------------
export function rgba_to_hex_str(sequence str)
------------------------------------------------------------------------
	str = str[find('(',str)+1..$-1]
	str = split(str,',')
	for i = 1 to 3 do
		str[i] = to_number(str[i])
	end for
	atom x = (str[1] * 256 * 256) + (str[2] * 256) + str[3]
	return sprintf("#%06x",x)
end function

------------------------------------------------------------------------
export function color_to_triplet(object x)
------------------------------------------------------------------------
x = color_to_rgba(x) 
x = rgba_to_string(x) 
x = rgba_to_hex_str(x) 
return x
end function

------------------------------------------------------------------------
-- [10] Registry - associates handles with classes
------------------------------------------------------------------------
enum HANDLE, CLASS
global object widget_registry = {}

------------------------------------------------------------------------
export procedure register(atom handle, integer class)
------------------------------------------------------------------------
if handle > 0 then
	if find({handle,class},widget_registry) then 
		ifdef REG then
			printf(1,"DUPE HANDLE: %d %s at %d\n",
				{handle,control[class][$],length(widget_registry)})
		end ifdef
    else
		widget_registry = prepend(widget_registry,{handle,class})
		ifdef REG then
			printf(1,"Registering %s handle %d as #%d\n",
				{control[class][$],handle,length(widget_registry)})
		end ifdef
	end if
end if

end procedure

------------------------------------------------------------------------
export function Deregister(object handle)
------------------------------------------------------------------------
if sequence(handle) then
	for i = 1 to length(handle) do
		Deregister(handle[i])
	end for
else
integer x = find(handle,vslice(widget_registry,1))
if x > 0 then
	ifdef DEREG then
		printf(1,"Removing %s handle %d\n",{classname(handle),handle})
	end ifdef
	widget_registry = remove(widget_registry,x) 
end if
end if
return 1
end function
constant deregister = call_back(routine_id("Deregister"))
------------------------------------------------------------------------
export function lookup_class(object handle)
------------------------------------------------------------------------
if not atom(handle) then
	return -1
end if

return vlookup(handle,widget_registry,HANDLE,CLASS,-1)

end function

------------------------------------------------------------------------
export function classname(atom handle)
------------------------------------------------------------------------
integer x = vlookup(handle,widget_registry,HANDLE,CLASS,-1)
if x > 0 then return control[x][$]
else return "Unknown"
end if
end function

------------------------------------------------------------------------
-- [11] Built-in message dialogs for convenience
------------------------------------------------------------------------

constant flags =  GTK_DIALOG_MODAL + GTK_DIALOG_DESTROY_WITH_PARENT

------------------------------------------------------------------------
global function Info(atom parent = 0, sequence title = "Info", 
	object primary_text = {}, object secondary_text = {},
	integer msg_type = GTK_MESSAGE_INFO, 
	integer btn_type = GTK_BUTTONS_OK, atom icon=0)
------------------------------------------------------------------------
	return MessageDialog(parent,title,primary_text,secondary_text,
		msg_type, btn_type, icon)
end function

------------------------------------------------------------------------
global function Warn(atom parent= 0, sequence title = "Warning", 
	object primary_text = {}, object secondary_text = {},
	integer msg_type = GTK_MESSAGE_WARNING, 
	integer btn_type = GTK_BUTTONS_CANCEL, atom icon = 0)
------------------------------------------------------------------------
	return MessageDialog(parent,title,primary_text,secondary_text,
		msg_type, btn_type, icon)	
end function

------------------------------------------------------------------------
global function Error(atom parent = 0, sequence title = "Error", 
	object primary_text = {}, object secondary_text = {},
	integer msg_type = GTK_MESSAGE_ERROR, 
	integer btn_type = GTK_BUTTONS_CLOSE, atom icon = 0)
------------------------------------------------------------------------
	return MessageDialog(parent,title,primary_text,secondary_text,
		msg_type, btn_type, icon)
end function

------------------------------------------------------------------------
global function Question(atom parent = 0, sequence title = "Question", 
	object primary_text = {}, object secondary_text = {},
	integer msg_type = GTK_MESSAGE_QUESTION, 
	integer btn_type = GTK_BUTTONS_YES_NO, atom icon = 0)
------------------------------------------------------------------------
return MessageDialog(parent,title,primary_text,secondary_text,
		msg_type, btn_type, icon)
end function

------------------------------------------------------------------------
global function MessageDialog(atom parent = 0, object title = "?", 
	object primary_text= 0, object secondary_text= 0,
	integer msg_type = 0, 
	integer btn_type = GTK_BUTTONS_NONE,
	object icon = 0)
------------------------------------------------------------------------
atom dlg = create(GtkMessageDialog,parent,flags,msg_type,btn_type,
		primary_text,secondary_text)

	set(dlg,"title",title)
	if msg_type = GTK_MESSAGE_QUESTION then
		set(dlg,"default response",MB_YES)
	end if
	
	if equal({},secondary_text) then
		set(dlg,"markup",primary_text)
	else
		set(dlg,"markup",sprintf("<b>%s</b>",{primary_text}))
		set(dlg,"format secondary markup",secondary_text)
	end if

	if icon != 0 then
		set(dlg,"image",icon) 
		set(icon,"show")
	end if
	
	set(dlg,"position",GTK_WIN_POS_MOUSE)
	
integer result = get(dlg,"run")
	set(dlg,"destroy")
	return result

end function

------------------------------------------------------------------------
global function Input(atom parent = 0, object title = "Input", 
	object primary_text= 0, object secondary_text= 0)
------------------------------------------------------------------------
-- dialog with an entry;

atom dlg = create(GtkDialog)
	set(dlg,"title",title)
	set(dlg,"add button","gtk-cancel",0)
	set(dlg,"add button","gtk-ok",1)

atom lbl1 = create(GtkLabel)
	if string(secondary_text) then
		set(lbl1,"markup",sprintf("<b>%s</b>",{primary_text}))
	else
		set(lbl1,"markup",primary_text)
	end if
		show(lbl1)

atom lbl2 = create(GtkLabel)
		set(lbl2,"markup",secondary_text)
		show(lbl2)

atom inp = create(GtkEntry)
	show(inp)

object ct = get(dlg,"content area")
	add(ct,{lbl1,lbl2,inp})

object result = -1
	if get(dlg,"run") = 1 then
		result = get(inp,"text")
		if not string(result) then result = "" end if
	end if

	set(dlg,"destroy")

return result

end function

------------------------------------------------------------------------
-- [12] Misc stuff;
------------------------------------------------------------------------

constant fn17 = define_c_func(GTK,"gtk_stock_list_ids",{},P)
--------------------------------
export function list_stock_ids()
--------------------------------
atom gs = c_func(fn17,{})
return gslist(gs)
end function

constant fn18 = define_c_func(GLIB,"g_list_length",{P},I)
constant fn19 = define_c_func(GLIB,"g_slist_nth_data",{P,I},P)
------------------------------------------------------------------------
export function gslist(atom gs)
------------------------------------------------------------------------
integer len = c_func(fn18,{gs})
object s = {}
atom data
for i = 0 to len-1 do
	data = c_func(fn19,{gs,i})
	s = append(s,peek_string(data))
end for
return s
end function

------------------------------------------------------------------------
export function deallocate_pointer_array(object array)
------------------------------------------------------------------------
integer x = find(0,peek4u({array,100})) -- get end of array
return peek4u({array,x-1}) -- return list of pointers
end function

------------------------------------------------------------------------
-- [13] Timeouts & Idle functions
------------------------------------------------------------------------
-- Timeout calls your Eu function every n/1000 sec, yielding to other 
-- computer tasks between calls.
 
constant fn20 = define_c_func(GLIB,"g_timeout_add",{I,P,P},I)
---------------------------------------------------------------------
export function timeout(integer interval, object func, atom data = 0)
---------------------------------------------------------------------
object tmp = func
if string(func) then
	func = routine_id(func)
	if func = -1 then
		crash("Timeout: can't find global Euphoria function %s\n",{tmp})
	end if
	func = call_back(func)
end if
return c_func(fn20,{interval,func,data})
end function

-- Idle runs as fast as possible, 
-- yielding to other tasks such as mouse moves, clicks, etc. 
-- Idle will use most of your cpu cycles - probably appropriate
-- for games, not normal apps.

constant fn21 = define_c_func(GLIB,"g_idle_add",{P,P},I)
----------------------------------------------
export function idle(object func, atom data=0)
----------------------------------------------
object tmp = func
if string(func) then
	func = routine_id(func)
	if func = -1 then
		crash("Idle: can't find global Euphoria function %s\n",{tmp})
	end if
	func = call_back(func)
end if
return c_func(fn21,{func,data})
end function

------------------------------------------------------------------------
-- [14] Other random uncategorized stuff;
------------------------------------------------------------------------
constant fn22 = define_c_proc(GTK,"gtk_drag_dest_set",{P,I,P,I,I})
------------------------------------------------------
export procedure set_drag_dest(atom ctl, integer flags)
------------------------------------------------------
atom str_ptr = allocate_string("text/uri-list",1)
atom drop_types_ptr = allocate(12)
poke4(drop_types_ptr,str_ptr)
poke4(drop_types_ptr+4,0)
poke4(drop_types_ptr+8,0)
c_proc(fn22,{ctl,flags,drop_types_ptr,1,GDK_ACTION_MOVE})
end procedure

constant fn23 = define_c_func(GTK,"gtk_selection_data_get_uris",{P},P)
------------------------------------------------------------------------
export function get_selection_uris(atom sel)
------------------------------------------------------------------------
object result = c_func(fn23,{sel})
result = gslist(result)
return result[1]
end function

constant fn24 = define_c_func(GTK,"gtk_show_uri",{P,S,P,P},B)
------------------------------------------------------------------------
export function show_uri(sequence uri)
------------------------------------------------------------------------
atom err = allocate(8) err = 0
return c_func(fn24,{0,allocate_string(uri,1),0,err})
end function

constant fn25 = define_c_func(GTK,"gtk_stock_lookup",{S,P},B)
------------------------------------------------------------------------
export function get_stock_item(sequence name, atom data)
------------------------------------------------------------------------
return c_func(fn25,{allocate_string(name,1),data})
end function

constant fn26 = define_c_func(PANGO,"pango_font_description_from_string",{S},P)
------------------------------------------------------------------------
export function pango_font(object descr)
------------------------------------------------------------------------
atom fnt
if string(descr) then 
	fnt = c_func(fn26,{allocate_string(descr,1)})
	register(fnt,PangoFontDescription)
	return fnt
else
	return 0
end if
end function

constant fn27 = define_c_proc(GTK,"gtk_disable_setlocale",{})
-------------------------------------------------------------
export procedure disable_setlocale()
-------------------------------------------------------------
c_proc(fn27,{})
end procedure

constant fn28 = define_c_func(GTK,"gtk_get_default_language",{},P)
------------------------------------------------------------------
export function get_default_language()
------------------------------------------------------------------
return peek_string(c_func(fn28,{}))
end function

constant fn29 = define_c_proc(GTK,"gtk_grab_add",{P})
------------------------------------------------------------------
export procedure grab_add(atom ctl)
------------------------------------------------------------------
c_proc(fn29,{ctl})
end procedure

constant fn30 = define_c_func(GTK,"gtk_grab_get_current",{},P)
--------------------------------------------------------------
export function grab_get_current()
--------------------------------------------------------------
return c_func(fn30,{})
end function

constant fn31 = define_c_proc(GTK,"gtk_grab_remove",{P})
--------------------------------------------------------
export procedure grab_remove(atom ctl)
--------------------------------------------------------
c_proc(fn31,{ctl})
end procedure

------------------------------------------------------------------------
-- [15] Types for use when declaring parameters passed to Eu routines;
------------------------------------------------------------------------
------------------------------------------------------------------------
export type ToolTip(atom x)
------------------------------------------------------------------------
	register(x,GtkToolTip)
return 1
end type

------------------------------------------------------------------------
export type PrintContext(atom x) -- set test150.ex for usage;
------------------------------------------------------------------------
	register(x,GtkPrintContext)
return 1
end type

export object language = 0
------------------------------------------------------------------------
export function _(sequence x)
------------------------------------------------------------------------
if atom(language) then return x end if
x=split(x) 
for i = 1 to length(x) do
	x[i] = translate(x[i],language,x[i])
end for
return join(x)
end function 

-------------------
----==[ END ]==----
-------------------
69.9