--====================================================================--
-- EuGTK - copyright 2012 by Irv Mullins, released under the LGPL
--====================================================================--

-- This file contains GTK 'widget' identifiers and method prototypes.

-- This library is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2 of the License, or (at your option) any later
-- version.
-- 
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- 
-- See the GNU Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Pl, Suite 330, Boston, MA  02111-1307 USA
-- http://www.gnu.org/licenses/gpl.html

namespace routines


ifdef WINDOWS then
	export constant 	version = "4.5.9_w32"
elsedef 
	export constant 	version = "4.5.9"
end ifdef

include GtkEnums.e
include std/dll.e
include std/get.e
include std/error.e
include std/text.e
include std/types.e
include std/machine.e
include std/sequence.e
include std/locale.e
include std/convert.e
include std/console.e
include std/datetime.e
include std/filesys.e


ifdef WINDOWS then
	
function call_back(object id)
		id={'+',id}
		return dll:call_back(id)
	   end function

function define_c_func(object lib,object routine_name,sequence args,atom return_type)
atom result
	result = dll:define_c_func(lib,routine_name,args,return_type)
	if result<0 then
		puts(1,"GtkRoutine.e : Debug Windows :"&routine_name&" not found.\n")
	end if
	return result
end function

function define_c_proc(object lib,object routine_name,sequence args)
atom result
	result = dll:define_c_proc(lib,routine_name,args)
	if result<0 then
		puts(1,"GtkRoutine.e : Debug Windows :"&routine_name&" not found.\n")
	end if
	return result
end function

end ifdef

-------------------------------------------------------------------------
-- shorthand tags for call/return params used in prototypes;
------------------------------------------------------------------------
export constant 
	P = C_POINTER,	I = E_INTEGER,	B = C_CHAR, 
	C = E_ATOM,		S = C_LONG,		A = E_SEQUENCE,
	X = E_OBJECT, 	F = C_FLOAT,	D = C_DOUBLE

export constant GTK =  open_dll(gtk_lib) 
if GTK = 0 then
	crash("Sorry, you don't have GTK library %s available!",{gtk_lib})
end if

ifdef WINDOWS then

export constant GOBJECT =  open_dll(gobject_lib) 
if GOBJECT = 0 then
	crash("Sorry, you don't have GOBJECT library %s available!",{gobject_lib})
end if

export constant GDKPIXBUF =  open_dll(gdkpixbuf_lib)
if GDKPIXBUF = 0 then
	crash("Sorry, you don't have GDK-PIXBUF library %s available!",{gdkpixbuf_lib})
end if

export constant GLIB =  open_dll(glib_lib)
if GLIB = 0 then
	crash("Sorry, you don't have GLIB library %s available!",{glib_lib})
end if

export constant PANGO =  open_dll(pango_lib)
if PANGO = 0 then
	crash("Sorry, you don't have PANGO library %s available!",{pango_lib})
end if

export constant GDK =  open_dll(gdk_lib)
if GDK = 0 then
	crash("Sorry, you don't have GDK library %s available!",{gdk_lib})
end if

export constant PANGOCAIRO =  open_dll(pangocairo_lib)
if PANGOCAIRO = 0 then
	crash("Sorry, you don't have PANGOCAIRO library %s available!",{pangocairo_lib})
end if

export constant CAIRO =  open_dll(cairo_lib)
if CAIRO = 0 then
	crash("Sorry, you don't have PANGOCAIRO library %s available!",{cairo_lib})
end if
chdir (APP_DIR)
elsedef

export constant GOBJECT=GTK
export constant GDKPIXBUF=GTK
export constant GLIB=GTK
export constant PANGO=GTK
export constant GDK=GTK
export constant PANGOCAIRO=GTK
end ifdef

constant gtype_from_name = define_c_func(GOBJECT,"g_type_from_name",{S},I)
global function type_from_name(sequence name)
return c_func(gtype_from_name,{allocate_string(name)})
end function

constant g_type_init = define_c_proc(GOBJECT,"g_type_init",{})
c_proc(g_type_init,{}) 

constant gpixtype = define_c_func(GDKPIXBUF,"gdk_pixbuf_get_type",{},P)
global constant gPIXBUF = c_func(gpixtype,{})

constant gboxtype = define_c_func(GTK,"gtk_combo_box_get_type",{},P)
global constant gCOMBO = c_func(gboxtype,{})

------------------------------------------------------------------------
export function decode_params(object params) -- convert to readable form
------------------------------------------------------------------------
if atom(params) then 
	if params < 0 then return "EuFunc" end if
	params = {params} 
end if
params = transmute(params,{{},0,P,I,B,C,S,A,X,D,F},
		{{},"Nul ","Ptr ","Int ","Bool ","Char ","Str ","Array ",
			"Other ","Dbl ","Flt "})
return params 
end function

------------------------------------------------------------------------
export function decode_gtypes(object params) -- convert to readable form
------------------------------------------------------------------------
if atom(params) then params = {params} end if
	params = transmute(params,{{},0,64,24,56,20,68,gPIXBUF},
		{{},"Nul ","gSTR ","gINT ","gFLOAT ","gBOOL ","gPTR ","gPIXBUF "})
return params
end function

--====================================================================--
-- These are the names of the GTK & Cairo objects
--====================================================================--
-- These need not be in any specific order, except that
-- GObject must be first, and GtkFinal must be last!

public enum 
	GObject,
	GInitiallyUnowned,
	GActionGroup,
	GApplication,
	GAppLaunchContext,
	GBoxed,
	GClosure,
	GIcon,
	GEmblem,
	GEmblemedIcon,
	GFile,
	GFileIcon,
	GThemedIcon,
	Gdk,
	GdkAppLaunchContext,
	AtkObject,
	PangoCairo,
	Cairo_t,
	CairoPattern,
	CairoPatternRGB,
	CairoPatternRGBA,
	CairoSurface,
	CairoImageSurface,
	CairoPDFSurface,
	CairoPSSurface,
	CairoSVGSurface,
	CairoLinearGradient,
	CairoRadialGradient,
	GdkCairo_t,
	GdkColor,
	GdkCursor,
	GdkDisplay,
	GdkDrawable,
	GdkEvent,
	GdkPangoContext,
	GdkPixbuf,
	GdkPixbufAnimation,
	GdkPixmap,
	GdkRGBA,
	GdkScreen,
	GdkWindow,
	GInterface,
	GSignal,
	GSignalHandler,
	GtkAboutDialog,
	GtkAccelerator,
	GtkAccelGroup,
	GtkAccelLabel,
	GtkAccelMap,
	GtkAccessible,
	GAppInfo,
	GtkAppChooserDialog,
	GtkAction,
	GtkActionable,
	GtkActionGroup,
	GtkActivatable,
	GtkAdjustment,
	GtkAlignment,
	GtkAppChooser,
	GtkAppChooserButton,
	GtkAppChooserWidget,
	GtkApplication,
	GtkApplicationWindow,
	GtkArrow,
	GtkAspectFrame,
	GtkAssistant,
	GtkOverlay,
	GtkBin,
	GtkBindingEntry,
	GtkBindings,
	GtkBindingSet,
	GtkBorder,
	GtkBox,
	GtkBuilder,
	GtkButton,
	GtkButtonBox,
	GtkCalendar,
	GtkCellArea,
	GtkCellAreaBox,
	GtkCellAreaContext,
	GtkCellEditable,
	GtkCellLayout,
	GtkCellRenderer,
	GtkCellRendererAccel,
	GtkCellRendererCombo,
	GtkCellRendererEditable,
	GtkCellRendererPixbuf,
	GtkCheckButton,
	GtkCellRendererProgress,
	GtkCellRendererSpin,
	GtkCellRendererSpinner,
	GtkCellRendererText,
	GtkCellRendererToggle,
	GtkCellView,
	GtkCheckMenuItem,
	GtkClipboard,
	GtkColorButton,
	GtkColorChooser,
	GtkColorSelection,
	GtkColorSelectionDialog,
	GtkComboBox,
	GtkComboBoxText,
	GtkContainer,
	GtkCssProvider,
	GtkCSSProvider,
	GtkCSSSection,
	GtkDialog,
	GtkDrawingArea,
	GtkEditable,
	GtkEntry,
	GtkEntryBuffer,
	GtkEntryCompletion,
	GtkEventBox,
	GtkExpander,
	GtkFileChooser,
	GtkFileChooserButton,
	GtkFileChooserDialog,
	GtkFileChooserWidget,
	GtkFileFilter,
	GtkFixed,
	GtkFontButton,
	GtkFontChooser,
	GtkFontChooserDialog,
	GtkFontChooserWidget,
	GtkFontSelectionDialog,
	GtkFrame,
	GtkGradient,
	GtkGrid,
	GtkHandleBox,
	GtkIconInfo,
	GtkIconTheme,
	GtkIconView,
	GtkImage,
	GtkImageMenuItem,
	GtkImContext,
	GtkImContextSimple,
	GtkImMulticontext,
	GtkInfoBar,
	GtkInputDialog,
	GtkInvisible,
	GtkItem,
	GtkLabel,
	GtkLayout,
	GtkLinkButton,
	GtkListStore,
	GtkLockButton,
	GtkMain,
	GtkMenu,
	GtkMenuBar,
	GtkMenuItem,
	GtkMenuShell,
	GtkMenuToolButton,
	GtkMessageDialog,
	GtkMisc,
	GtkMountOperation,
	GtkNotebook,
	GtkNumerableIcon,
	GtkOffscreenWindow,
	GtkOrientable,
	GtkPageSetup,
	GtkPageSetupUnixDialog,
	GtkPaned,
	GtkPaperSize,
	GtkPlug,
	GtkPrintBackend,
	GtkPrintContext,
	GtkPrinter,
	GtkPrinterOption,
	GtkPrinterOptionSet,
	GtkPrinterOptionWidget,
	GtkPrintJob,
	GtkPrintOperation,
	GtkPrintRun,
	GtkPrintSettings,
	GtkPrintUnixDialog,
	GtkProgressBar,
	GtkRadioAction,
	GtkRadioButton,
	GtkRadioMenuItem,
	GtkRadioToolButton,
	GtkRange,
	GtkRecentAction,
	GtkRecentChooser,
	GtkRecentChooserDialog,
	GtkRecentChooserMenu,
	GtkRecentChooserWidget,
	GtkRecentFilter,
	GtkRecentInfo,
	GtkRecentManager,
	GtkRequisition,
	GtkRuler,
	GtkScale,
	GtkScaleButton,
	GtkScrollable,
	GtkScrollbar,
	GtkScrolledWindow,
	GtkSelection,
	GtkSelectionData,
	GtkSeparator,
	GtkSeparatorMenuItem,
	GtkSeparatorToolItem,
	GtkSettings,
	GtkSizeGroup,
	GtkSocket,
	GtkSpinButton,
	GtkSpinner,
	GtkStatusbar,
	GtkStatusIcon,
	GtkStock,
	GtkStyleContext,
	GtkStyleProvider,
	GtkSwitch,
	GtkTargetEntry,
	GtkTargetList,
	GtkTargetTable,
	GtkTextAttributes,
	GtkTextBuffer,
	GtkTextIter,
	GtkTextMark,
	GtkTextTag,
	GtkTextTagTable,
	GtkTextView,
	GtkToggleAction,
	GtkToggleButton,
	GtkToggleToolButton,
	GtkToolBar,
	GtkToolButton,
	GtkToolItem,
	GtkToolItemGroup,
	GtkToolPalette,
	GtkToolShell,
	GtkToolTip,
	GtkTree,
	GtkTreeDragDest,
	GtkTreeDragSource,
	GtkTreeIter,
	GtkTreeModel,
	GtkTreeModelFilter,
	GtkTreeModelSort,
	GtkTreePath,
	GtkTreeRowReference,
	GtkTreeSelection,
	GtkTreeSortable,
	GtkTreeStore,
	GtkTreeView,
	GtkTreeViewColumn,
	GtkViewport,
	GtkVolumeButton,
	GtkWidget,
	GtkWidgetPath,
	GtkWindow,
	GtkWindowGroup,
	PangoAttrForeground,
	PangoAttrList,
	PangoCairoLayout,
	PangoContext,
	PangoFont,
	PangoFontDescription,
	PangoFontFace,
	PangoFontFamily,
	PangoFontMap,
	PangoFontSet,
	PangoFontsetSimple,
	PangoLayout,
	PangoRenderer,
GtkFinal -- keep this at the end of the list!

object fn = repeat(0,20)

------------------------------------------------------------------------
-- Main control structure which will be filled in by the declarations 
-- which follow; indexed by the enums listed above.
------------------------------------------------------------------------
export sequence control = repeat(0,GtkFinal) 

------------------------------------------------------------------------
-- Control method declarations, order is not important
------------------------------------------------------------------------
control[GObject] = {"g_object",{0},
	{"new",{I,S,I},P},
	{"set",{P,S,P,P}},
	{"set_property",{P,S,P},-routine_id("setObjectProperty")},
	{"get_property",{P,S},-routine_id("getObjectProperty")},
	{"get_data",{P,S},S},
	{"set_data",{P,S,P}},
	{"set_data_full",{P,S,S,P}},
	{"steal_data",{P,S},P},
"GObject"}

fn[2] = define_c_proc(GOBJECT,"g_object_set",{P,P,P,P})
------------------------------------------------------------------------
global function setObjectProperty(atom handle, object a, object b)
------------------------------------------------------------------------
atom err = allocate(16) err = 0
	if string(a) then a = allocate_string(a,1) end if
	if string(b) then b = allocate_string(b,1) end if
	c_proc(fn[2],{handle,a,b,err})
	return 0
end function

fn[3] = define_c_proc(GOBJECT,"g_object_get_property",{P,S,P})
------------------------------------------------------------------------
global function getObjectProperty(atom handle, sequence prop, atom vt)
------------------------------------------------------------------------
c_proc(fn[3],{handle,allocate_string(prop),vt})
return vt
end function

control[GInitiallyUnowned] = {" ",
{GObject},
"GInitiallyUnowned"}

control[GInterface] = {" ",
{0}, -- empty to complete hierarchy for some ctls
"GInterface"}

control[GClosure] = {"g_cclosure",
{GObject},
	{"new",{P,P,P},P},
"GClosure"}

control[GFile] = {"g_file",
{GInterface,GObject},
	{"new_for_path",{S},P},
	{"new_for_uri",{S},P},
	{"get_parse_name",{P},S},
	{"parse_name",{S},P},
"GFile"}

control[GIcon] = {"g_icon",
{GInterface},
	{"hash",{P},I},
	{"equal",{P,P},B},
	{"to_string",{P},S},
	{"new_for_string",{S,P},P},
"GIcon"}

control[GFileIcon] = {"g_file_icon",
{GObject},
	{"new",{P},P},
	{"get_file",{P},P},
"GFileIcon"}

control[GThemedIcon] =  {"g_themed_icon",
{GObject},
	{"new",{S},P},
	{"new_with_default_fallbacks",{S},P},
	{"get_names",{P},A},
"GThemedIcon"}

control[GEmblem] = {"g_emblem",
{GObject},
	{"new",{P},P},
	{"get_icon",{P},P},
"GEmblem"}

control[GEmblemedIcon] = {"g_emblemed_icon",
{GObject},
	{"new",{P,P},P},
"GEmblemedIcon"}

control[GtkRecentAction] = {"gtk_recent_action",
{GtkAction},
"GtkRecentAction"}

control[GtkActivatable] = {"gtk_activatable",
{GInterface},
	{"do_set_related_action",{P,P}},
	{"get_related_action",{P},P},
	{"get_use_action_appearance",{P},B},
	{"sync_action_properties",{P,P}},
	{"set_related_action",{P,P}},
	{"set_use_action_appearance",{P,B}},
"GtkActivatable"}

control[GSignal] = {"g_signal",
{Gdk},
	{"connect",{P,S,P,P},I},
	{"connect_after",{P,S,P,P},I},
	{"connect_swapped",{P,S,P,P},I},
	{"emit_by_name",{P,S}},
	{"lookup",{S,I},I},
	{"name",{I},S},
	{"stop_emission",{P,I,I}},
	{"stop_emission_by_name",{P,S}},
"GSignal"}

control[GSignalHandler] = {"g_signal_handler",
{Gdk},
	{"block",{P,I}},
	{"unblock",{P,I}},
	{"unblock_by_func",{P,P,P}},
	{"connect_data",{P,S,P,P,P,I},P},
	{"disconnect",{P,I}},
	{"is_connected",{P,I},B},
"GSignalHandler"}

control[GdkAppLaunchContext] = {"gdk_app_launch_context",
{Gdk},
	{"new",{},P},
	{"set_display",{P,P}},
	{"set_screen",{P,P}},
	{"set_desktop",{P,I}},
	{"set_timestamp",{P,I}},
	{"set_icon",{P,P}},
	{"set_icon_name",{P,S}},
"GdkAppLaunchContext"}

control[GdkEvent] = {"gtk_event",
{Gdk},
	{"events_pending",{},B},
	{"peek",{},P},
	{"get",{},P},
	{"new",{I},P},
	{"copy",{P},P},
	{"free",{P}},
	{"get_axis",{P,I,D},B},
	{"get_button",{P,I},B},
	{"get_click_count",{P,I},B},
	{"get_coords",{P,D,D},B},
	{"get_keycode",{P,I},B},
	{"get_keyval",{P,I},B},
	{"get_root_coords",{P,D,D},B},
	{"get_scroll_direction",{P,I},B},
	{"get_state",{P,I},B},
	{"get_time",{P},I},
	{"request_motions",{P}},
	{"get_angle",{P,P,D},B},
	{"get_center",{P,P,D,D},B},
	{"get_distance",{P,P,D},B},
	{"handler_set",{P,P,P}},
	{"get_show_events",{},B},
	{"set_show_events",{B}},
	{"set_screen",{P,P}},
	{"get_screen",{P},P},
	{"get_device",{P},P},
	{"set_device",{P,P}},
	{"get_source_device",{P},P},
	{"set_source_device",{P,P}},
"GdkEvent"}

control[GdkColor] = {"gdk_color",
{Gdk},
	{"copy",{P},P},
	{"free",{P}},
	{"parse",{S,P},B},
	{"equal",{P,P},B},
	{"hash",{P},I},
	{"to_string",{P},S},
"GdkColor"}

control[GtkMountOperation] = {"gtk_mount_operation",
{GObject},
	{"new",{P},P},
"GtkMountOperation"}

control[GtkOrientable] = {"gtk_orientable",
{GInterface},
	{"set_orientation",{P,I}},
	{"get_orientation",{P},I},
"GtkOrientable"}

control[GdkWindow] = {"gdk_window",
{GObject},
	{"new",{P,P,I},P},
	{"destroy",{P}},
	{"get_window_type",{P},I},
	{"get_display",{P},P},
	{"get_screen",{P},P},
	{"get_visual",{P},P},
	{"show",{P}},
	{"show_unraised",{P}},
	{"hide",{P}},
	{"is_destroyed",{P},B},
	{"is_visible",{P},B},
	{"is_viewable",{P},B},
	{"is_input_only",{P},B},
	{"is_shaped",{P},B},
	{"get_state",{P},I},
	{"flush",{P}},
	{"shape_combine_region",{P,P,I,I}},
	{"set_cursor",{P,P}},
	{"get_cursor",{P},P},
	{"get_geometry",{P,I,I,I,I}},
	{"get_width",{P},I},
	{"get_height",{P},I},
	{"get_origin",{P,I,I}},
"GdkWindow"}

control[Gdk] = {"gdk",GObject,
	{"draw_layout",{P,P,I,I,P}},
	{"screen_width",{},I},
	{"screen_height",{},I},
	{"notify_startup_complete",{}},
	{"notify_startup_complete_with_id",{S}},
	{"get_program_class",{},S},
	{"set_program_class",{S}},
	{"get_display",{},S},
	{"flush",{}},
	{"pointer_grab",{P,B,I,P,P,I},I},
	{"pointer_ungrab",{I}},
	{"pointer_is_grabbed",{},B},
	{"set_double_click_time",{I}},
	{"keyboard_grab",{P,B,I},I},
	{"keyboard_ungrab",{I}},
	{"beep",{}},
	{"error_trap_push",{}},
	{"error_trap_pop",{},I},
"Gdk"}

control[GdkDisplay] = {"gdk_display",
{GObject},
	{"open",{S},P},
	{"new",{},-routine_id("newDisplay")},
	{"get_default",{},P},
	{"get_name",{P},S},
	{"get_n_screens",{P},I},
	{"get_screen",{P,I},P},
	{"get_default_screen",{P},P},
	{"pointer_ungrab",{P,I}},
	{"keyboard_ungrab",{P,I}},
	{"pointer_is_grabbed",{P},B},
	{"beep",{}},
	{"sync",{P}},
	{"flush",{P}},
	{"close",{P}},
	{"is_closed",{P},B},
	{"list_devices",{P},A},
	{"get_event",{P},P},
	{"peek_event",{P},P},
	{"put_event",{P,P}},
	{"get_pointer",{P,P,I,I,I}},
	{"get_window_at_pointer",{P,I,I},P},
	{"warp_pointer",{P,P,I,I}},
	{"supports_cursor_color",{P},B},
	{"supports_cursor_alpha",{P},B},
	{"get_default_cursor_size",{P},I},
	{"get_maximal_cursor_size",{P,I,I}},
	{"get_default_group",{P},P},
	{"supports_selection_notification",{P},B},
	{"request_selection_notification",{P,P},B},
	{"supports_clipboard_persistence",{P},B},
	{"supports_shapes",{P},B},
	{"supports_input_shapes",{P},B},
	{"supports_composite",{P},B},
"GdkDisplay"}

fn[4] = define_c_func(GDK,"gdk_display_get_default",{},P)
-----------------------------------------------------------
function newDisplay()
	return c_func(fn[4],{})
end function

control[GdkScreen] = {"gdk_screen",
{GObject},
	{"new",{},-routine_id("getScreen")},
	{"get_default",{},P},
	{"get_default_colormap",{P},P},
	{"set_default_colormap",{P,P}},
	{"get_system_colormap",{P},P},
	{"get_system_visual",{P},P},
	{"get_rgba_colormap",{P},P},
	{"get_rgba_visual",{P},P},
	{"screen_is_composited",{P},B},
	{"get_root_window",{P},P},
	{"get_display",{P},P},
	{"get_number",{P},I},
	{"get_width",{P},I},
	{"get_height",{P},I},
	{"list_visuals",{P},A},
	{"get_toplevel_windows",{P},A},
	{"make_display_name",{P},S},
	{"get_n_monitors",{P},I},
	{"get_primary_monitor",{P},I},
	{"get_monitor_geometry",{P,I,P}},
	{"get_monitor_at_point",{P,I,I},I},
	{"get_monitor_at_window",{P,P},I},
	{"get_monitor_height_mm",{P},I},
	{"get_monitor_width_mm",{P},I},
	{"get_monitor_plug_name",{P,I},S},
	{"broadcast_client_message",{P,P}},
	{"get_font_options",{P},P},
	{"set_font_options",{P,P}},
	{"get_resolution",{P},D},
	{"set_resolution",{P,D}},
	{"get_active_window",{P},P},
	{"get_window_stack",{P},A},
"GdkScreen"}

fn[5] = define_c_func(GDK,"gdk_display_get_screen",{},P)
------------------------------------------------------------------------
-- gets current screen, rather than creating a new one
------------------------------------------------------------------------
function getScreen()
	return c_func(fn[5],{})
end function

control[GdkPixmap] = {"gdk_pixmap",
{Gdk},
	{"new",{P,I,I,I},P},
	{"get_size",{P,I,I}},
"GdkPixmap"}

control[PangoCairoLayout] = {"pango_cairo",
{PangoLayout},
	{"create_layout",{P},P},
	{"show_layout",{P,P},-routine_id("ShowLayout")},
"PangoCairoLayout"}

fn[6] = define_c_proc(PANGOCAIRO,"pango_cairo_show_layout",{P,P})
------------------------------------------------------------------------
function ShowLayout(atom cr, atom layout)
------------------------------------------------------------------------
	c_proc(fn[6],{layout,cr})
return 1
end function

control[CairoPattern] = {"cairo_pattern",
{Cairo_t},
	{"add_color_stop_rgb",{P,D,D,D,D}},
	{"add_color_stop_rgba",{P,D,D,D,D,D}},
	{"get_color_stop_count",{P,I},I},
	{"get_color_stop_rgba",{P,I,D,D,D,D,D},I},
"CairoLinearGradient"}

control[CairoLinearGradient] = {"cairo_pattern",
{CairoPattern},
	{"create_linear",{D,D,D,D},P},
"CairoLinearGradient"}

control[CairoRadialGradient] = {"cairo_pattern",
{CairoPattern},
	{"create_radial",{D,D,D,D,D,D},P},
"CairoRadialGradient"}

control[CairoPatternRGB] = {"cairo_pattern",
{CairoPattern},
	{"create_rgb",{D,D,D},P},
"CairoPatternRGB"}

control[CairoPatternRGBA] = {"cairo_pattern",
{CairoPattern},
	{"create_rgba",{D,D,D,D},P},
"CairoPatternRGBA"}

control[GdkCairo_t] = {"gdk_cairo",
{CairoImageSurface},
	{"create",{P},P},
"Cairo_t"}

control[CairoSurface] = {"cairo_surface",
{Cairo_t},
	{"write_to_png",{P,S},I},
"CairoSurface"}

control[CairoImageSurface] = {"cairo_image_surface",
{CairoSurface},
	{"create",{I,I,I},P},
	{"create_from_png",{S},P},
	{"get_format",{P},I},
	{"get_width",{P},I},
	{"get_height",{P},I},
	{"get_stride",{P},I},
"CairoImageSurface"}

control[CairoPSSurface] = {"cairo_ps_surface",
{CairoSurface},
	{"create",{S,D,D},P},
	{"set_size",{P,D,D}},
	{"set_eps",{P,B}},
	{"get_eps",{P},B},
	{"restrict_to_level",{P,I}},
"CairoPSSurface"}

control[CairoPDFSurface] = {"cairo_pdf_surface",
{CairoSurface},
	{"create",{S,D,D},P},
	{"set_size",{P,D,D}},
	{"restrict_to_version",{P,I}},
"CairoPDFSurface"}

control[CairoSVGSurface] = {"cairo_svg_surface",
{CairoSurface},
	{"create",{S,D,D},P},
	{"restrict_to_version",{P,I}},
"CairoSVGSurface"}

control[Cairo_t] = {"cairo",
{0},
	{"create",{P},P},
	{"reference",{P},P},
	{"destroy",{P}},
	{"status",{P},I},
	{"save",{P}},
	{"restore",{P}},
	{"get_target",{P},P},
	{"push_group",{P}},
	{"push_group_with_content",{P,P}},
	{"pop_group",{P},P},
	{"pop_group_to_source",{P}},
	{"get_group_target",{P},P},
	{"set_source_rgba",{P,D,D,D,D}},
	{"set_source_rgb",{P,D,D,D}},
	{"set_source",{P,P}},
	{"get_source",{P},P},
	{"set_source_surface",{P,P,D,D}},
	{"set_antialias",{P,I}},
	{"get_antialias",{P},I},
	{"set_dash",{P,D,I,D}},
	{"get_dash_count",{P},I},
	{"get_dash",{P,D,D}},
	{"set_fill_rule",{P,I}},
	{"get_fill_rule",{P},I},
	{"set_line_cap",{P,I}},
	{"get_line_cap",{P},I},
	{"set_line_join",{P},I},
	{"get_line_join",{P},I},
	{"set_line_width",{P,D}},
	{"get_line_width",{P},D},
	{"set_miter_limit",{P,I}},
	{"get_miter_limit",{P},I},
	{"set_operator",{P,I}},
	{"get_operator",{P},I},
	{"set_pattern",{P,P}},
	{"set_tolerance",{P,D}},
	{"get_tolerance",{P},D},
	{"clip",{P}},
	{"clip_preserve",{P}},
	{"clip_extents",{P,D,D,D,D}},
	{"in_clip",{P,D,D},B},
	{"reset_clip",{P}},
	{"rectangle_list_destroy",{P}},
	{"fill",{P}},
	{"fill_preserve",{P}},
	{"fill_extents",{P,D,D,D,D}},
	{"in_fill",{P,D,D},B},
	{"mask",{P,P}},
	{"mask_surface",{P,P,D,D}},
	{"paint",{P}},
	{"paint_with_alpha",{P,D}},
	{"stroke",{P}},
	{"stroke_preserve",{P}},
	{"stroke_extents",{P,D,D,D,D}},
	{"in_stroke",{P,D,D},B},
	{"copy_page",{P}},
	{"show_page",{P}},
	{"copy_path",{P},P},
	{"copy_path_flat",{P},P},
	{"path_destroy",{P}},
	{"append_path",{P,P}},
	{"has_current_point",{P},B},
	{"get_current_point",{P,D,D}},
	{"new_path",{P}},
	{"new_sub_path",{P}},
	{"close_path",{P}},
	{"set_user_data",{P,S,P,P},I},
	{"get_user_data",{P,S}},
	{"arc",{P,D,D,D,D,D}},
	{"arc_negative",{P,D,D,D,D,D}},
	{"move_to",{P,D,D}},
	{"rel_move_to",{P,D,D}},
	{"line_to",{P,D,D}},
	{"rel_line_to",{P,D,D}},
	{"rectangle",{P,D,D,D,D}},
	{"glyph_path",{P,I,I}},
	{"text_path",{P,S}},
	{"curve_to",{P,D,D,D,D,D,D}},
	{"rel_curve_to",{P,D,D,D,D,D,D}},
	{"path_extents",{P,D,D,D,D}},
	{"set_font_face",{P,S}},
	{"show_text",{P,S}},
	{"device_get_type",{P},I},
	{"device_status",{P},I},
	{"status_to_string",{I},S},
	{"translate",{P,D,D}},
	{"scale",{P,D,D}},
	{"rotate",{P,D}},
	{"transform",{P,P}},
	{"set_matrix",{P,P}},
	{"get_matrix",{P,P}},
	{"identity_matrix",{P}},
	{"user_to_device",{P,D,D}},
	{"user_to_device_distance",{P,D,D}},
	{"device_to_user",{P,D,D}},
	{"device_to_user_distance",{P,D,D}},
	{"version",{},I},
	{"version_string",{},S},
	{"set_font_size",{P,D}},
	{"set_font_matrix",{P,P}},
	{"get_font_matrix",{P,P}},
	{"set_font_options",{P,P}},
	{"get_font_options",{P,P}},
	{"set_font_face",{P,P}},
	{"get_font_face",{P},P},
	{"set_scaled_font",{P,P}},
	{"get_scaled_font",{P},P},
	{"show_text",{P,S}},
	{"show_glyphs",{P,P}},
	{"show_text_glyphs",{P,S,I,P,I,P,I,I}},
	{"font_extents",{P,P}},
	{"text_extents",{P,S,P}},
	{"glyph_extents",{P,P,I,P}},
	{"toy_font_face_create",{S,I,I},P},
	{"toy_font_face_get_slant",{P},I},
	{"toy_font_face_get_weight",{P},I},
	{"glyph_allocate",{I},P},
	{"glyph_free",{P}},
	{"text_cluster_allocate",{I},P},
	{"text_cluster_free",{P}},
	{"set_layout_path",{P,P},-routine_id("SetLayoutPath")},
"Cairo_t"}

fn[7] = define_c_proc(PANGOCAIRO,"pango_cairo_layout_path",{P,P})
------------------------------------------------------------------------
function SetLayoutPath(atom cr, atom layout)
------------------------------------------------------------------------
	c_proc(fn[7],{cr,layout})
	return 1
end function

control[GdkCursor] = {"gdk_cursor",
{GObject},
	{"new",{I},P},
	{"new_from_pixbuf",{P,P,I,I},P},
	{"new_from_name",{P,S},P},
	{"new_for_display",{P,I},P},
	{"get_display",{P},P},
	{"get_image",{P},P},
	{"get_cursor_type",{P},I},
	{"ref",{P},P},
	{"unref",{P}},
"GdkCursor"}

control[GtkTextIter] = {"gtk_text_iter",
{GBoxed},
	{"get_buffer",{P},P},
	{"copy",{P},P},
	{"free",{P}},
	{"get_offset",{P},I},
	{"get_line",{P},I},
	{"get_line_offset",{P},I},
	{"get_line_index",{P},I},
	{"get_visible_line_index",{P},I},
	{"get_visible_line_offset",{P},I},
	{"get_char",{P},I},
	{"get_slice",{P,P},S},
	{"get_text",{P,P},S},
	{"get_visible_slice",{P,P},S},
	{"get_visible_text",{P,P},S},
	{"get_pixbuf",{P},P},
	{"get_marks",{P},A},
	{"get_toggled_tags",{P,B},A},
	{"get_child_anchor",{P},P},
	{"begins_tag",{P,P},B},
	{"ends_tag",{P,P},B},
	{"toggles_tag",{P,P},B},
	{"has_tag",{P,P},B},
	{"get_tags",{P},A},
	{"editable",{P,B},B},
	{"can_insert",{P,B},B},
	{"starts_word",{P},B},
	{"ends_word",{P},B},
	{"inside_word",{P},B},
	{"starts_line",{P},B},
	{"ends_line",{P},B},
	{"starts_sentence",{P},B},
	{"ends_sentence",{P},B},
	{"inside_sentence",{P},B},
	{"is_cursor_position",{P},B},
	{"get_chars_in_line",{P},I},
	{"get_bytes_in_line",{P},I},
	{"get_attributes",{P,P},B},
	{"get_language",{P},P},
	{"is_end",{P},B},
	{"is_start",{P},B},
	{"forward_char",{P},B},
	{"backward_char",{P},B},
	{"forward_chars",{P,I},B},
	{"backward_chars",{P,I},B},
	{"forward_line",{P},B},
	{"backward_line",{P},B},
	{"forward_lines",{P,I},B},
	{"backward_lines",{P,I},B},
	{"forward_word_ends",{P,I},B},
	{"backward_word_starts",{P,I},B},
	{"forward_word_end",{P},B},
	{"backward_word_start",{P},B},
	{"forward_cursor_position",{P},B},
	{"backward_cursor_position",{P},B},
	{"forward_cursor_positions",{P,I},B},
	{"backward_cursor_positions",{P,I},B},
	{"backward_sentence_start",{P},B},
	{"backward_sentence_starts",{P,I},B},
	{"forward_sentence_end",{P},B},
	{"forward_sentence_ends",{P,I},B},
	{"forward_visible_word_ends",{P,I},B},
	{"backward_visible_word_starts",{P,I},B},
	{"forward_visible_word_end",{P},B},
	{"backward_visible_word_start",{P},B},
	{"forward_visible_cursor_position",{P},B},
	{"backward_visible_cursor_position",{P},B},
	{"forward_visible_cursor_positions",{P,I},B},
	{"backward_visible_cursor_positions",{P,I},B},
	{"forward_visible_line",{P},B},
	{"backward_visible_line",{P},B},
	{"forward visible_lines",{P,I},B},
	{"backward_visible_lines",{P,I},B},
	{"set_offset",{P,I}},
	{"set_line",{P,I}},
	{"set_line_offset",{P,I}},
	{"set_line_index",{P,I}},
	{"set_visible_line_index",{P,I}},
	{"set_visible_line_offset",{P,I}},
	{"forward_to_end",{P}},
	{"forward_to_line_end",{P},B},
	{"forward_to_tag_toggle",{P,P},B},
	{"forward_find_char",{P,P,P,P},B},
	{"backward_find_char",{P,P,P,P},B},
	{"forward_search",{P,S,I,P,P,P},B},
	{"backward_search",{P,S,I,P,P,P},B},
	{"equal",{P,P},B},
	{"compare",{P,P},I},
	{"in_range",{P,P,P},B},
	{"order",{P,P}},
"GtkTextIter"}

control[GtkTextMark] = {"gtk_text_mark",
{GObject},
	{"new",{S,B},P},
	{"set_visible",{P,B}},
	{"get_visible",{P},B},
	{"get_deleted",{P},B},
	{"get_name",{P},S},
	{"get_buffer",{P},P},
	{"get_left_gravity",{P},B},
"GtkTextMark"}

control[GtkTextTag] = {"gtk_text_tag",
{GObject},
	{"new",{S},P},
	{"get_priority",{P},I},
	{"set_priority",{P,I}},
	{"event",{P,P,P,P},B},
"GtkTextTag"}

control[GtkTextTagTable] = {"gtk_text_tag_table",
{GObject},
	{"new",{},P},
	{"add",{P,P}},
	{"remove",{P,P}},
	{"lookup",{P,S},P},
	{"foreach",{P,P,P}},
	{"get_size",{P},I},
"GtkTextTagTable"}

control[PangoAttrList] = {"pango_attr_list",
{GObject},
	{"new",{},P},
	{"insert",{P,P}},
"PangoAttrList"}

control[PangoAttrForeground] = {"pango_attr_foreground",
{GObject},
	{"new",{I,I,I},P},
"PangoAttrForeground"}

control[PangoRenderer] = {"pango_renderer",
{GObject},
	{"new",{P},P},
	{"get_default",{P},P},
	{"set_drawable",{P,P}},
	{"set_gc",{P,P}},
"PangoRenderer"}

control[GtkCssProvider] = {"gtk_css_provider",
{GObject},
	{"new",{},P},
	{"get_default",{},P},
	{"get_named",{S,S},P},
	{"load_from_data",{P,S,I,P},B},
	{"load_from_file",{P,S},B},
	{"load_from_path",{P,S,P},B},
	{"to_string",{P},S},
"GtkCssProvider"}

control[GtkAccessible] = {"gtk_accessible",
{AtkObject},
	{"connect_widget_destroyed",{P}},
	{"get_widget",{P},P},
	{"set_widget",{P,P}},
"GtkAccessible"}

control[GtkWidgetPath] = {"gtk_widget_path",
{GObject},
	{"new",{},P},
	{"to_string",{P},S},
	{"length",{P},I},
	{"is_type",{P,I},B},
	{"has_parent",{P,I},B},
	{"get_object_type",{P},I},
	{"free",{P}},
	{"unref",{P}},
	{"ref",{P}},
	{"copy",{P},P},
	{"append_for_widget",{P,P},I},
	{"append_with_siblings",{P,P,I},I},
	{"append_type",{P,I},I},
	{"prepend_type",{P,I}},
"GtkWidgetPath"}

control[GtkWidget] = {"gtk_widget",
{GInitiallyUnowned},
	{"activate",{P},B},
	{"add_accelerator",{P,S,P,I,I,I}},
	{"add_events",{P,I}},
	{"add_mnemonic_label",{P,P}},
	{"can_activate_accel",{P,I},B},
	{"child_focus",{P,I},B},
	{"child_notify",{P,S}},
	{"class_find_style_property",{P,S},P},
	{"class_install_style_property",{P,P}},
	{"class_install_style_property_parser",{P,P,P}},
	{"class_list_style_properties",{P,I},A},
	{"compute_expand",{P,I},B},
	{"create_pango_context",{P},P},
	{"create_pango_layout",{P,S},P},
	{"destroyed",{P,P}},
	{"destroy",{P}},
	{"draw",{P,P}},
	{"error_bell",{P}},
	{"event",{P,P},B},
	{"freeze_child_notify",{P}},
	{"get_accessible",{P},P},
	{"get_allocation",{P,P}},
	{"get_allocated_width",{P},I},
	{"get_allocated_height",{P},I},
	{"get_ancestor",{P,I},P},
	{"get_app_paintable",{P},B},
	{"get_can_default",{P},B},
	{"get_can_focus",{P},B},
	{"get_child_visible",{P},B},
	{"get_clipboard",{P,I},P},
	{"get_composite_name",{P},S},
	{"get_default_direction",{},I},
	{"get_direction",{P},I},
	{"get_display",{P},P},
	{"get_double_buffered",{P},B},
	{"get_events",{P},I},
	{"get_halign",{P},I},
	{"get_has_window",{P},B},
	{"get_has_tooltip",{P},B},
	{"get_hexpand",{P},B},
	{"get_hexpand_set",{P},B},
	{"get_margin_bottom",{P},I},
	{"get_margin_top",{P},I},
	{"get_margin_left",{P},I},
	{"get_margin_right",{P},I},
	{"get_name",{P},S},
	{"get_no_show_all",{P},B},
	{"get_pango_context",{P},P},
	{"get_modifier_mask",{P,I},I},
	{"get_parent_window",{P},P},
	{"get_parent",{P},P},
	{"get_path",{P},P},
	{"get_pointer",{P,I,I}},
	{"get_preferred_size",{P,P,P}},
	{"get_preferred_height",{P,I,I}},
	{"get_preferred_height_for_width",{P,I,I,I}},
	{"get_preferred_width",{P,I,I}},
	{"get_preferred_width_for_height",{P,I,I,I}},
	{"get_receives_default",{P},B},
	{"get_request_mode",{P},I},
	{"get_root_window",{P},P},
	{"get_screen",{P},P},
	{"get_sensitive",{P},B},
	{"get_settings",{P},P},
	{"get_size_request",{P,I,I}},
	{"get_state_flags",{P},I},
	{"get_style_context",{P},P},
	{"get_support_multidevice",{P},B},
	{"get_tooltip_markup",{P},S},
	{"get_tooltip_text",{P},S},
	{"get_tooltip_window",{P},P},
	{"get_toplevel",{P},P},
	{"get_type",{P},P},
	{"get_valign",{P},I},
	{"get_vexpand",{P},B},
	{"get_vexpand_set",{P},B},
	{"get_visible",{P},B},
	{"get_visual",{P},P},
	{"get_window",{P},P},
	{"grab_default",{P}},
	{"grab_focus",{P}},
	{"has_default",{P},B},
	{"has_screen",{P},B},
	{"has_focus",{P},B},
	{"has_grab",{P},B},
	{"hide_on_delete",{P},B},
	{"hide",{P}},
	{"in_destruction",{P},B},
	{"input_shape_combine_region",{P,P}},
	{"intersect",{P,P,P},B},
	{"is_ancestor",{P,P},B},
	{"is_composited",{P},B},
	{"is_drawable",{P},B},
	{"is_focus",{P},B},
	{"is_sensitive",{P},B},
	{"is_toplevel",{P},B},
	{"keynav_failed",{P,I},B},
	{"list_accel_closures",{P},A},
	{"list_mnemonic_labels",{P},A},
	{"map",{P}},
	{"mnemonic_activate",{P,B},B},
	{"new",{I,X},P},
	{"override_color",{P,I,C}},
	{"override_background_color",{P,I,C}},
	{"override_cursor",{P,C,C}},
	{"override_font",{P,X}},
	{"override_symbolic_color",{P,S,C}},
	{"path_to_string",{P},S},
	{"pop_composite_child",{}},
	{"push_composite_child",{}},
	{"queue_compute_expand",{P}},
	{"queue_draw_area",{P,I,I,I,I}},
	{"queue_draw",{P}},
	{"queue_draw_region",{P,P}},
	{"queue_resize_no_redraw",{P}},
	{"queue_resize",{P}},
	{"realize",{P}},
	{"region_intersect",{P,P},P},
	{"remove_accelerator",{P,P,I,I},B},
	{"remove_mnemonic_label",{P,P}},
	{"render_icon_pixbuf",{P,S,I},P},
	{"reparent",{P,P}},
	{"reset_style",{P}},
	{"send_expose",{P,P},I},
	{"set_accel_path",{P,S,P}},
	{"set_allocation",{P,P}},
	{"set_app_paintable",{P,B}},
	{"set_can_default",{P,B}},
	{"set_can_focus",{P,B}},
	{"set_child_visible",{P,B}},
	{"set_composite_name",{P,S}},
	{"set_default_direction",{I}},
	{"set_device_enabled",{P,P,B}},
	{"set_device_events",{P,P,I}},
	{"set_direction",{P,I}},
	{"set_double_buffered",{P,B}},
	{"set_events",{P,I}},
	{"set_halign",{P,I}},
	{"set_has_tooltip",{P,B}},
	{"set_has_window",{P,B}},
	{"set_hexpand",{P,B}},
	{"set_haxpand_set",{P,B}},
	{"set_margin_bottom",{P,I}},
	{"set_margin_top",{P,I}},
	{"set_margin_left",{P,I}},
	{"set_margin_right",{P,I}},
	{"set_name",{P,S}},
	{"set_no_show_all",{P,B}},
	{"set_parent",{P,P}},
	{"set_parent_window",{P,P}},
	{"set_receives_default",{P,B}},
	{"set_redraw_on_allocate",{P,B}},
	{"set_sensitive",{P,B}},
	{"set_size_request",{P,I,I}},
	{"set_support_multidevice",{P,B}},
	{"set_tooltip_markup",{P,S}},
	{"set_tooltip_text",{P,S}},
	{"set_tooltip_window",{P,P}},
	{"set_valign",{P,I}},
	{"set_vexpand",{P,B}},
	{"set_vexpand_set",{P,B}},
	{"set_visible",{P,B}},
	{"set_window",{P,P}},
	{"show_all",{P}},
	{"show_now",{P}},
	{"show",{P}},
	{"size_allocate",{P,P}},
	{"style_get",{P,X}},
	{"style_get_property",{P,S,I}},
	{"style_get_valist",{P,X}},
	{"thaw_child_notify",{P}},
	{"translate_coordinates",{P,P,I,I,I,I},B},
	{"trigger_tooltip_query",{P}},
	{"unmap",{P}},
	{"unparent",{P}},
	{"unrealize",{P}},
	{"set_state_flags",{P,I,B}},
	{"unset_state_flags",{P,I}},
	{"send_expose",{P,P},I},
	{"send_focus_change",{P,P},B}, 
"GtkWidget"}

control[GtkEditable] = {"gtk_editable",
{GInterface},
	{"select_region",{P,I,I}},
	{"get_selection_bounds",{P,I,I},B},
	{"insert_text",{P,S,I,I}},
	{"delete_text",{P,I,I}},
	{"get_chars",{P,I,I},S},
	{"cut_clipboard",{P}},
	{"copy_clipboard",{P}},
	{"paste_clipboard",{P}},
	{"delete_selection",{P}},
	{"set_position",{P,I}},
	{"get_position",{P},I},
	{"set_editable",{P,B}},
	{"get_editable",{P},B},
"GtkEditable"}

control[GtkClipboard] = {"gtk_clipboard",
{GObject},
	{"get",{P},P},
	{"get_for_display",{P,P},P},
	{"get_display",{P},P},
	{"get_owner",{P},P},
	{"set_text",{P,S,I}},
	{"set_image",{P,P}},
	{"set_with_data",{P,P,I,P,P,P},B},
	{"set_with_owner",{P,P,I,P,P,P},B},
	{"request_contents",{P,P,P,P}},
	{"request_text",{P,P,P}},
	{"request_image",{P,P,P}},
	{"request_targets",{P,P,P}},
	{"request_uris",{P,P,P}},
	{"request_rich_text",{P,P,P,P}},
	{"wait_for_contents",{P,P},P},
	{"wait_for_text",{P},S},
	{"wait_for_image",{P},P},
	{"wait_for_rich_text",{P,P,P,I},P},
	{"wait_for_uris",{P},A},
	{"wait_is_text_available",{P},B},
	{"wait_is_image_available",{P},B},
	{"wait_is_rich_text_available",{P,P},B},
	{"wait_is_uris_available",{P},B},
	{"wait_for_targets",{P,A,I},B},
	{"wait_is_target_available",{P,P},B},
	{"set_can_store",{P,A,I}},
	{"store",{P}},
	{"clear",{P}},
"GtkClipboard"}

control[GtkContainer] = {"gtk_container",
{GtkWidget},
	{"add",{P,P}},
	{"add_with_properties",{P,P,X}},
	{"check_resize",{P}},
	{"child_get",{P,P,X}},
	{"child_get_property",{P,P,X}},
	{"child_get_valist",{P,P,X}},
	{"child_set",{P,P,X}},
	{"child_set_property",{P,P,X}},
	{"child_set_valist",{P,P,X}},
	{"child_type",{P},I},
	{"class_find_child_property",{P,S},P},
	{"class_install_child_property",{P,I,P}},
	{"class_list_child_properties",{P,I},P},
	{"forall",{P,P,P}},
	{"foreach",{P,P,P}},
	{"get_border_width",{P},I},
	{"get_children",{P},A},
	{"get_focus_chain",{P,A},B},
	{"get_focus_hadjustment",{P},P},
	{"get_focus_vadjustment",{P},P},
	{"get_resize_mode",{P},I},
	{"get_type",{P},P},
	{"remove",{P,P}},
	{"resize_children",{P}},
	{"set_border_width",{P,I}},
	{"set_focus_chain",{P,A}},
	{"set_focus_child",{P,P}},
	{"set_focus_hadjustment",{P,P}},
	{"set_focus_vadjustment",{P,P}},
	{"set_reallocate_redraws",{P,B}},
	{"set_resize_mode",{P,I}},
	{"unset_focus_chain",{P}},
	{"get_focus_child",{P},P},
"GtkContainer"}

control[GtkWindow] = {"gtk_window",
{GtkBin},
	{"new",{I},P},
	{"move",{P,I,I}},
	{"get_type",{},I},
	{"activate_default",{P},B},
	{"activate_focus",{P},B},
	{"activate_key",{P,P},B},
	{"add_accel_group",{P,P}},
	{"add_mnemonic",{P,I,P}},
	{"begin_move_drag",{P,I,I,I}},
	{"begin_resize_drag",{P,I,I,I,I,I}},
	{"deiconify",{P}},
	{"fullscreen",{P}},
	{"get_accept_focus",{P},B},
	{"get_application",{P},P},
	{"get_attached_to",{P},P},
	{"get_decorated",{P},B},
	{"get_default_icon_list",{},A},
	{"get_default_icon_name",{},S},
	{"get_default_size",{P,I,I}},
	{"get_default_widget",{P},P},
	{"get_deletable",{P},B},
	{"get_destroy_with_parent",{P},B},
	{"get_focus_on_map",{P},B},
	{"get_focus",{P},P},
	{"get_gravity",{P},I},
	{"get_group",{P},P},
	{"get_has_resize_grip",{P},B},
	{"get_hide_titlebar_when_maximized",{P},B},
	{"get_icon_list",{P},A},
	{"get_icon_name",{P},S},
	{"get_icon",{P},P},
	{"get_mnemonic_modifier",{P},I},
	{"get_mnemonics_visible",{P},B},
	{"get_modal",{P},B},
	{"get_opacity",{P},D},
	{"get_position",{P,I,I}},
	{"get_resizable",{P},B},
	{"get_resize_grip_area",{P,P},B},
	{"get_role",{P},S},
	{"get_screen",{P},P},
	{"get_size",{P,I,I}},
	{"get_skip_pager_hint",{P},B},
	{"get_skip_taskbar_hint",{P},B},
	{"get_title",{P},S},
	{"get_transient_for",{P},P},
	{"get_type_hint",{P},I},
	{"get_urgency_hint",{P},B},
	{"get_window_type",{P},I},
	{"has_group",{P},B},
	{"has_toplevel_focus",{P},B},
	{"iconify",{P}},
	{"is_active",{P},B},
	{"list_toplevels",{},A},
	{"maximize",{P}},
	{"mnemonic_activate",{P,I,I},B},
	{"parse_geometry",{P,S},B},
	{"present",{P}},
	{"present_with_time",{P,I}},
	{"propagate_key_event",{P,P},B},
	{"remove_accel_group",{P,P}},
	{"remove_mnemonic",{P,I,P}},
	{"reshow_with_initial_size",{P}},
	{"resize",{P,I,I}},
	{"resize_grip_is_visible",{P},B},
	{"resize_to_geometry",{P,I,I}},
	{"set_accept_focus",{P,B}},
	{"set_auto_startup_notification",{P,B}},
	{"set_application",{P,P}},
	{"set_attached_to",{P,P}},
	{"set_decorated",{P,B}},
	{"set_default_geometry",{P,I,I}},
	{"set_default_icon",{I,I},-routine_id("setDefaultIcon")}, 
	{"set_default",{P,P}},
	{"set_default_size",{P,I,I}},
	{"set_deletable",{P,B}},
	{"set_destroy_with_parent",{P,B}},
	{"set_focus_on_map",{P,B}},
	{"set_focus",{P,P}},
	{"set_focus_visible",{P,B}},-- 3.2
	{"set_geometry_hints",{P,P,P,I}},
	{"set_gravity",{P,I}},
	{"set_has_resize_grip",{P,B}},
	{"set_has_user_ref_count",{P,B}},
	{"set_hide_titlebar_when_maximized",{P,B}},
	{"set_icon_from_file",{P,S,P},B},
	{"set_icon_list",{P,A}},
	{"set_icon_name",{P,S}},
	{"set_icon",{P,P}},
	{"set_keep_above",{P,B}},
	{"set_keep_below",{P,B}},
	{"set_mnemonic_modifier",{P,I}},
	{"set_mnemonics_visible",{P,B}},
	{"set_modal",{P,B}},
	{"set_opacity",{P,D}},
	{"set_position",{P,I}},
	{"set_resizable",{P,B}},
	{"set_role",{P,S}},
	{"set_screen",{P,P}},
	{"set_skip_pager_hint",{P,B}},
	{"set_skip_taskbar_hint",{P,B}},
	{"set_startup_id",{P,S}},
	{"set_title",{P,S}},
	{"set_transient_for",{P,P}},
	{"set_type_hint",{P,I}},
	{"set_urgency_hint",{P,B}},
	{"stick",{P}},
	{"unfullscreen",{P}},
	{"unmaximize",{P}},
	{"unstick",{P}},
"GtkWindow"}

------------------------------------------------------------------------
function setDefaultIcon(atom junk, object icon)
------------------------------------------------------------------------
	if string(icon) then 
		c_proc(define_c_proc(GTK,"gtk_window_set_default_icon_from_file",{P}),
		{allocate_string(icon,1)})
	else
		c_proc(define_c_proc(GTK,"gtk_window_set_icon",{P}),{icon})
	end if
return 1
end function

control[GtkButton] = {"gtk_button",
{GtkActivatable,GtkBin},
	{"new",{},P},
	{"get_alignment",{P},-routine_id("GetButtonAlignment")},
	{"get_event_window",{P},P},
	{"get_focus_on_click",{P},B},
	{"get_image",{P},P},
	{"get_image_position",{P},I},
	{"get_label",{P},S},
	{"get_relief",{P},I},
	{"get_use_stock",{P},B},
	{"get_use_underline",{P},B},
	{"clicked",{P}},
	{"new_from_stock",{S},P},
	{"new_with_label",{S},P},
	{"new_with_mnemonic",{S},P},
	{"set_alignment",{P,F,F}},
	{"set_focus_on_click",{P,B}},
	{"set_image",{P,P}},
	{"set_image_position",{P,I}},
	{"set_label",{P,S}},
	{"set_relief",{P,I}},
	{"set_use_stock",{P,B}},
	{"set_use_underline",{P,B}},
"GtkButton"}

fn[8] = define_c_proc(GTK,"gtk_button_get_alignment",{P,P,P})
------------------------------------------------------------------------
function GetButtonAlignment(atom ctl)
------------------------------------------------------------------------
	atom x = allocate(16), y = allocate(16)
	c_proc(fn[8],{ctl,x,y})
	return {float32_to_atom(peek({x,4})),float32_to_atom(peek({y,4}))}
end function

control[GtkFontChooser] = {"gtk_font_chooser",
{GInterface,GObject},
	{"get_font_family",{P},S},
	{"get_font_face",{P},S},
	{"get_font_size",{P},I},
	{"get_font",{P},S},
	{"set_font",{P,S}},
	{"get_font_desc",{P},P},
	{"set_font_desc",{P,P}},
	{"get_preview_text",{P},S},
	{"set_preview_text",{P,S}},
	{"get_show_preview_entry",{P},B},
	{"set_show_preview_entry",{P,B}},
	{"set_filter_func",{P,P,P,P}},
"GtkFontChooser"}

control[GtkFontChooserWidget] = {"gtk_font_chooser_widget",
{GtkFontChooser,GtkOrientable,GtkBox},
"GtkFontChooserWidget"}

control[GtkFontChooserDialog] = {"gtk_font_chooser_dialog",
{GtkFontChooser,GtkDialog},
	{"new",{S,P},P},
"GtkFontChooserDialog"}

control[GtkFontSelectionDialog] = {"gtk_font_selection_dialog",
{GtkDialog},
	{"new",{S},P},
	{"set_font_name",{P,S}},
	{"get_font_name",{P},S},
	{"set_preview_text",{P,S}},
	{"get_preview_text",{P},S},
"GtkFontSelectionDialog"}

control[GtkAccelerator] = {"gtk_accelerator",
{GObject},
	{"parse",{S,I,I}},
	{"valid",{I,I},B},
	{"name",{I,I},S},
	{"get_label",{I,I},S},
	{"parse_with_keycode",{P,I,I,I}},
	{"name_with_keycode",{P,I,I,I},S},
	{"set_default_mod_mask",{I}},
	{"get_default_mod_mask",{},I},
	{"get_label_with_keycode",{P,I,I,I},S},
"GtkAccelerator"}

control[GtkAccelGroup] = {"gtk_accel_group",
{GObject},
	{"new",{},P},
	{"connect",{P,I,I,I,P}},
	{"disconnect",{P,I},B},
	{"disconnect_key",{P,I,I},B},
	{"lock",{P}},
	{"unlock",{P}},
	{"get_is_locked",{P},B},
	{"from_accel_closure",{I},P},
	{"get_modifier_mask",{P},I},
	{"activate",{P,I,I},B},
"GtkAccelGroup"}

control[GtkAccelLabel] = {"gtk_accel_label",
{GtkLabel},
	{"new",{S},P},
	{"get_accel_widget",{P},P},
	{"get_accel_width",{P},I},
	{"refetch",{P},B},
	{"set_accel_closure",{P,P}},
	{"set_accel_widget",{P,P}},
"GtkAccelLabel"}

control[GtkBox] = {"gtk_box",
{GtkOrientable,GtkContainer},
	{"new",{I,I},P},
	{"get_homogeneous",{P},B},
	{"get_spacing",{P},I},
	{"pack_end",{P,P,B,B,I}},
	{"pack_start",{P,P,B,B,I}},
	{"query_child_packing",{P,P,B,B,I,I}},
	{"reorder_child",{P,P,I}},
	{"set_child_packing",{P,P,B,B,I,I}},
	{"set_homogeneous",{P,B}},
	{"set_spacing",{P,I}},
"GtkBox"}

control[GtkBin] = {"gtk_bin",
{GtkContainer},
	{"get_child",{P},P},
"GtkBin"}

control[GtkImage] = {"gtk_image",
{GtkMisc},
	{"new",{S},P},
	{"clear",{P}},
	{"get_animation",{P},P},
	{"get_gicon",{P,P,I}},
	{"get_icon_name",{P,P,I}},
	{"get_icon_set",{P,P,I}},
	{"get_pixbuf",{P},P},
	{"get_pixel_size",{P},I},
	{"get_stock",{P,S,I}},
	{"get_storage_type",{P},I},
	{"new_from_file",{S},P},
	{"new_from_gicon",{P,I},P},
	{"new_from_icon_name",{S,I},P},
	{"new_from_icon_set",{P,I},P},
	{"new_from_pixbuf",{P},P},
	{"new_from_stock",{S,I},P},
	{"new_from_animation",{P},P},
	{"set_from_animation",{P,P}},
	{"set_from_file",{P,S}},
	{"set_from_gicon",{P,P,I}},
	{"set_from_icon_name",{P,S,I}},
	{"set_from_icon_set",{P,P,I}},
	{"set_from_pixbuf",{P,P}},
	{"set_from_stock",{P,S,I}},
	{"set_pixel_size",{P,I}},
"GtkImage"}

control[GtkLabel] = {"gtk_label",
{GtkMisc},
	{"new",{S},P},
	{"get_angle",{P},D},
	{"get_attributes",{P},A},
	{"get_current_uri",{P},S},
	{"get_ellipsize",{P},I},
	{"get_justify",{P},I},
	{"get_label",{P},S},
	{"get_layout_offsets",{P,I,I}},
	{"get_layout",{P},P},
	{"get_line_wrap_mode",{P},I},
	{"get_line_wrap",{P},B},
	{"get_max_width_chars",{P},I},
	{"get_mnemonic_keyval",{P},I},
	{"get_mnemonic_widget",{P},P},
	{"get_selectable",{P},B},
	{"get_selection_bounds",{P,I,I},B},
	{"get_single_line_mode",{P},B},
	{"get_text",{P},S},
	{"get_track_visited_links",{P},B},
	{"get_use_markup",{P},B},
	{"get_use_underline",{P},B},
	{"get_width_chars",{P},I},
	{"new_with_mnemonic",{S},P},
	{"select_region",{P,I,I}},
	{"set_angle",{P,D}},
	{"set_attributes",{P,P}},
	{"set_ellipsize",{P,I}},
	{"set_justify",{P,I}},
	{"set_label",{P,S}},
	{"set_line_wrap",{P,B}},
	{"set_line_wrap_mode",{P,I}},
	{"set_markup",{P,S}},
	{"set_markup_with_mnemonic",{P,S}},
	{"set_max_width_chars",{P,I}},
	{"set_mnemonic_widget",{P,P}},
	{"set_pattern",{P,S}}, 
	{"set_selectable",{P,B}},
	{"set_single_line_mode",{P,B}},
	{"set_text",{P,S}},
	{"set_text_with_mnemonic",{P,S}},
	{"set_track_visited_links",{P,B}},
	{"set_use_markup",{P,B}},
	{"set_use_underline",{P,B}},
	{"set_width_chars",{P,I}},
"GtkLabel"}

control[GtkButtonBox] = {"gtk_button_box",
{GtkOrientable,GtkBox},
	{"new",{I},P},
	{"get_child_secondary",{P,P},B},
	{"get_child_non_homogeneous",{P,P},B},
	{"get_layout",{P},I},
	{"set_child_secondary",{P,P,B}},
	{"set_child_non_homogeneous",{P,P,B}},
	{"set_layout",{P,I}},
"GtkButtonBox"}

control[GdkPixbuf] = {"gdk_pixbuf",
{GObject},
	{"new_from_file_at_size",{S,I,I,P},P},
	{"new_from_file_at_scale",{S,I,I,B,P},P},
	{"new_from_file",{S,P},P},
	{"get_file_info",{S,I,I},P},
	{"add_alpha",{P,B,I,I,I},P},
	{"copy_area",{P,I,I,I,I,P,I,I}},
	{"apply_embedded_orientation",{P},P},
	{"rotate_simple",{P,I},P},
	{"flip",{P,I},P},
	{"scale_simple",{P,I,I,I},P},
"GdkPixbuf"}

control[GtkDialog] = {"gtk_dialog",
{GtkWindow},
	{"add_action_widget",{P,P,I}},
	{"add_button",{P,S,I},P},
	{"get_response_for_widget",{P,P},I},
	{"get_action_area",{P},P},
	{"get_content_area",{P},P},
	{"get_widget_for_response",{P,I},P},
	{"new",{},P},
	{"new_with_buttons",{S,P,I,A}},
	{"response",{P,I}},
	{"run",{P},I},
	{"set_alternative_button_order_from_array",{P,I,A}},
	{"set_default_response",{P,I}},
	{"set_response_sensitive",{P,I,B}},
"GtkDialog"}

control[GtkAboutDialog] = {"gtk_about_dialog",
	{GtkDialog},
	{"get_artists",{P},A},
	{"get_authors",{P},A},
	{"get_comments",{P},S},
	{"get_copyright",{P},S},
	{"get_documenters",{P},A},
	{"get_license",{P},S},
	{"get_license_type",{P},I},
	{"get_logo_icon_name",{P},S},
	{"get_logo",{P},P},
	{"get_program_name",{P},S},
	{"get_translator_credits",{P},S},
	{"get_version",{P},S},
	{"get_website",{P},S},
	{"get_website_label",{P},S},
	{"get_wrap_license",{P},B},
	{"new",{},P},
	{"set_artists",{P,A}},
	{"set_authors",{P,A}},
	{"set_comments",{P,S}},
	{"set_copyright",{P,S}},
	{"set_documenters",{P,A}},
	{"set_email_hook",{P,P,P},P},
	{"set_license",{P,S}},
	{"set_license_type",{P,I}},
	{"set_logo",{P,P}},
	{"set_logo_icon_name",{P,S}},
	{"set_program_name",{P,S}},
	{"set_translator_credits",{P,S}},
	{"set_url_hook",{P,P,P},P},
	{"set_version",{P,S}},
	{"set_website",{P,S}},
	{"set_website_label",{P,S}},
	{"set_wrap_license",{P,B}},
	{"add_credit_section",{P,S,A}},
"GtkAboutDialog"}

control[GtkApplication] = {"gtk_application",
{GActionGroup,GApplication},
	{"new",{S,I},P},
	{"add_window",{P,P}},
	{"remove_window",{P,P}},
	{"get_windows",{P},A},
	{"inhibit",{P,P,I,S},I},
	{"uninhibit",{P,I}},
	{"is_inhibited",{P,I},B},
	{"end_session",{P,I,B},B},
	{"get_app_menu",{P},P},
	{"set_app_menu",{P,P}},
	{"get_menubar",{P},P},
	{"set_menubar",{P,P}},	
"GtkApplication"}

control[GtkApplicationWindow] = {"gtk_application_window",
{GtkWindow},
	{"new",{P},P},
	{"set_show_menubar",{P,B}},
	{"get_show_menubar",{P},B},
"GtkApplicationWindow"}

control[GtkAppChooser] = {"gtk_app_chooser",
{GtkWidget,GInterface},
	{"get_app_info",{P},P},
	{"get_content_type",{P},S},
	{"refresh",{P}},
"GtkAppChooser"}

control[GAppLaunchContext] = {"g_app_launch_context",
{GObject},
	{"new",{},P},
	{"get_display",{P,P,A},S},
	{"get_startup_notify_id",{P,P,A},S},
	{"failed",{P,S}},
"GAppLaunchContext"}

control[GAppInfo] = {"g_app_info",
{GObject},
	{"get_name",{P},S},
	{"get_display_name",{P},S},
	{"get_description",{P},S},
	{"get_executable",{P},S},
	{"get_commandline",{P},S},
	{"supports_files",{P},B},
	{"supports_uris",{P},B},
	{"get_icon",{P},P},
	{"launch",{P,A,P,P},B},
	{"launch_uris",{P,A,P,P},B},
	{"should_show",{P},B},
	{"can_delete",{P},B},
	{"delete",{P},B},
	{"reset_type_associations",{S}},
	{"set_as_default_for_type",{P,S,P},B},
	{"set_as_default_for_extension",{P,S,P},B},
	{"set_as_last_used_for_type",{P,S,P},B},
	{"add_supports_type",{P,S,P},B},
	{"can_remove_supports_type",{P},B},
	{"remove_supports_type",{P,S,P},B},
	{"get_all",{},A},
	{"get_all_for_type",{S},A},
	{"get_default_for_type",{S,B},P},
	{"get_default_for_uri_scheme",{S},P},
	{"get_fallback_for_type",{S},A},
	{"get_recommended_for_type",{S},A},
	{"launch_default_for_uri",{S,P,P},B},
	{"get_id",{P},S},
"GAppInfo"}

control[GtkAppChooserDialog] = {"gtk_app_chooser_dialog",
	{GtkAppChooser,GtkDialog},
	{"new",{P,I,P},P},
	{"new_for_content_type",{P,I,S},P},
	{"get_widget",{P},P},
	{"set_heading",{P,S}},
	{"get_heading",{P},S},
"GtkAppChooserDialog"}

control[GtkAppChooserWidget] = {"gtk_app_chooser_widget",
	{GtkOrientable,GtkAppChooser,GtkBox},
	{"new",{S},P},
	{"set_show_default",{P,B}},
	{"get_show_default",{P},B},
	{"set_show_recommended",{P,B}},
	{"get_show_recommended",{P},B},
	{"set_show_fallback",{P,B}},
	{"get_show_fallback",{P},B},
	{"set_show_other",{P,B}},
	{"get_show_other",{P},B},
	{"set_show_all",{P,B}},
	{"get_show_all",{P},B},
	{"set_default_text",{P,S}},
	{"get_default_text",{P},S},
"GtkFileChooserWidget"}

control[GtkAppChooserButton] = {"gtk_app_chooser_button",
	{GtkCellLayout,GtkCellEditable,GtkAppChooser,GtkComboBox},
	{"new",{S},P},
	{"append_custom_item",{P,S,S,P}},
	{"append_separator",{P}},
	{"set_active_custom_item",{P,S}},
	{"get_show_default_item",{P},B},
	{"set_show_default_item",{P,B}},
	{"get_show_dialog_item",{P},B},
	{"set_show_dialog_item",{P,B}},
	{"get_heading",{P},S},
	{"set_heading",{P,S}},
"GtkAppChooserButton"}

control[GtkColorChooser] = {"gtk_color_chooser",
{GInterface,GObject},
	{"get_rgba",{P,P}},
	{"set_rgba",{P,P}},
	{"get_use_alpha",{P},B},
	{"set_use_alpha",{P,B}},
	{"add_palette",{P,B,I,I,P}},
"GtkColorChooser"}

control[GtkColorSelection] = {"gtk_color_selection",
{GtkOrientable,GtkBox},
	{"new",{},P},
	{"get_current_rgba",{P},-routine_id("getCurrentRGBA")},
	{"set_current_rgba",{P,C}},
	{"get_current_alpha",{P},I},
	{"set_current_alpha",{P,I}},
	{"set_has_palette",{P,B}},
	{"get_has_palette",{P},B},
	{"palette_from_string",{S,A,I},B},
	{"palette_to_string",{P,I},S},
	{"set_has_opacity_control",{P,B}},
	{"get_has_opacity_control",{P},B},
	{"get_previous_rgba",{P,P}},
	{"set_previous_rgba",{P,P}},
	{"set_previous_alpha",{P,I}},
	{"get_previous_alpha",{P},I},
	{"is_adjusting",{P},B},
"GtkColorSelection"}

control[GtkColorSelectionDialog] = {"gtk_color_selection_dialog",
{GtkDialog},
	{"new",{S},P},
	{"get_current_color",{P},-routine_id("getCurrentColor")},
	{"get_color_selection",{P},-routine_id("getColorSelection")},
"GtkColorSelectionDialog"}

fn[9] = define_c_func(GTK,"gtk_color_selection_dialog_get_color_selection",{P},P)
---------------------------------------------------------------------------------
function getColorSelection(atom dlg)
------------------------------------------------------------------------
	return c_func(fn[9],{dlg})
end function

fn[10] = define_c_proc(GTK,"gtk_color_selection_get_current_rgba",{P,P})
fn[11] = define_c_func(GDK,"gdk_rgba_to_string",{P},S)
------------------------------------------------------------------------
global function getCurrentRGBA(atom sel)
------------------------------------------------------------------------
	atom rgba = allocate(32)
	c_proc(fn[10],{sel,rgba}) 
	return peek_string(c_func(fn[11],{rgba}))
end function

control[GtkFileChooserDialog] = {"gtk_file_chooser_dialog",
	{GtkFileChooser,GtkDialog},
	{"new",{S,P,I,P},P},
"GtkFileChooserDialog"}

control[GtkFileChooser] = {"gtk_file_chooser",
{GInterface},
	{"get_filename",{P},S},
	{"add_filter",{P,P}},
	{"add_shortcut_folder",{P,S,P},B},
	{"add_shortcut_folder_uri",{P,S,P},B},
	{"error_quark",{P}},
	{"get_action",{P},I},
	{"get_create_folders",{P},B},
	{"get_current_folder",{P},S},
	{"get_current_folder_file",{P},P},
	{"get_current_folder_uri",{P},S},
	{"get_do_overwrite_confirmation",{P},B},
	{"get_extra_widget",{P},P},
	{"get_filenames",{P},A},
	{"get_files",{P},A},
	{"get_file",{P},P},
	{"get_filter",{P},P},
	{"get_local_only",{P},B},
	{"get_preview_file",{P},P},
	{"get_preview_filename",{P},S},
	{"get_preview_uri",{P},S},
	{"get_preview_widget_active",{P},B},
	{"get_preview_widget",{P},P},
	{"get_select_multiple",{P},B},
	{"get_show_hidden",{P},B},
	{"get_type",{P},P},
	{"get_uri",{P},S},
	{"get_uris",{P},A},
	{"get_use_preview_label",{P},B},
	{"list_filters",{P},A},
	{"list_shortcut_folders",{P},A},
	{"list_shortcut_folder_uris",{P},A},
	{"remove_filter",{P,P}},
	{"remove_shortcut_folder",{P,S,P},B},
	{"remove_shortcut_folder_uri",{P,S,P},B},
	{"select_all",{P}},
	{"select_file",{P,P,P},B},
	{"select_filename",{P,S},B},
	{"select_uri",{P,S},B},
	{"set_action",{P,I}},
	{"set_create_folders",{P,B}},
	{"set_current_folder",{P,S},B},
	{"set_current_folder_file",{P,P,P},B},
	{"set_current_folder_uri",{P,S},B},
	{"set_current_name",{P,S}},
	{"set_do_overwrite_confirmation",{P,B}},
	{"set_extra_widget",{P,P}},
	{"set_file",{P,P,P},B},
	{"set_filename",{P,S}},
	{"set_filter",{P,P}},
	{"set_local_only",{P,B}},
	{"set_preview_widget_active",{P,B}},
	{"set_preview_widget",{P,P}},
	{"set_select_multiple",{P,B}},
	{"set_show_hidden",{P,B}},
	{"set_uri",{P,S}},
	{"set_use_preview_label",{P,B}},
	{"unselect_all",{P}},
	{"unselect_file",{P,P}},
	{"unselect_filename",{P,S}},
	{"unselect_uri",{P,S}},
"GtkFileChooser"}

control[GtkFileChooserWidget] = {"gtk_file_chooser_widget",
{GtkOrientable,GtkFileChooser,GtkBox},
	{"new",{I},P},
	{"new_with_backend",{I,S},P},
"GtkFileChooserWidget"}

control[GtkFileChooserButton] = {"gtk_file_chooser_button",
{GtkFileChooser,GtkOrientable,GtkBox},
	{"get_focus_on_click",{P},B},
	{"get_title",{P},S},
	{"get_width_chars",{P},I},
	{"new_with_backend",{S,I,S},P},
	{"new_with_dialog",{P},P},
	{"new",{S,I},P},
	{"set_focus_on_click",{P,B}},
	{"set_title",{P,S}},
	{"set_width_chars",{P,I}},
"GtkFileChooserButton"}

control[GtkFixed] = {"gtk_fixed",
{GtkContainer},
	{"move",{P,P,I,I}},
	{"new",{},P},
	{"put",{P,P,I,I}},
"GtkFixed"}

control[GtkFontButton] = {"gtk_font_button",
{GtkActivatable,GtkFontChooser,GtkButton},
	{"get_font_name",{P},S},
	{"get_show_size",{P},B},
	{"get_show_style",{P},B},
	{"get_title",{P},S},
	{"get_use_font",{P},B},
	{"get_use_size",{P},B},
	{"new",{},P},
	{"new_with_font",{S},P},
	{"set_font_name",{P,S},B},
	{"set_show_size",{P,B}},
	{"set_show_style",{P,B}},
	{"set_title",{P,S}},
	{"set_use_font",{P,B}},
	{"set_use_size",{P,B}},
"GtkFontButton"}

control[GtkFrame] = {"gtk_frame",
{GtkBin},
	{"get_label",{P},S},
	{"get_label_align",{P},-routine_id("GetLabelAlign")},
	{"get_label_widget",{P},P},
	{"get_shadow_type",{P},I},
	{"new",{S},P},
	{"set_label",{P,S}},
	{"set_label_align",{P,F,F}},
	{"set_label_widget",{P,P}},
	{"set_shadow_type",{P,I}},
"GtkFrame"}

fn[12] = define_c_proc(GTK,"gtk_frame_get_label_align",{P,P,P})
------------------------------------------------------------------------
function GetLabelAlign(atom frame)
------------------------------------------------------------------------
	atom x = allocate(16), y = allocate(16)
	c_proc(fn[12],{frame,x,y})
	return {float32_to_atom(peek({x,4})),float32_to_atom(peek({y,4}))}
end function

control[GtkScrollable] = {"gtk_scrollable",
{GInterface},
	{"get_hadjustment",{P},P},
	{"set_hadjustment",{P,P}},
	{"get_vadjustment",{P},P},
	{"set_vadjustment",{P,P}},
	{"get_hscroll_policy",{P},I},
	{"set_hscroll_policy",{P,I}},
	{"get_vscroll_policy",{P},I},
	{"set_vscroll_policy",{P,I}},
"GtkScrollable"}

control[GtkScrollbar] = {"gtk_scrollbar",
{GtkOrientable,GtkRange},
	{"new",{I,P},P},
"GtkScrollbar"}

control[GtkSeparator] = {"gtk_separator",
{GtkOrientable,GtkWidget},
{"new",{I},P}, 
"GtkSeparator"}

control[GtkStatusIcon] = {"gtk_status_icon",
{GObject},
	{"new",{},P},
	{"new_from_stock",{S},P},
	{"new_from_file",{S},P},
	{"new_from_pixbuf",{P},P},
	{"new_from_icon_name",{S},P},
	{"new_from_gicon",{P},P},
	{"position_menu",{P,I,I,B,P}},
	{"set_from_stock",{P,S}},
	{"set_from_file",{P,S}},
	{"set_from_pixbuf",{P,P}},
	{"set_from_icon_name",{P,S}},
	{"set_from_gicon",{P,P}},
	{"set_has_tooltip",{P,B}},
	{"set_name",{P,S}},
	{"set_title",{P,S}},
	{"set_visible",{P,B}},
	{"get_visible",{P},B},
	{"get_geometry",{P,P,P,I},B},
	{"get_storage_type",{P},I},
	{"get_pixbuf",{P},P},
	{"get_stock",{P},S},
	{"get_icon_name",{P},S},
	{"get_gicon",{P},P},
	{"get_has_tooltip",{P},B},
	{"get_size",{P},I},
	{"get_screen",{P},P},
	{"set_screen",{P,P}},
	{"get_title",{P,S}},
	{"set_tooltip_text",{P,S}},
	{"get_tooltip_text",{P},S},
	{"set_tooltip_markup",{P,S}},
	{"get_tooltip_markup",{P},S},
	{"get_visible",{P},B},
	{"is_embedded",{P},B},
	{"position_menu",{P,I,I,B,P}},
	{"get_geometry",{P,P,P,P},B},
	{"get_x11_window_id",{P},I},
"GtkStatusIcon"}

control[GtkIconView] = {"gtk_icon_view",
	{GtkCellLayout,GtkScrollable,GtkContainer},
	{"convert_widget_to_bin_window_coords",{P,I,I,I,I}},
	{"create_drag_icon",{P,P},P},
	{"enable_model_drag_dest",{P,P,I,I}},
	{"enable_model_drag_source",{P,I,P,I,I}},
	{"get_columns",{P},I},
	{"get_column_spacing",{P},I},
	{"get_cursor",{P,P,P},B},
	{"get_dest_item_at_pos",{P,I,I,P,I},B},
	{"get_drag_dest_item",{P,P,I}},
	{"get_item_at_pos",{P,I,I,P,P},B},
	{"get_item_column",{P,P},I},
	{"get_item_orientation",{P},I},
	{"get_item_padding",{P},I},
	{"get_item_row",{P,P},I},
	{"get_item_width",{P},I},
	{"get_margin",{P},I},
	{"get_markup_column",{P},I},
	{"get_model",{P},P},
	{"get_orientation",{P},I},
	{"get_path_at_pos",{P,I,I},P},
	{"get_pixbuf_column",{P},I},
	{"get_reorderable",{P},B},
	{"get_row_spacing",{P},I},
	{"get_selected_items",{P},A},
	{"get_selection_mode",{P},I},
	{"get_spacing",{P},I},
	{"get_text_column",{P},I},
	{"get_tooltip_column",{P},I},
	{"get_tooltip_context",{P,I,I,B,P,P,P},B},
	{"get_visible_range",{P,P,P},B},
	{"item_activated",{P,P}},
	{"new",{},P},
	{"new_with_model",{P},P},
	{"path_is_selected",{P,P},B},
	{"scroll_to_path",{P,P,B,F,F}},
	{"select_all",{P}},
	{"selected_foreach",{P,P,P}},
	{"select_path",{P,P}},
	{"set_columns",{P,I}},
	{"set_column_spacing",{P,I}},
	{"set_cursor",{P,P,P,B}},
	{"set_drag_dest_item",{P,P,I}},
	{"set_item_orientation",{P,I}},
	{"set_item_padding",{P,I}},
	{"set_item_width",{P,I}},
	{"set_margin",{P,I}},
	{"set_markup_column",{P,I}},
	{"set_model",{P,P}},
	{"set_pixbuf_column",{P,I}},
	{"set_reorderable",{P,B}},
	{"set_row_spacing",{P,I}},
	{"set_selection_mode",{P,I}},
	{"set_spacing",{P,I}},
	{"set_text_column",{P,I}},
	{"set_tooltip_cell",{P,P,P,P}},
	{"set_tooltip_column",{P,I}},
	{"set_tooltip_item",{P,P,P}},
	{"unselect_all",{P}},
	{"unselect_path",{P,P}},
	{"unset_model_drag_dest",{P}},
	{"unset_model_drag_source",{P}},
"GtkIconView"}

control[GtkIconTheme] = {"gtk_icon_theme",
{GObject},
	{"add_builtin_icon",{P,I,P}},
	{"append_search_path",{P,S}},
	{"choose_icon",{P,S,I,I},P},
	{"error_quark",{P}},
	{"get_default",{},P},
	{"get_example_icon_name",{P},S},
	{"get_for_screen",{P},P},
	{"get_icon_sizes",{P,S},I},
	{"get_search_path",{P,P,I}},
	{"get_search_path",{P},P},
	{"has_icon",{P,S},B},
	{"list_contexts",{P},A},
	{"list_icons",{P,S},A},
	{"load_icon",{P,S,I,I,P},P},
	{"lookup_icon",{P,S,I,I},P},
	{"lookup_by_gicon",{P,P,I,I},P},
	{"new",{},P},
	{"prepend_search_path",{P,S}},
	{"rescan_if_needed",{P},B},
	{"set_custom_theme",{P,S}},
	{"set_screen",{P,P}},
	{"set_search_path",{P,S,I}},
	{"set_search_path",{P}},
"GtkIconTheme"}

control[GtkImageMenuItem] = {"gtk_image_menu_item",
{GtkActivatable,GtkMenuItem},
	{"get_image",{P},P},
	{"new_from_stock",{S,P},P},
	{"new",{},P},
	{"new_with_label",{S},P},
	{"new_with_mnemonic",{S},P},
	{"set_image",{P,P}},
"GtkImageMenuItem"}

control[GtkInvisible] = {"gtk_invisible",
{GtkWidget},
	{"new",{},P},
	{"new_for_screen",{P},P},
	{"set_screen",{P,P}},
	{"get_screen",{P},P},
"GtkInvisible"}

control[GtkSizeGroup] = {"gtk_size_group",
{GObject},
	{"new",{I},P},
	{"set_mode",{P,I}},
	{"get_mode",{P},I},
	{"set_ignore_hidden",{P,B}},
	{"get_ignore_hidden",{P},B},
	{"add_widget",{P,P}},
	{"remove_widget",{P,P}},
	{"get_widgets",{P},A},
"GtkSizeGroup"}

control[GtkLayout] = {"gtk_layout",
{GtkScrollable,GtkContainer},
	{"get_bin_window",{P},P},
	{"get_size",{P,I,I}},
	{"move",{P,P,I,I}},
	{"new",{P,P},P},
	{"put",{P,P,I,I}},
	{"set_size",{P,I,I}},
"GtkLayout"}

control[GInterface] = {"",0,
"GInterface"}

control[GtkNumerableIcon] = {"gtk_numerable_icon",
{GIcon,GEmblemedIcon,GObject},
	{"new",{P},P},
	{"new_with_style_context",{P,I},P},
	{"get_background_gicon",{P},P},
	{"set_background_gicon",{P,P}},
	{"get_background_icon_name",{P},S},
	{"set_background_icon_name",{P,S}},
	{"get_count",{P},I},
	{"set_count",{P,I}},
	{"get_label",{P},S},
	{"set_label",{P,S}},
	{"get_style_context",{P},P},
	{"set_style_context",{P,P}},
"GtkNumerableIcon"}

control[GtkLinkButton] = {"gtk_link_button",
	{GtkActivatable,GtkButton},
	{"new",{S},P},
	{"new_with_label",{S,S},P},
	{"get_uri",{P},S},
	{"set_uri",{P,S}},
	{"get_visited",{P},B},
	{"set_visited",{P,B}},
"GtkLinkButton"}

control[GtkListStore] = {"gtk_list_store",
{GtkTreeModel,GtkTreeSortable,GObject},
	{"append",{P,P}},
	{"clear",{P}},
	{"insert",{P,P,I}},
	{"insert_after",{P,P,P}},
	{"insert_before",{P,P,P}},
	{"insert_with_values",{P,P,I,P,I,P,I,P,I,P,I,P}},
	{"insert_with_valuesv",{P,P,I,I,A,I}},
	{"move_after",{P,P,P}},
	{"move_before",{P,P,P}},
	{"new",{P},-routine_id("createListStore")},
	{"newv",{P},P}, 
	{"prepend",{P,P}},
	{"remove",{P,P},B},
	{"reorder",{P,A}},
	{"set_col_data",{P,P,I,P},-routine_id("setListStoreColumnData")},
	{"set_row_data",{P,P,P},-routine_id("setListStoreRowData")},
	{"get_col_data",{P,P,I},-routine_id("getListStoreColumnData")},
	{"set_column_types",{P,I,A}},
	{"set_value",{P,P,I,P}},
	{"swap",{P,P,P}},
"GtkListStore"}

------------------------------------------------------------------------
global function createListStore(object params)
------------------------------------------------------------------------
integer fail = FALSE

if atom(params) then 
	crash("Error: parameters must be a {sequence}")
end if

for i = 1 to length(params) do
	switch params[i] do
		case gVOID,gINTERFACE,gENUM,gBOXED,gPARAM,gVARIANT then
		fail = TRUE params = params[i]
	end switch
end for

if fail then 
	crash("Error: cannot create a list store with column type %d \n",params)
end if

object proto = I & repeat(P,length(params))
params = length(params) & params
atom x = define_c_func(GTK,"gtk_list_store_new",proto,P)
return c_func(x,params)

end function

constant fn_n_cols = define_c_func(GTK,"gtk_tree_model_get_n_columns",{P},I)
--------------------------------------------------------------------------
global function setListStoreRowData(atom self, atom iter, object data)
--------------------------------------------------------------------------
integer max_col =  c_func(fn_n_cols,{self})
if max_col > length(data) then max_col = length(data) end if

for i = 1 to max_col do
	iter = setListStoreColumnData(self,iter,i,data[i])
end for

return iter
end function

constant fn_col_type = define_c_func(GTK,"gtk_tree_model_get_column_type",{P,I},I)
--------------------------------------------------------------------------------------
global function setListStoreColumnData(atom self, atom iter, integer col, object data)
--------------------------------------------------------------------------------------
object proto = {P,P,I,P,I}
object params
atom fn

if iter < 0 then -- invalid (no row selected?)
	return 1
end if

integer max_col = c_func(fn_n_cols,{self})
if col > max_col or col < 1 then 
	crash("Set column data: column %d does not exist! Max = %d",{col,max_col})
end if

integer ct = c_func(fn_col_type,{self,col-1})

switch ct do
	case gFLOAT then proto={P,P,I,D,I} if string(data) then data = to_number(data) end if
	case gSTR then proto = {P,P,I,S,I} 
	case gINT then proto = {P,P,I,P,I} if string(data) then data = to_number(data) end if
end switch

if string(data) then data = allocate_string(data,1) end if

fn = define_c_proc(GTK,"gtk_list_store_set",proto)
params = {self,iter,col-1,data,-1}
c_proc(fn,params)

return iter
end function

control[GtkMain] = {"gtk",GObject,
	{"check_version",{P}},
	{"disable_setlocale",{}},
	{"events_pending",{},B},
	{"false",{},B},
	{"get_current_event",{},P},
	{"get_current_event_state",{I},B},
	{"get_current_event_time",{},I},
	{"get_default_language",{},P},
	{"get_event_widget",{I},P},
	{"get_option_group",{B},P},
	{"grab_add",{P}},
	{"grab_get_current",{},P},
	{"grab_remove",{P}},
	{"init_add",{P,P}},
	{"init_check",{I,A},B},
	{"init",{I,P}},
	{"init_with_args",{I,A,S,P,S,P},B},
	{"main_do_event",{P}},
	{"main_iteration_do",{B},B},
	{"main_iteration",{},B},
	{"main_level",{},I},
	{"main",{}},
	{"main_quit",{}},
	{"parse_args",{I,A},B},
	{"propagate_event",{P,P}},
	{"quit_add_destroy",{I,P}},
	{"quit_add_full",{I,P,P,P,P},I},
	{"quit_add",{I,P,P},I},
	{"quit_remove_by_data",{P}},
	{"quit_remove",{I}},
	{"set_locale",{},S},
	{"true",{},B},
"GtkMain"}

control[GtkMenuBar] = {"gtk_menu_bar",
{GtkMenuShell},
	{"get_child_pack_direction",{P},I},
	{"get_pack_direction",{P},I},
	{"new",{},P},
	{"new_from_model",{P},P},
	{"set_child_pack_direction",{P,I}},
	{"set_pack_direction",{P,I}},
"GtkMenuBar"}

control[GtkMenu] = {"gtk_menu",
{GtkMenuShell},
	{"new",{},P},
	{"new_from_model",{P},P},
	{"attach",{P,P,I,I,I,I}},
	{"attach_to_widget",{P,P,P}},
	{"detach",{P}},
	{"get_accel_group",{P},P},
	{"get_accel_path",{P},S},
	{"get_active",{P},P},
	{"get_attach_widget",{P},P},
	{"get_for_attach_widget",{P},A},
	{"get_monitor",{P},I},
	{"get_reserve_toggle_size",{P},B},
	{"get_tearoff_state",{P},B},
	{"get_title",{P},S},
	{"popdown",{P}},
	{"popup",{P,P,P,P,P,I,I}},
	{"popup_for_device",{P,P,P,P,P,P,P,I,I}},
	{"reorder_child",{P,P,I}},
	{"reposition",{P}},
	{"set_accel_group",{P,P}},
	{"set_accel_path",{P,S}},
	{"set_active",{P,I}},
	{"set_monitor",{P,I}},
	{"set_reserve_toggle_size",{P,B}},
	{"set_screen",{P,P}},
	{"set_tearoff_state",{P,B}},
	{"set_title",{P,S}},
"GtkMenu"}

control[GtkMenuItem] = {"gtk_menu_item",
{GtkActivatable,GtkBin},
	{"new",{},P},
	{"activate",{P}},
	{"deselect",{P}},
	{"get_label",{P},S},
	{"get_reserve_indicator",{P},B},
	{"get_submenu",{P},P},
	{"get_use_underline",{P},B},
	{"new_with_mnemonic",{S},P},
	{"new_with_label",{S},P},
	{"select",{P}},
	{"set_accel_path",{P,S}},
	{"set_label",{P,S}},
	{"set_reserve_indicator",{P,B}},
	{"set_submenu",{P,P}},
	{"set_use_underline",{P,B}},
	{"toggle_size_allocate",{P,I}},
	{"toggle_size_request",{P,I}},
"GtkMenuItem"}

control[GtkMenuShell] = {"gtk_menu_shell",
{GtkContainer},
	{"activate_item",{P,P,B}},
	{"append",{P,P}},
	{"cancel",{P}},
	{"deactivate",{P}},
	{"deselect",{P}},
	{"get_take_focus",{P},B},
	{"get_parent_shell",{P},P},
	{"get_selected_item",{P},P},
	{"insert",{P,P,I}},
	{"prepend",{P,P}},
	{"select_first",{P,B}},
	{"select_item",{P,P}},
	{"set_take_focus",{P,B}},
"GtkMenuShell"}

control[GtkMenuToolButton] = {"gtk_menu_tool_button",
{GtkActivatable,GtkToolButton},
	{"get_menu",{P},P},
	{"new",{P,S},P},
	{"new_from_stock",{S},P},
	{"set_arrow_tooltip_markup",{P,S}},
	{"set_arrow_tooltip_text",{P,S}},
	{"set_menu",{P,P}},
"GtkMenuToolButton"}

control[GtkMessageDialog] = {"gtk_message_dialog",
	{GtkDialog},
	{"format_secondary_markup",{P,S,S}},
	{"format_secondary_text",{P,S,S}},
	{"get_image",{P},P},
	{"get_message_area",{P},P},
	{"new",{P,I,I,I,S,S},P},
	{"new_with_markup",{P,I,I,I,S,S},P},
	{"set_image",{P,P}},
	{"set_markup",{P,S}},
"GtkMessageDialog"}

control[GtkMisc] = {"gtk_misc",
{GtkWidget},
	{"get_alignment",{P},-routine_id("GetAlignment")},
	{"get_padding",{P,I,I}},
	{"set_alignment",{P,D,D}},
	{"set_padding",{P,I,I}},
"GtkMisc"}

fn[13] = define_c_proc(GTK,"gtk_misc_get_alignment",{P,P,P})
------------------------------------------------------------------------
function GetAlignment(atom ctl)
------------------------------------------------------------------------
	atom x = allocate(16), y = allocate(16)
	c_proc(fn[13],{ctl,x,y})
	return {float32_to_atom(peek({x,4})),float32_to_atom(peek({y,4}))}
end function

control[GtkGrid] = {"gtk_grid",
{GtkOrientable,GtkContainer,GtkWidget},
	{"new",{},P},
	{"attach",{P,P,I,I,I,I}},
	{"attach_next_to",{P,P,P,I,I,I}},
	{"get_child_at",{P,I,I},P},
	{"insert_row",{P,I}},
	{"insert_column",{P,I}},
	{"insert_next_to",{P,P,I}},
	{"set_row_homogeneous",{P,B}},
	{"get_row_homogeneous",{P},B},
	{"set_row_spacing",{P,I}},
	{"get_row_spacing",{P},I},
	{"set_column_homogeneous",{P,B}},
	{"get_colunm_homogeneous",{P},B},
	{"set_column_spacing",{P,I}},
	{"get_column_spacing",{P},I},
"GtkGrid"}

control[GtkCellEditable] = {"gtk_cell_editable",
{GInterface},
	{"start_editing",{P,P}},
	{"editing_done",{P}},
	{"remove_widget",{P}},
"GtkCellEditable"}

control[GtkCellLayout] = {"gtk_cell_layout",
{GInterface,GObject},
	{"pack_start",{P,P,B}},
	{"pack_end",{P,P,B}},
	{"get_area",{P},P},
	{"get_cells",{P},A},
	{"reorder",{P,P,I}},
	{"clear",{P}},
	{"set_attributes",{P,P,P}},
	{"add_attribute",{P,P,S,I}},
	{"set_cell_data_func",{P,P,P,P,P}},
	{"clear_attributes",{P,P}},
"GtkCellLayout"}

control[GtkOffscreenWindow] = {"gtk_offscreen_window",
{GtkWindow},
	{"new",{},P},
	{"get_surface",{P},P},
	{"get_pixbuf",{P},P},
"GtkOffscreenWindow"}

control[GtkInvisible] = {"gtk_invisible",
{GtkWidget},
	{"new",{},P},
	{"new_for_screen",{P},P},
	{"set_screen",{P,P}},
	{"get_screen",{P},P},
"GtkInvisible"}

control[GtkNotebook] = {"gtk_notebook",
{GtkContainer},
	{"append_page_menu",{P,P,P,P},I},
	{"append_page",{P,P,P},I},
	{"get_action_widget",{P,I},P},
	{"get_current_page",{P},I},
	{"get_group",{P},P},
	{"get_group_name",{P},S},
	{"get_menu_label",{P,P},P},
	{"get_menu_label_text",{P,P},S},
	{"get_n_pages",{P},I},
	{"get_nth_page",{P,I},P},
	{"get_scrollable",{P},B},
	{"get_show_border",{P},B},
	{"get_show_tabs",{P},B},
	{"get_tab_detachable",{P,P},B},
	{"get_tab_label",{P,P},P},
	{"get_tab_label_text",{P,P},S},
	{"get_tab_pos",{P},I},
	{"get_tab_reorderable",{P,P},B},
	{"insert_page_menu",{P,P,P,P,I},I},
	{"insert_page",{P,P,P,I},I},
	{"new",{},P},
	{"next_page",{P}},
	{"page_num",{P,P},I},
	{"popup_disable",{P}},
	{"popup_enable",{P}},
	{"prepend_page_menu",{P,P,P,P},I},
	{"prepend_page",{P,P,P},I},
	{"prev_page",{P}},
	{"query_tab_label_packing",{P,P,B,B,I}},
	{"remove_page",{P,I}},
	{"reorder_child",{P,P,I}},
	{"set_action_widget",{P,P,I}},
	{"set_current_page",{P,I}},
	{"set_group",{P,P}},
	{"set_group_name",{P,S}},
	{"set_menu_label",{P,P,P}},
	{"set_menu_label_text",{P,P,S}},
	{"set_scrollable",{P,B}},
	{"set_show_border",{P,B}},
	{"set_show_tabs",{P,B}},
	{"set_tab_detachable",{P,P,B}},
	{"set_tab_label",{P,P,P}},
	{"set_tab_label_packing",{P,P,B,B,I}},
	{"set_tab_label_text",{P,P,S}},
	{"set_tab_pos",{P,I}},
	{"set_tab_reorderable",{P,P,B}},
	{"set_window_creation_hook",{P,P,P}},
"GtkNotebook"}

control[GtkPageSetup] = {"gtk_page_setup",
{GObject},
	{"copy",{P},P},
	{"get_bottom_margin",{P,I},D},
	{"get_left_margin",{P,I},D},
	{"get_orientation",{P},I},
	{"get_page_height",{P},P},
	{"get_page_width",{P,I},D},
	{"get_paper_height",{P,I},D},
	{"get_paper_size",{P},P},
	{"get_paper_width",{P,I},D},
	{"get_right_margin",{P,I},D},
	{"get_top_margin",{P,I},D},
	{"new_from_file",{S,P},P},
	{"new_from_key_file",{P,S,P},P},
	{"new",{},P},
	{"set_bottom_margin",{P,D,I}},
	{"set_left_margin",{P,D,I}},
	{"set_orientation",{P,I}},
	{"set_paper_size_and_default_margins",{P,P}},
	{"set_paper_size",{P,P}},
	{"set_right_margin",{P,D,I}},
	{"set_top_margin",{P,D,I}},
	{"to_file",{P,S,P},B},
	{"to_key_file",{P,P,S}},
	{"run",{P,P,P,P},-routine_id("runPageSetupDialog")},
"GtkPageSetup"}

------------------------------------------------------------------------------
function runPageSetupDialog(atom self, atom parent, atom setup, atom settings)
------------------------------------------------------------------------------
return c_func(define_c_func(GTK,"gtk_print_run_page_setup_dialog",{P,P,P},P),
	{parent,setup,settings})
end function

control[GtkPageSetupUnixDialog] = {"gtk_page_setup_unix_dialog",
	{GtkDialog},
	{"get_page_setup",{P},P},
	{"get_print_settings",{P},P},
	{"new",{S,P},P},
	{"set_page_setup",{P,P}},
	{"set_print_settings",{P,P}},
"GtkPageSetupUnixDialog"}

control[GtkPaned] = {"gtk_paned",
{GtkOrientable,GtkContainer},
	{"new",{I},P},
	{"add1",{P,P}},
	{"add2",{P,P}},
	{"compute_position",{P}},
	{"get_child1",{P},P},
	{"get_child2",{P},P},
	{"get_position",{P},I},
	{"pack1",{P,P,B,B}},
	{"pack2",{P,P,B,B}},
	{"set_position",{P,I}},
"GtkPaned"}

control[GtkSwitch] = {"gtk_switch",
{GtkActivatable,GtkWidget},
	{"new",{},P},
	{"set_active",{P,B}},
	{"get_active",{P},B},
"GtkSwitch"}

control[GtkPlug] = {"gtk_plug",
{GtkWindow},
	{"construct_for_display",{P,P,I}},
	{"construct",{P,I}},
	{"get_embedded",{P},B},
	{"get_socket_window",{P},P},
	{"get_id",{P},I},
	{"new",{I},P},
	{"new_for_display",{P,I},P},
"GtkPlug"}

control[GtkSocket] = {"gtk_socket",
{GtkContainer},
	{"new",{},P},
	{"add_id",{P,I}},
	{"get_id",{P},I},
	{"get_plug_window",{P},P},
"GtkSocket"}

control[GtkPrintContext] = {"gtk_print_context",
{GObject},
	{"create_pango_context",{P},P},
	{"create_pango_layout",{P},P},
	{"get_cairo_context",{P},P},
	{"get_dpi_x",{P},D},
	{"get_dpi_y",{P},D},
	{"get_hard_margins",{P,D,D,D,D},B},
	{"get_height",{P},D},
	{"get_page_setup",{P},P},
	{"get_pango_fontmap",{P},P},
	{"get_width",{P},D},
	{"set_cairo_context",{P,I,D,D}},
"GtkPrintContext"}

control[GtkPaperSize] = {"gtk_paper_size",
{GObject},
	{"new",{S},P},
	{"new_custom",{S,S,D,D,I},P},
	{"new_from_key_file",{P,S,P},P},
	{"new_from_ppd",{P,S,D,D},P},
	{"to_key_file",{P,P,S}},
	{"get_paper_sizes",{B},A},
	{"get_name",{P},S},
	{"get_display_name",{P},S},
	{"get_ppd_name",{P},S},
	{"get_width",{P,I},D},
	{"get_height",{P,I},D},
	{"is_custom",{P},B},
	{"is_equal",{P,P},B},
	{"set_size",{P,D,D,I}},
	{"get_default_top_margin",{P,I},D},
	{"get_default_bottom_margin",{P,I},D},
	{"get_default_left_margin",{P,I},D},
	{"get_default_right_margin",{P,I},D},
	{"get_default",{},S},
"GtkPaperSize"}

control[GtkPrinter] = {"gtk_printer",
{GObject},
	{"accepts_pdf",{P},B},
	{"accepts_ps",{P},B},
	{"compare",{P,P},I},
	{"get_backend",{P},P},
	{"get_capabilities",{P},I},
	{"get_description",{P},S},
	{"get_hard_margins",{P,D,D,D,D},B},
	{"get_icon_name",{P},S},
	{"get_job_count",{P},I},
	{"get_location",{P},S},
	{"get_name",{P},S},
	{"get_state_message",{P},S},
	{"has_details",{P},B},
	{"is_active",{P},B},
	{"is_default",{P},B},
	{"is_new",{P},B},
	{"is_virtual",{P},B},
	{"list_papers",{P},A},
	{"new",{S,P,B},P},
	{"request_details",{P}},
	{"set_description",{P}},
	{"set_has_details",{P}},
	{"set_icon_name",{P}},
	{"set_is_active",{P}},
	{"set_is_default",{P}},
	{"set_is_new",{P}},
	{"set_job_count",{P}},
	{"set_location",{P}},
	{"set_state_message",{P}},
"GtkPrinter"}

control[GtkPrintJob] = {"gtk_print_job",
{GObject},
	{"new",{S,P,P,P},P},
	{"get_collate",{P},B},
	{"set_collate",{P,B}},
	{"get_reverse",{P},B},
	{"set_reverse",{P,B}},
	{"get_pages",{P},I},
	{"set_pages",{P,I}},
	{"get_page_ranges",{P,I},P},
	{"set_page_ranges",{P,P,I}},
	{"get_page_set",{P},I},
	{"set_page_set",{P,I}},
	{"get_num_copies",{P},I},
	{"set_num_copies",{P,I}},
	{"get_scale",{P},D},
	{"set_scale",{P,D}},
	{"get_n_up",{P},I},
	{"set_n_up",{P,I}},
	{"get_n_up_layout",{P},I},
	{"set_n_up_layout",{P,I}},
	{"get_rotate",{P},B},
	{"set_rotate",{P,B}},
	{"get_printer",{P},P},
	{"get_settings",{P},P},
	{"get_status",{P},I},
	{"get_surface",{P,P},P},
	{"get_title",{P},S},
	{"get_track_print_status",{P},B},
	{"send",{P,P,P,P}},
	{"set_source_file",{P,S,P},B},
	{"set_status",{P}},
	{"set_track_print_status",{P,B}},
"GtkPrintJob"}

control[GtkPrintOperation] = {"gtk_print_operation",
{GObject},
	{"cancel",{P}},
	{"get_default_page_setup",{P},P},
	{"get_error",{P,P}},
	{"get_embed_page_setup",{P},B},
	{"get_has_selection",{P},B},
	{"get_manual_capabilities",{P},I},
	{"get_n_pages_to_print",{P},I},
	{"get_page_setup_set",{P},B},
	{"get_print_settings",{P},P},
	{"get_status",{P},I},
	{"get_status_string",{P},S},
	{"get_support_selection",{P},B},
	{"is_finished",{P},B},
	{"new",{},P},
	{"run",{P,I,P,P},I},
	{"set_allow_async",{P,B}},
	{"set_current_page",{P,I}},
	{"set_custom_tab_label",{P,S}},
	{"set_default_page_setup",{P,P}},
	{"set_embed_page_setup",{P,B}},
	{"set_export_filename",{P,S}},
	{"set_has_selection",{P,B}},
	{"set_job_name",{P,S}},
	{"set_n_pages",{P,I}},
	{"set_print_settings",{P,P}},
	{"set_show_progress",{P,B}},
	{"set_support_selection",{P,B}},
	{"set_track_print_status",{P,B}},
	{"set_unit",{P,I}},
	{"set_use_full_page",{P,B}},
	{"preview_end_preview",{P}},
	{"preview_get_type",{P},P},
	{"preview_is_selected",{P,I},B},
	{"preview_render_page",{P,I}},
"GtkPrintOperation"}

control[GtkPrintSettings] = {"gtk_print_settings",
{GObject},
	{"copy",{P},P},
	{"foreach",{P,P,P}},
	{"get_bool",{P,S},B},
	{"get_collate",{P},B},
	{"get_default_source",{P},S},
	{"get_dither",{P},S},
	{"get_double",{P,S},D},
	{"get_double_with_default",{P,S,D},D},
	{"get_duplex",{P},B},
	{"get_finishings",{P},S},
	{"get_int",{P,S},I},
	{"get_int_with_default",{P,S,I},I},
	{"get_length",{P,S,I},D},
	{"get_media_type",{P},S},
	{"get_n_copies",{P},I},
	{"get_number_up",{P},I},
	{"get_orientation",{P},P},
	{"get_output_bin",{P},S},
	{"get",{P,S},S},
	{"get_page_ranges",{P,I},P},
	{"get_page_set",{P},I},
	{"get_paper_height",{P,I},D},
	{"get_paper_size",{P},P},
	{"get_paper_width",{P,I},D},
	{"get_printer",{P},S},
	{"get_print_pages",{P},I},
	{"get_quality",{P},I},
	{"get_resolution",{P},I},
	{"get_reverse",{P},B},
	{"get_scale",{P},D},
	{"get_use_color",{P},B},
	{"has_key",{P,S},B},
	{"new_from_file",{S,P},P},
	{"new_from_key_file",{P,S,P},P},
	{"new",{},P},
	{"set_bool",{P,S,B}},
	{"set_collate",{P,B}},
	{"set_default_source",{P,S}},
	{"set_dither",{P,S}},
	{"set_double",{P,S,D}},
	{"set_duplex",{P,I}},
	{"set_finishings",{P,S}},
	{"set_int",{P,S,I}},
	{"set_length",{P,S,D,I}},
	{"set_media_type",{P,S}},
	{"set_n_copies",{P,I}},
	{"set_number_up",{P,I}},
	{"set_number_up_layout",{P,I}},
	{"set_orientation",{P,I}},
	{"set_output_bin",{P,S}},
	{"set",{P,S,S}},
	{"set_page_ranges",{P,I},-routine_id("setPageRanges")},
	{"set_page_set",{P,I}},
	{"set_paper_height",{P,D,I}},
	{"set_paper_size",{P,P}},
	{"set_paper_width",{P,D,I}},
	{"set_printer",{P,S}},
	{"set_print_pages",{P,I}},
	{"set_quality",{P,I}},
	{"set_resolution",{P,I}},
	{"set_reverse",{P,B}},
	{"set_scale",{P,D}},
	{"set_use_color",{P,B}},
	{"to_file",{P,S,P},B},
	{"to_key_file",{P,P,S}},
	{"unset",{P,S}},
"GtkPrintSettings"}

------------------------------------------------------------------------
function setPageRanges(atom x, object r)
------------------------------------------------------------------------
integer n
atom z
if atom(r[1]) then 
	z = allocate(8) n = 1 poke4(z,r-1)
else n = length(r) 
	z = allocate(8*n)
	for i = 0 to n-1 do
		poke4(z+(i*8),r[i+1]-1)
	end for
end if
c_proc(define_c_proc(GTK,"gtk_print_settings_set_page_ranges",{P,P,I}),{x,z,n})
return 1
end function

control[GtkPrintUnixDialog] = {"gtk_print_unix_dialog",
	{GtkDialog},
	{"add_custom_tab",{P,P,P}},
	{"get_current_page",{P},I},
	{"get_embed_page_setup",{P},B},
	{"get_has_selection",{P},B},
	{"get_manual_capabilities",{P},I},
	{"get_page_setup",{P},P},
	{"get_page_setup_set",{P},B},
	{"get_selected_printer",{P},P},
	{"get_settings",{P},P},
	{"get_support_selection",{P},B},
	{"new",{S,P},P},
	{"set_current_page",{P,I}},
	{"set_embed_page_setup",{P,B}},
	{"set_has_selection",{P,B}},
	{"set_manual_capabilities",{P}},
	{"set_page_setup",{P,P}},
	{"set_settings",{P,P}},
	{"set_support_selection",{P,B}},
"GtkPrintUnixDialog"}

control[GtkProgressBar] = {"gtk_progress_bar",
{GtkOrientable,GtkWidget},
	{"get_ellipsize",{P},I},
	{"get_fraction",{P},D},
	{"get_inverted",{P},B},
	{"get_pulse_step",{P},D},
	{"get_show_text",{P},B},
	{"get_text",{P},S},
	{"new",{},P},
	{"pulse",{P}},
	{"set_ellipsize",{P,I}},
	{"set_fraction",{P,D}},
	{"set_inverted",{P,B}},
	{"set_pulse_step",{P,D}},
	{"set_show_text",{P,B}},
	{"set_text",{P,S}},
"GtkProgressBar"}

control[GtkRadioAction] = {"gtk_radio_action",
{GtkToggleAction},
	{"get_current_value",{P},I},
	{"get_group",{P},P},
	{"new",{S,S,S,S,I},P},
	{"set_current_value",{P,I}},
	{"set_group",{P,P}},
	{"join_group",{P,P}},
"GtkRadioAction"}

control[GtkRadioButton] = {"gtk_radio_button",
{GtkActivatable,GtkCheckButton},
	{"new_with_mnemonic_from_widget",{P,S},P},
	{"new_with_mnemonic",{P,S},P},
	{"new_with_label_from_widget",{P,S},P},
	{"new_with_label",{P,S},P},
	{"new_from_widget",{P},P},
	{"new",{P},P},
	{"get_group",{P},P},
	{"set_group",{P,P}},
	{"join_group",{P,P}},
"GtkRadioButton"}

control[GtkRadioMenuItem] = {"gtk_radio_menu_item",
{GtkActivatable,GtkCheckMenuItem},
	{"get_group",{P},A},
	{"new",{P},P},
	{"new_from_widget",{P},P},
	{"new_with_label",{P,S},P},
	{"new_with_mnemonic",{P,S},P},
	{"new_with_label_from_widget",{P,S},P},
	{"new_with_mnemonic_from_widget",{P,S},P},
	{"set_group",{P,P}},
"GtkRadioMenuItem"}

control[GtkRadioToolButton] = {"gtk_radio_tool_button",
{GtkActivatable,GtkToggleToolButton},
	{"get_group",{P},P},
	{"new_from_stock",{P,S},P},
	{"new_from_widget",{P},P},
	{"new",{P},P},
	{"new_with_stock_from_widget",{P,S},P},
	{"set_group",{P,P}},
"GtkRadioToolButton"}

control[GtkCellView] = {"gtk_cell_view",
{GtkCellLayout,GtkOrientable,GtkWidget},
	{"new",{},P},
	{"new_with_text",{S},P},
	{"new_with_markup",{S},P},
	{"new_with_pixbuf",{P},P},
	{"set_model",{P,P}},
	{"set_displayed_row",{P,P}},
	{"get_displayed_row",{P},P},
	{"set_background_color",{P,C}},
"GtkCellView"}

control[GtkAction] = {"gtk_action",
{GObject},
	{"new",{S,S,S,S},P},
	{"get_name",{P},S},
	{"is_sensitive",{P},B},
	{"get_sensitive",{P},B},
	{"set_sensitive",{P,B}},
	{"is_visible",{P},B},
	{"get_visible",{P},B},
	{"set_visible",{P,B}},
	{"activate",{P}},
	{"create_icon",{P,I},P},
	{"create_menu_item",{P},P},
	{"create_tool_item",{P},P},
	{"create_menu",{P},P},
	{"get_proxies",{P},A},
	{"connect_accelerator",{P}},
	{"disconnect_accelerator",{P}},
	{"block_activate",{P}},
	{"unblock_activate",{P}},
	{"get_accel_path",{P},S},
	{"set_accel_path",{P,S}},
	{"get_accel_closure",{P},P},
	{"set_accel_group",{P,P}},
	{"set_label",{P,S}},
	{"get_label",{P},S},
	{"set_short_label",{P,S}},
	{"get_short_label",{P},S},
	{"set_tooltip",{P,S}},
	{"get_tooltip",{P},S},
	{"set_stock_id",{P,S}},
	{"get_stock_id",{P},S},
	{"set_gicon",{P,P}},
	{"set_always_show_image",{P,B}},
	{"get_always_show_image",{P},B},
	{"get_gicon",{P},P},
	{"set_icon_name",{P,S}},
	{"get_icon_name",{P},S},
	{"set_visible_horizontal",{P,B}},
	{"get_visible_horizontal",{P},B},
	{"set_visible_vertical",{P,B}},
	{"get_visible_vertical",{P},B},
	{"set_is_important",{P,B}},
	{"get_is_important",{P},B},
	{"block_activate",{P}},
	{"unblock_activate",{P}},
	{"set_always_show_image",{P,B}},	
	{"get_always_show_image",{P},B},
"GtkAction"}

control[GtkActionGroup] = {"gtk_action_group",
{GObject},
"GtkActionGroup"}

control[GtkToggleAction] = {"gtk_toggle_action",
{GtkAction},
	{"get_active",{P},B},
	{"get_draw_as_radio",{P},B},
	{"new",{S,S,S,S},P},
	{"set_active",{P,B}},
	{"set_draw_as_radio",{P,B}},
	{"toggled",{P}},
"GtkToggleAction"}

control[GtkToggleButton] = {"gtk_toggle_button",
{GtkActivatable,GtkButton},
	{"new_with_mnemonic",{S},P},
	{"new_with_label",{S},P},
	{"new",{},P},
	{"get_active",{P},B},
	{"get_inconsistent",{P},B},
	{"get_mode",{P},B},
	{"set_active",{P,B}},
	{"set_inconsistent",{P,B}},
	{"set_mode",{P,B}},
	{"toggled",{P}},
"GtkToggleButton"}

control[GtkActionable] = {"gtk_actionable",
{GInterface},
	{"get_action_name",{P},S},
	{"set_action_name",{P,S}},
	{"get_action_target_value",{P},P},
	{"set_action_target_value",{P,P}},
	{"set_action_target",{P,S,P}},
	{"set_detailed_action_name",{P,S}},
"GtkActionable"}

control[GtkToggleToolButton] = {"gtk_toggle_tool_button",
{GtkActivatable,GtkToolButton},
	{"new_from_stock",{S},P},
	{"new",{},P},
	{"set_active",{P,B}},
	{"get_active",{P},B},
"GtkToggleToolButton"}

control[GtkToolBar] = {"gtk_toolbar",
{GtkToolShell,GtkOrientable,GtkContainer},
	{"get_drop_index",{P,I,I},I},
	{"get_icon_size",{P},I},
	{"get_item_index",{P,P},I},
	{"get_n_items",{P},I},
	{"get_nth_item",{P,I},P},
	{"get_orientation",{P},I},
	{"get_relief_style",{P},I},
	{"get_show_arrow",{P},B},
	{"get_style",{P},P},
	{"get_tooltips",{P},B},
	{"insert",{P,P,I}},
	{"new",{},P},
	{"set_drop_highlight_item",{P,P,I}},
	{"set_icon_size",{P,I}},
	{"set_show_arrow",{P,B}},
	{"set_style",{P,I}},
	{"set_tooltips",{P,B}},
	{"unset_icon_size",{P}},
	{"unset_style",{P}},
"GtkToolBar"}

control[GtkToolButton] = {"gtk_tool_button",
{GtkActivatable,GtkToolItem,GtkBin,GtkContainer,GtkWidget},
	{"get_icon_name",{P},S},
	{"get_icon_widget",{P},P},
	{"get_label",{P},S},
	{"get_label_widget",{P},P},
	{"get_stock_id",{P},S},
	{"get_type",{P},P},
	{"get_use_underline",{P},B},
	{"new_from_stock",{S},P},
	{"new",{P,S},P},
	{"set_icon_name",{P,S}},
	{"set_icon_widget",{P,P}},
	{"set_label",{P,S}},
	{"set_label_widget",{P,P}},
	{"set_stock_id",{P,S}},
	{"set_use_underline",{P,B}},
"GtkToolButton"}

control[GtkToolItemGroup] = {"gtk_tool_item_group",
{GtkToolShell,GtkContainer},
	{"get_collapsed",{P},B},
	{"get_drop_item",{P,I,I},P},
	{"get_ellipsize",{P},B},
	{"get_item_position",{P,P},I},
	{"get_n_items",{P},I},
	{"get_label",{P},S},
	{"get_label_widget",{P},P},
	{"get_nth_item",{P,I},P},
	{"get_header_relief",{P},I},
	{"insert",{P,P,I}},
	{"new",{S},P},
	{"set_collapsed",{P,B}},
	{"set_ellipsize",{P,I}},
	{"set_item_position",{P,P,I}},
	{"set_label",{P,S}},
	{"set_label_widget",{P,P}},
	{"set_header_relief",{P,I}},
"GtkToolItemGroup"}

control[GtkToolItem] = {"gtk_tool_item",
{GtkActivatable,GtkBin},
	{"get_expand",{P},B},
	{"get_ellipsize_mode",{P},I},
	{"get_homogeneous",{P},B},
	{"get_icon_size",{P},I},
	{"get_is_important",{P},B},
	{"get_orientation",{P},I},
	{"get_proxy_menu_item",{P,S},P},
	{"get_relief_style",{P},I},
	{"get_text_alignment",{P},F},
	{"get_text_orientation",{P},I},
	{"get_text_size_group",{P},P},
	{"get_toolbar_style",{P},I},
	{"get_use_drag_window",{P},B},
	{"get_visible_horizontal",{P},B},
	{"get_visible_vertical",{P},B},
	{"new",{},P},
	{"rebuild_menu",{P}},
	{"retrieve_proxy_menu_item",{P},P},
	{"set_expand",{P,B}},
	{"set_homogeneous",{P,B}},
	{"set_is_important",{P,B}},
	{"set_proxy_menu_item",{P,S,P}},
	{"set_tooltip_markup",{P,S}},
	{"set_tooltip_text",{P,S}},
	{"set_use_drag_window",{P,B}},
	{"set_visible_horizontal",{P,B}},
	{"set_visible_vertical",{P,B}},
"GtkToolItem"}

control[GtkGradient] = {"gtk_gradient",
{0},
	{"new_linear",{D,D,D,D},P},
	{"new_radial",{D,D,D,D,D,D},P},
	{"add_color_stop",{P,D,P}},
	{"resolve",{P,P,P},B},
	{"to_string",{P},S},
"GtkGradient"}

control[GtkToolTip] = {"gtk_tooltip",
{GObject},
	{"set_custom",{P,P}},
	{"set_icon_from_stock",{P,S,I}},
	{"set_icon",{P,P}},
	{"set_icon_from_gicon",{P,P,I}},
	{"set_icon_from_icon_name",{P,S,I}},
	{"set_markup",{P,S}},
	{"set_text",{P,S}},
	{"set_tip_area",{P,P}},
	{"trigger_tooltip_query",{P}},
"GtkToolTip"}

control[GtkOverlay] = {"gtk_overlay",
{GtkBin},
	{"new",{},P},
"GtkOverlay"}

control[GtkInfoBar] = {"gtk_info_bar",
{GtkOrientable,GtkBox},
	{"new",{},P},
	{"new_with_buttons",{P,X},P},
	{"add_action_widget",{P,P,I}},
	{"add_button",{P,S,I},P},
	{"add_buttons",{P,X}},
	{"get_action_area",{P},P},
	{"get_content_area",{P},P},
	{"get_message_type",{P},I},
	{"response",{P,I}},
	{"set_default_response",{P,I}},
	{"set_message_type",{P,I}},
	{"set_response_sensitive",{P,I,B}},
"GtkInfoBar"}

control[GtkTreeModel] = {"gtk_tree_model",
{GInterface},
	{"foreach",{P,P,P}},
	{"get",{P,P,I,P,I}},
	{"get_column_type",{P,I},I},
	{"get_flags",{P},P},
	{"get_iter",{P,P,P},B},
	{"get_iter_first",{P,P},B},
	{"get_iter_from_string",{P,P,S},B},
	{"get_n_columns",{P},P},
	{"get",{P,P,P,P,P}},
	{"get_path",{P},P},
	{"get_string_from_iter",{P,P},S},
	{"get_type",{P},P},
	{"get_valist",{P},P},
	{"get_value",{P,P,I,P}},
	{"iter_children",{P}},
	{"iter_has_child",{P},B},
	{"iter_n_children",{P,P},I},
	{"iter_next",{P,P},B},
	{"iter_nth_child",{P,P,P,I},B},
	{"iter_parent",{P}},
	{"ref_node",{P}},
	{"row_changed",{P}},
	{"row_deleted",{P}},
	{"row_has_child_toggled",{P}},
	{"row_inserted",{P}},
	{"rows_reordered",{P}},
	{"unref_node",{P}},
	{"destroy",{P}},
"GtkTreeModel"}

control[GtkTreeModelFilter] = {"gtk_tree_model_filter",
{GtkTreeModel,GtkTreeDragSource,GObject},
	{"clear_cache",{P}},
	{"convert_child_iter_to_iter",{P,P,P},B},
	{"convert_child_path_to_path",{P,P},P},
	{"convert_iter_to_child_iter",{P,P,P}},
	{"convert_path_to_child_path",{P,P},P},
	{"get_model",{P},P},
	{"new",{P,P},P},
	{"refilter",{P}},
	{"set_modify_func",{P,I,A,P,P,P}},
	{"set_visible_column",{P,I}},
	{"set_visible_func",{P,P,P}},
"GtkTreeModelFilter"}

control[GtkTreeModelSort] = {"gtk_tree_model_sort",
	{GtkTreeModel,GtkTreeSortable,GtkTreeDragSource,GObject},
	{"clear_cache",{P}},
	{"convert_child_iter_to_iter",{P,P,P}},
	{"convert_child_path_to_path",{P,P},P},
	{"convert_iter_to_child_iter",{P,P,P}},
	{"convert_path_to_child_path",{P,P},P},
	{"get_model",{P},P},
	{"iter_is_valid",{P,I},B},
	{"new_with_model",{P},P},
	{"reset_default_sort_func",{P}},
"GtkTreeModelSort"}

control[GtkTreePath] = {"gtk_tree_path",
{0},
	{"new",{},P},
	{"new_from_string",{S},P},
	{"new_from_indices",{X},P},
	{"to_string",{P},S},
	{"append_index",{P,I}},
	{"prepend_index",{P,I}},
	{"get_depth",{P},I},
	{"get_indices",{P},A},
	{"get_indices_with_depth",{P,I},A},
	{"free",{P}},
	{"copy",{P},P},
	{"compare",{P,P},I},
	{"next",{P}},
	{"prev",{P},B},
	{"up",{P},B},
	{"down",{P}},
	{"is_ancestor",{P,P},B},
	{"is_descendant",{P,P},B},
"GtkTreePath"}

control[GtkTreeRowReference] = {"gtk_tree_row_reference",
{GInterface},
	{"new",{P,P},P},
	{"get_model",{P},P},
	{"get_path",{P},P},
	{"valid",{P},B},
	{"free",{P}},
	{"copy",{P},P},
	{"inserted",{P,P}},
	{"deleted",{P,P}},
	{"reordered",{P,P,P,I}},
"GtkTreeRowReference"}

control[GtkWindowGroup] = {"gtk_window_group",
{GObject},
	{"add_window",{P,P}},
	{"list_windows",{P},A},
	{"new",{},P},
	{"remove_window",{P,P}},
	{"get_current_grab",{P},P},
	{"get_current_device_grab",{P,P},P},
"GtkWindowGroup"}

control[GtkViewport] = {"gtk_viewport",
{GtkScrollable,GtkBin},
	{"get_bin_window",{P},P},
	{"get_shadow_type",{P},I},
	{"get_view_window",{P},P}, 
	{"new",{P,P},P},
	{"set_shadow_type",{P,I}},
"GtkViewport"}

control[GtkFileFilter] = {"gtk_file_filter",
{GInitiallyUnowned},
	{"add_custom",{P,I,P,P,P}},
	{"add_mime_type",{P,S}},
	{"add_pattern",{P,S}},
	{"add_pixbuf_formats",{P}},
	{"filter",{P,P}},
	{"get_name",{P},S},
	{"get_needed",{P},I},
	{"new",{},P},
	{"set_name",{P,S}},
"GtkFileFilter"}

control[GtkTreeViewColumn] = {"gtk_tree_view_column",
{GtkCellLayout,GInitiallyUnowned},
	{"add_attribute",{P,P,S,I}},
	{"cell_get_position",{P,P,I,I},B},
	{"cell_get_size",{P,P,I,I,I,I}},
	{"cell_is_visible",{P},B},
	{"cell_set_cell_data",{P,P,P,B,B}},
	{"clear_attributes",{P,P}},
	{"clear",{P}},
	{"clicked",{P}},
	{"focus_cell",{P,P}},
	{"get_alignment",{P},F},
	{"get_button",{P},P},
	{"get_cell_renderers",{P},P},
	{"get_clickable",{P},B},
	{"get_expand",{P},B},
	{"get_fixed_width",{P},I},
	{"get_max_width",{P},I},
	{"get_min_width",{P},I},
	{"get_reorderable",{P},B},
	{"get_resizable",{P},B},
	{"get_sizing",{P},I},
	{"get_sort_column_id",{P},I},
	{"get_sort_indicator",{P},B},
	{"get_sort_order",{P},I},
	{"get_spacing",{P},I},
	{"get_title",{P},S},
	{"get_tree_view",{P},P},
	{"get_visible",{P},B},
	{"get_widget",{P},P},
	{"get_width",{P},I},
	{"new",{P,P},-routine_id("createTreeViewColumn")},
	{"new_with_attributes",{S,P,P,P,P,P,P},P},
	{"pack_end",{P,P,B}},
	{"pack_start",{P,P,B}},
	{"queue_resize",{P}},
	{"set_alignment",{P,F}},
	{"set_attributes",{P,P,S,I}},
	{"set_cell_data_func",{P,P,P,P,P}},
	{"set_clickable",{P,B}},
	{"set_expand",{P,B}},
	{"set_fixed_width",{P,I}},
	{"set_max_width",{P,I}},
	{"set_min_width",{P,I}},
	{"set_reorderable",{P,B}},
	{"set_resizable",{P,B}},
	{"set_sizing",{P,I}},
	{"set_sort_column_id",{P,I}},
	{"set_sort_indicator",{P,B}},
	{"set_sort_order",{P,I}},
	{"set_spacing",{P,I}},
	{"set_title",{P,S}},
	{"set_visible",{P,B}},
	{"set_widget",{P,P}},
"GtkTreeViewColumn"}

constant fn_tvc_new = define_c_func(GTK,"gtk_tree_view_column_new",{},P)
constant fn_tvc_pack_start = define_c_proc(GTK,"gtk_tree_view_column_pack_start",{P,P,I})
constant fn_tvc_add_attr = define_c_proc(GTK,"gtk_tree_view_column_add_attribute",{P,P,P,I})
constant fn_gobj_set = define_c_proc(GOBJECT,"g_object_set",{P,P,P,P})
constant fn_set_align = define_c_proc(GTK,"gtk_cell_renderer_set_alignment",{P,F,F})
constant fn_set_pad = define_c_proc(GTK,"gtk_cell_renderer_set_padding",{P,I,I})
constant fn_set_size = define_c_proc(GTK,"gtk_cell_renderer_set_fixed_size",{P,I})
constant fn_set_vis = define_c_proc(GTK,"gtk_cell_renderer_set_visible",{P,B})
constant fn_set_sen = define_c_proc(GTK,"gtk_cell_renderer_set_sensitive",{P,B})
constant fn_connect = define_c_func(GOBJECT,"g_signal_connect_data",{P,P,P,P,I,I},I)
-----------------------------------------------------------------------------
global function createTreeViewColumn(object attrs, object params={})
-----------------------------------------------------------------------------
atom renderer = -1
atom col = c_func(fn_tvc_new,{})
atom fn 
object test
object key, sig=0, funk=0, data=0
object funkname = "?"

attrs = keyvalues(text:format(attrs,params))

for i = 1 to length(attrs) do
	key = allocate_string(attrs[i][1],1)

	switch attrs[i][1] do
		case "p[1]","p[2]","p[3]","p[4]","p[5]","p[6]" then
			crash("Error: need a property=value pair here")

		case "type" then
			fn = define_c_func(GTK,sprintf("gtk_cell_renderer_%s_new",{attrs[i][2]}),{},P)
			if fn = -1 then
				crash("Error: no such thing as a '%s' renderer!",{attrs[i][2]})
			end if

			renderer = c_func(fn,{})
			c_proc(fn_tvc_pack_start,{col,renderer,TRUE})
			
			ifdef TVC then
				printf(1,"Created new cell renderer %s [%d]\n",{attrs[i][2],renderer})
			end ifdef
	
		case "alignment" then 
				c_proc(fn_set_align,{renderer,to_number(attrs[i][2][1]),to_number(attrs[i][2][2])})

		case "padding" then
				c_proc(fn_set_pad,{renderer,to_number(attrs[i][2][1]),to_number(attrs[i][2][2])})

		case "visible" then
				c_proc(fn_set_vis,{renderer,to_number(attrs[i][2])})

		case "sensitive" then
				c_proc(fn_set_sen,{renderer,to_number(attrs[i][2])})

		case "stock-size" then 
			if match("#",attrs[i][2]) then
				c_proc(fn_tvc_add_attr,
					{col,renderer,key,to_number(attrs[i][2])-1})
			else
				c_proc(fn_gobj_set,{renderer,key,to_number(attrs[i][2]),0})
			end if

		case "signal" then sig = attrs[i][2] 
		case "function" then funkname = attrs[i][2] 
		case "data" then data = attrs[i][2]

		case else
			if match("#",attrs[i][2]) then
				ifdef TVC then
					printf(1,"\tObtains value of %s from model col: #%d\n",{peek_string(key),to_number(attrs[i][2])})
				end ifdef
				c_proc(fn_tvc_add_attr,
					{col,renderer,key,to_number(attrs[i][2])-1})
			else 
				if equal("TRUE",attrs[i][2]) then attrs[i][2] = "1" end if
				if equal("FALSE",attrs[i][2]) then attrs[i][2] = "0" end if
				test = value(attrs[i][2])
				if test[1] = 0 then
					attrs[i][2] = test[2]
				else
					if match("$",attrs[i][2]) = 1 then
						attrs[i][2] = attrs[i][2][2..$]
					end if
					attrs[i][2] = allocate_string(attrs[i][2],1)
				end if
				ifdef TVC then
					printf(1,"\tSetting %s to %d\n",{peek_string(key),attrs[i][2]})
				end ifdef
				c_proc(fn_gobj_set,
					{renderer,key,attrs[i][2],0})
			end if
	end switch

end for

-- set up callback to Eu function;
if string(sig) then
	test = value(funkname) 
	if test[1] =  0 then
		funk = test[2]
	else
		funk = call_back(routine_id(funkname)) 
	end if
	
	if string(data) then
		test = value(data)
		if test[1] = 0 then
			data = test[2]
		else
			data = 0
		end if
	end if
	ifdef TVC then
		printf(1,"Connecting signal '%s' to func %s [%d] with data %d\n",{sig,funkname,funk,data})
	end ifdef
	c_func(fn_connect,{renderer,allocate_string(sig,1),funk,data,0,0})
	
end if

return col
end function

control[GtkTreeView] = {"gtk_tree_view",
{GtkScrollable,GtkContainer},
	{"append_column",{P,P},I},
	{"collapse_all",{P}},
	{"collapse_row",{P,P},B},
	{"columns_autosize",{P}},
	{"convert_bin_window_to_tree_coords",{P,I,I,I,I}},
	{"convert_bin_window_to_widget_coords",{P,I,I,I,I}},
	{"convert_tree_to_bin_window_coords",{P,I,I,I,I}},
	{"convert_tree_to_widget_coords",{P,I,I,I,I}},
	{"convert_widget_to_bin_window_coords",{P,I,I,I,I}},
	{"convert_widget_to_tree_coords",{P,I,I,I,I}},
	{"create_row_drag_icon",{P,P},P},
	{"enable_model_drag_dest",{P,P,I,P}},
	{"enable_model_drag_source",{P,P,P,I,P}},
	{"expand_all",{P}},
	{"expand_row",{P,P,B},B},
	{"expand_to_path",{P,P}},
	{"get_background_area",{P,P,P,P}},
	{"get_bin_window",{P},P},
	{"get_cell_area",{P,P,P,P}},
	{"get_column",{P,I},P},
	{"get_columns",{P},P},
	{"get_cursor",{P,P,P}},
	{"get_dest_row_at_pos",{P,I,I,P,I},B},
	{"get_drag_dest_row",{P,P,I}},
	{"get_enable_search",{P},B},
	{"get_enable_tree_lines",{P},B},
	{"get_expander_column",{P},P},
	{"get_fixed_height_mode",{P},B},
	{"get_grid_lines",{P},I},
	{"get_headers_clickable",{P},B},
	{"get_headers_visible",{P},B},
	{"get_hover_expand",{P},B},
	{"get_hover_selection",{P},B},
	{"get_level_indentation",{P},I},
	{"get_n_columns",{P},I},
	{"get_model",{P},P},
	{"get_path_at_pos",{P,I,I,P,P,I,I},B},
	{"get_reorderable",{P},B},
	{"get_row_separator_func",{P},P},
	{"get_rubber_banding",{P},B},
	{"get_rules_hint",{P},B},
	{"get_search_column",{P},I},
	{"get_search_entry",{P},P},
	{"get_search_equal_func",{P},P},
	{"get_search_position_func",{P},P},
	{"get_selection",{P},P},
	{"get_show_expanders",{P},B},
	{"get_tooltip_column",{P},I},
	{"get_tooltip_context",{P,I,I,B,P,P,P},B},
	{"get_visible_range",{P,P,P},B},
	{"get_visible_rect",{P,P}},
	{"insert_column",{P,P,I},I},
	{"insert_columns",{P,P},-routine_id("setTreeViewInsertColumns")},
	{"insert_column_with_attributes",{P,I,S,P,S,I,S,I,S,I,S,I},I},
	{"insert_column_with_data_func",{P,I,S,P,P,P,P},I},
	{"is_rubber_banding_active",{P},B},
	{"map_expanded_rows",{P,P,P}},
	{"move_column_after",{P,P,P}},
	{"new",{},P},
	{"new_with_model",{P},P},
	{"remove_column",{P,P},I},
	{"row_activated",{P,P,P}},
	{"row_expanded",{P,P},B},
	{"scroll_to_cell",{P,P,P,B,F,F}},
	{"scroll_to_point",{P,I,I}},
	{"set_column_drag_function",{P,P,P,P}},
	{"set_cursor_on_cell",{P,P,P,P,B}},
	{"set_cursor",{P,P,P,B}},
	{"set_destroy_count_func",{P,P,P,P}},
	{"set_drag_dest_row",{P,P,I}},
	{"set_enable_search",{P,B}},
	{"set_enable_tree_lines",{P,B}},
	{"set_expander_column",{P,P}},
	{"set_fixed_height_mode",{P,B}},
	{"set_grid_lines",{P,I}},
	{"set_headers_clickable",{P,B}},
	{"set_headers_visible",{P,B}},
	{"set_hover_expand",{P,B}},
	{"set_hover_selection",{P,B}},
	{"set_level_indentation",{P,I}},
	{"set_model",{P,P}},
	{"set_reorderable",{P,B}},
	{"set_row_separator_func",{P,P,P,P}},
	{"set_rubber_banding",{P,B}},
	{"set_rules_hint",{P,B}},
	{"set_search_column",{P,I}},
	{"set_search_entry",{P,P}},
	{"set_search_equal_func",{P,P,P,P}},
	{"set_search_position_func",{P,P,P,P}},
	{"set_show_expanders",{P,B}},
	{"set_tooltip_cell",{P,P,P,P,P}},
	{"set_tooltip_column",{P,I}},
	{"set_tooltip_row",{P,P,P}},
	{"unset_rows_drag_dest",{P}},
	{"unset_rows_drag_source",{P}},
"GtkTreeView"}

constant fn_tv_insert_col = define_c_func(GTK,"gtk_tree_view_insert_column",{P,P,I},I)
-----------------------------------------------------------------------------------------
global function setTreeViewInsertColumns(atom tv, object cols)
-----------------------------------------------------------------------------------------
for i = 1 to length(cols) do
	 c_func(fn_tv_insert_col,{tv,cols[i],i-1})
end for
return 1
end function

control[GtkTreeStore] = {"gtk_tree_store",
	{GtkTreeSortable,GtkTreeModel,GObject},
	{"new",{X},-routine_id("createTreeStore")},
	{"append",{P,P,P}},
	{"clear",{P}},
	{"insert_after",{P,P,P,P}},
	{"insert_before",{P,P,P,P}},
	{"insert",{P,P,P,I}},
	{"insert_with_values",{P,P,P,I,I,P,I,P,I,P,I,P}},
	{"insert_with_valuesv",{P,P,P,I,I,A,I}},
	{"is_ancestor",{P,P,P},B},
	{"iter_depth",{P,P},I},
	{"iter_is_valid",{P,P},B},
	{"move_after",{P,P,P}},
	{"move_before",{P,P,P}},
	{"newv",{I,A},P},
	{"prepend",{P,P,P}},
	{"remove",{P,P},B},
	{"reorder",{P,P,I}},
	{"set_column_types",{P,I,A}},
	{"set",{P,P,I,S,I}},
	{"set_valist",{P,P,P}},
	{"set_value",{P,P,I,P}},
	{"set_valuesv",{P,P,I,A,I}},
	{"set_col_data",{P,P,I,P},-routine_id("setTreeStoreColumnData")},
	{"set_row_data",{P,P,P},-routine_id("setTreeStoreRowData")},
	{"swap",{P,P,P}},
"GtkTreeStore"}

------------------------------------------------------------------------
global function createTreeStore(object params)
------------------------------------------------------------------------
integer fail = FALSE

if atom(params) then 
	crash("Error: parameters must be a {sequence}")
end if

for i = 1 to length(params) do
	switch params[i] do
		case gVOID,gINTERFACE,gENUM,gBOXED,gPARAM,gVARIANT then
		fail = TRUE params = params[i]
	end switch
end for

if fail then 
	crash("Error: cannot create a list store with column type %d \n",params)
end if

object proto = I & repeat(P,length(params))
params = length(params) & params
atom x = define_c_func(GTK,"gtk_tree_store_new",proto,P)
return c_func(x,params)

end function

--------------------------------------------------------------------------
global function setTreeStoreRowData(atom self, atom iter, object data)
--------------------------------------------------------------------------
integer max_col =  c_func(fn_n_cols,{self})

if max_col > length(data) then max_col = length(data) end if

for i = 1 to max_col do
	iter = setTreeStoreColumnData(self,iter,i,data[i])
end for

return iter
end function

--------------------------------------------------------------------------------------
global function setTreeStoreColumnData(atom self, atom iter, integer col, object data)
--------------------------------------------------------------------------------------
object proto = {P,P,I,P,I}
object params
atom fn

integer max_col = c_func(fn_n_cols,{self})
if col > max_col or col < 1 then 
	crash("Set column data: column %d does not exist! Max = %d",{col,max_col})
end if

integer ct = c_func(fn_col_type,{self,col-1})

switch ct do
	case gFLOAT then proto = {P,P,I,D,I} 
	case gSTR then proto = {P,P,I,S,I}
end switch

if string(data) then data = allocate_string(data,1) end if

fn = define_c_proc(GTK,"gtk_tree_store_set",proto)
params = {self,iter,col-1,data,-1}
c_proc(fn,params)

return iter
end function

control[GtkCellArea] = {"gtk_cell_area",
{GtkCellLayout,GInitiallyUnowned},
	{"add",{P,P}},
	{"remove",{P,P}},
	{"has_renderer",{P,P},B},
	{"foreach",{P,P,P}},
	{"foreach_alloc",{P,P,P,P,P,P,P}},
	{"event",{P,P,P,P,P,I},I},
	{"render",{P,P,P,P,P,P,P,B}},
	{"get_cell_allocation",{P,P,P,P,P,P}},
	{"get_cell_at_position",{P,P,P,P,I,I,P},P},
	{"create_context",{P},P},
	{"copy_context",{P,P},P},
	{"get_request_mode",{P},P},
	{"get_preferred_width",{P,P,P,I,I}},
	{"get_preferred_height",{P,P,P,I,I}},
	{"get_preferred_width_for_height",{P,P,P,I,I,I}},
	{"get_preferred_height_for_width",{P,P,P,I,I,I}},
	{"get_current_path_string",{P},S},
	{"apply_attributes",{P,P,P,B,B}},
	{"attribute_connect",{P,P,S,I}},
	{"attribute_disconnect",{P,P,S}},
	{"install_cell_property",{P,I,P}},
	{"find_cell_property",{P,S},P},
	{"list_cell_properties",{P,I},P},
	{"add_with_properties",{P,P,S,I,I}},
	{"cell_set",{P,P,S,I,I}},
	{"cell_get",{P,P,S,I,I}},
	{"cell_set_property",{P,P,S,P}},
	{"cell_get_property",{P,P,S,P}},
	{"is_activatable",{P},B},
	{"activate",{P,P,P,P,I,B},B},
	{"focus",{P,I},B},
	{"set_focus_cell",{P,P}},
	{"get_focus_cell",{P},P},
	{"add_focus_sibling",{P,P,P}},
	{"is_focus_sibling",{P,P,P},B},
	{"get_focus_siblings",{P,P},A},
	{"get_focus_from_sibling",{P,P},P},
	{"get_edited_cell",{P},P},
	{"get_edit_widget",{P},P},
	{"activate_cell",{P,P,P,P,P,I},B},
	{"stop_editing",{P,B}},
	{"inner_cell_area",{P,P,P,P}},
	{"request_renderer",{P,P,I,P,I,I,I}},
"GtkCellArea"}

control[GtkCellAreaBox] = {"gtk_cell_area_box",
{GtkCellArea,GtkCellLayout,GtkOrientable,GObject},
	{"new",{},P},
	{"pack_start",{P,P,B,B,B}},
	{"pack_end",{P,P,B,B,B}},
	{"get_spacing",{P},I},
	{"set_spacing",{P,I}},
"GtkCellAreaBox"}

control[GtkCellAreaContext] = {"gtk_cell_area_context",
{GObject},
	{"get_area",{P},P},
	{"allocate",{P,I,I}},
	{"reset",{P}},
	{"get_preferred_width",{P,I,I}},
	{"get_preferred_height",{P,I,I}},
	{"get_preferred_height_for_width",{P,I,I,I}},
	{"get_preferred_width_for_height",{P,I,I,I}},
	{"get_allocation",{P,I,I}},
	{"push_preferred_width",{P,I,I}},
	{"push_preferred_height",{P,I,I}},
"GtkCellAreaContext"}

control[GtkTreeSortable] = {"gtk_tree_sortable",
{GInterface},
	{"get_sort_column_id",{P,I,I},B},
	{"get_type",{P},P},
	{"has_default_sort_func",{P},B},
	{"set_default_sort_func",{P,P,P,P}},
	{"set_sort_column_id",{P,I,I}},
	{"set_sort_func",{P,I,P,P,P}},
	{"sort_column_changed",{P}},
"GtkTreeSortable"}

control[GtkTreeSelection] = {"gtk_tree_selection",
{GObject},
	{"count_selected_rows",{P},I},
	{"get_mode",{P},I},
	{"get_selected",{P,P,P},B},
	{"get_selected_rows",{P,P},P},
	{"get_tree_view",{P},P},
	{"get_user_data",{P},P},
	{"iter_is_selected",{P}},
	{"path_is_selected",{P}},
	{"get",{P,P,I},S},
	{"select_all",{P}},
	{"selected_foreach",{P,P,P}},
	{"select_iter",{P}},
	{"select_path",{P}},
	{"select_range",{P}},
	{"set_mode",{P,I}},
	{"set_select_function",{P}},
	{"unselect_all",{P}},
	{"unselect_iter",{P}},
	{"unselect_path",{P}},
	{"unselect_range",{P}},
	{"get_data_from_col",{P,I},-routine_id("getSelectedItem")},
	{"get_data_from_column",{P,I},-routine_id("getSelectedItem")},
	{"get_data_from_row",{P},-routine_id("getSelectedRowData")},
	{"get_iter",{P},-routine_id("getIterFromSelection")},
	{"get_iter_from_selection",{P},-routine_id("getIterFromSelection")},
"GtkTreeSelection"}

constant fn_ts_get_sel = define_c_func(GTK,"gtk_tree_selection_get_selected",{P,P,P},B)
constant fn_tm_get_col_type = define_c_func(GTK,"gtk_tree_model_get_column_type",{P,I},I)
constant fn_tm_get_val = define_c_proc(GTK,"gtk_tree_model_get_value",{P,P,I,P})
------------------------------------------------------------------------
global function getSelectedItem(atom selection, integer col)
------------------------------------------------------------------------
atom mdl = allocate(16) poke4(mdl,{0,0,0,0})
atom iter = allocate(16) poke4(iter,{0,0,0,0})
object val = allocate(16) poke4(val,{0,0,0,0})
object result=-1
integer ct
col-=1
if c_func(fn_ts_get_sel,{selection,mdl,iter}) then
	ct = c_func(fn_tm_get_col_type,{peek4u(mdl),col})
	c_proc(fn_tm_get_val,{peek4u(mdl),iter,col,val})
	
ifdef WINDOWS then
		val = peek4u({val,4})
elsedef
		val = peek4u({val,2})
end ifdef
	
ifdef WINDOWS then	
	if val[1] = gSTR then 
		if val[3] != 0 then
			result = peek_string(val[3])
		end if
		else result = val[3]
	end if

elsedef

	if val[1] = gSTR then 
		if val[2] != 0 then
			result = peek_string(val[2])
		end if
		else result = val[2]
	end if
end ifdef
	

end if
return result
end function

------------------------------------------------------------------------
global function getSelectedRowData(atom selection)
------------------------------------------------------------------------
atom mdl = allocate(16)
atom iter = allocate(16)
integer n
object result = 0
if c_func(fn_ts_get_sel,{selection,mdl,iter}) then
	n = c_func(fn_n_cols,{peek4u(mdl)})
	result = repeat(0,n)
	for i = 1 to n do
		result[i] = getSelectedItem(selection,i)
	end for
end if
return result
end function

------------------------------------------------------------------------
global function getIterFromSelection(atom selection)
------------------------------------------------------------------------
atom mdl = allocate(16)
atom iter = allocate(16)
if c_func(fn_ts_get_sel,{selection,mdl,iter}) then
	return iter
end if
return -1
end function

control[GtkEntry] = {"gtk_entry",
{GtkEditable,GtkCellEditable,GtkWidget},
	{"new",{},P},
	{"get_activates_default",{P},B},
	{"get_alignment",{P},F},	
	{"get_buffer",{P},P},
	{"get_completion",{P},P},
	{"get_cursor_hadjustment",{P},P},
	{"get_has_frame",{P},B},
	{"get_current_icon_drag_source",{P},I},
	{"get_icon_area",{P,I,P}},
	{"get_icon_activatable",{P,I},B},
	{"get_icon_at_pos",{P,I,I},I},
	{"get_icon_gicon",{P,I},P},
	{"get_icon_name",{P,I},S},
	{"get_icon_pixbuf",{P,I},P},
	{"get_icon_sensitive",{P,I},B},
	{"get_icon_stock",{P,I},S},
	{"get_icon_storage_type",{P,I},I},
	{"get_icon_tooltip_markup",{P,I},S},
	{"get_icon_tooltip_text",{P,I},S},
	{"get_inner_border",{P},P},
	{"get_invisible_char",{P},I},
	{"get_layout_offsets",{P,I,I}},
	{"get_layout",{P},P},
	{"get_max_length",{P},I},
	{"get_overwrite_mode",{P},B},
	{"get_progress_fraction",{P},D},
	{"get_progress_pulse_step",{P},D},
	{"get_placeholder_text",{P},S},
	{"get_text",{P},S},
	{"get_text_area",{P,P}}, 
	{"get_text_length",{P},I},
	{"get_type",{},I},
	{"get_visibility",{P},B},
	{"get_width_chars",{P},I},
	{"im_context_filter_keypress",{P,P},B},
	{"layout_index_to_text_index",{P,I},I},
	{"new_with_buffer",{P},P},
	{"progress_pulse",{P}},
	{"reset_im_context",{P}},
	{"set_activates_default",{P,B}},
	{"set_alignment",{P,F}},
	{"set_buffer",{P,P}},
	{"set_completion",{P,P}},
	{"set_cursor_hadjustment",{P,P}},
	{"set_has_frame",{P,B}},
	{"set_icon_activatable",{P,I,B}},
	{"set_icon_drag_source",{P,I,P,I}},
	{"set_icon_from_pixbuf",{P,I,P}},
	{"set_icon_from_stock",{P,I,S}},
	{"set_icon_from_icon_name",{P,I,S}},
	{"set_icon_from_gicon",{P,I,P}},
	{"set_icon_sensitive",{P,I,B}},
	{"set_icon_tooltip_markup",{P,I,S}},
	{"set_icon_tooltip_text",{P,I,S}},
	{"set_inner_border",{P,P}},
	{"set_invisible_char",{P,I}},
	{"set_max_length",{P,I}},
	{"set_overwrite_mode",{P,B}},
	{"set_placeholder_text",{P,S}},
	{"set_progress_fraction",{P,D}},
	{"set_progress_pulse_step",{P,D}},
	{"set_text",{P,S}},
	{"set_visibility",{P,B}},
	{"set_width_chars",{P,I}},
	{"text_index_to_layout_index",{P,I},I},
	{"unset_invisible_char",{P}},
"GtkEntry"}

control[GtkEntryBuffer] = {"gtk_entry_buffer",
{GObject},
	{"new",{S,I},P},
	{"get_text",{P},S},
	{"set_text",{P,S,I}},
	{"get_bytes",{P},I},
	{"get_length",{P},I},
	{"get_max_length",{P},I},
	{"set_max_length",{P,I}},
	{"insert_text",{P,I,S,I},I},
	{"delete_text",{P,I,I},I},
	{"emit_deleted_text",{P,I,I}},
	{"emit_inserted_text",{P,I,S,I}},
"GtkEntryBuffer"}

control[GtkEntryCompletion] = {"gtk_entry_completion",
{GtkCellLayout,GObject},
	{"new",{},P},
	{"new_with_area",{P},P},
	{"complete",{P}},
	{"delete_action",{P,I}},
	{"get_completion_prefix",{P},S},
	{"get_entry",{P},P},
	{"get_inline_completion",{P},B},
	{"get_inline_selection",{P},B},
	{"get_minimum_key_length",{P},I},
	{"get_model",{P},P},
	{"get_popup_completion",{P},B},
	{"get_popup_set_width",{P},B},
	{"get_popup_single_match",{P},B},
	{"get_text_column",{P},I},
	{"insert_action_markup",{P,I,S}},
	{"insert_action_text",{P,I,S}},
	{"insert_prefix",{P}},
	{"set_inline_completion",{P,B}},
	{"set_inline_selection",{P,B}},
	{"set_match_func",{P,P,P,P}},
	{"set_minimum_key_length",{P,I}},
	{"set_model",{P,P}},
	{"set_popup_completion",{P,B}},
	{"set_popup_set_width",{P,B}},
	{"set_popup_single_match",{P,B}},
	{"set_text_column",{P,I}},
	{"compute_prefix",{P,S},S},
"GtkEntryCompletion"}

control[GtkEventBox] = {"gtk_event_box",
{GtkBin},
	{"get_above_child",{P},B},
	{"get_visible_window",{P},B},
	{"new",{},P},
	{"set_above_child",{P,B}},
	{"set_visible_window",{P,B}},
"GtkEventBox"}

control[GtkExpander] = {"gtk_expander",
{GtkBin},
	{"get_expanded",{P},B},
	{"get_label",{P},S},
	{"get_label_fill",{P},B},
	{"get_label_widget",{P},P},
	{"get_resize_toplevel",{P},B},
	{"get_spacing",{P},I},
	{"get_use_markup",{P},B},
	{"get_use_underline",{P},B},
	{"new",{S},P},
	{"new_with_mnemonic",{S},P},
	{"set_expanded",{P,B}},
	{"set_label",{P,S}},
	{"set_label_fill",{P,B}},
	{"set_label_widget",{P,P}},
	{"set_resize_toplevel",{P,B}},
	{"set_spacing",{P,I}},
	{"set_use_markup",{P,B}},
	{"set_use_underline",{P,B}},
"GtkExpander"}

control[GtkStock] = {"gtk_stock",
{0},
	{"add",{P,I}},
	{"add_static",{P,I}},
	{"item_copy",{P},P},
	{"item_free",{P}},
	{"list_ids",{},A},
	{"lookup",{S,P},B},
	{"set_translate_func",{S,P,P,P}},
"GtkStock"}

control[GtkSpinButton] = {"gtk_spin_button",
{GtkEditable,GtkCellEditable,GtkEntry},
	{"new",{P,D,I},P},
	{"new_with_range",{D,D,D},P},
	{"configure",{P,P,D,I}},
	{"get_adjustment",{P},P},
	{"get_digits",{P},I},
	{"get_increments",{P,D,D}},
	{"get_numeric",{P},B},
	{"get_range",{P,D,D}},
	{"get_snap_to_ticks",{P},B},
	{"get_update_policy",{P},I},
	{"get_value",{P},D},
	{"get_value_as_int",{P},I},
	{"get_wrap",{P},B},
	{"set_adjustment",{P,P}},
	{"set_digits",{P,I}},
	{"set_increments",{P,D,D}},
	{"set_numeric",{P,B}},
	{"set_range",{P,D,D}},
	{"set_snap_to_ticks",{P,B}},
	{"set_update_policy",{P,I}},
	{"set_value",{P,D}},
	{"set_wrap",{P,B}},
	{"spin",{P,I,D}},
	{"update",{P}},
"GtkSpinButton"}

control[GtkSpinner] = {"gtk_spinner",
{GtkWidget},
	{"new",{},P},
	{"start",{P}},
	{"stop",{P}},
"GtkSpinner"}

control[GtkAdjustment] = {"gtk_adjustment",
{GInitiallyUnowned},
	{"changed",{P}},
	{"clamp_page",{P,D,D}},
	{"configure",{P,D,D,D,D,D,D}},
	{"get_lower",{P},D},
	{"set_lower",{P,D}},
	{"get_upper",{P},D},
	{"set_upper",{P,D}},
	{"get_page_increment",{P},D},
	{"set_page_increment",{P,D}},
	{"get_page_size",{P},D},
	{"set_page_size",{P,D}},
	{"get_step_increment",{P},D},
	{"set_step_increment",{P,D}},
	{"get_value",{P},D},
	{"new",{D,D,D,D,D,D},P},
	{"set_value",{P,D}},
	{"value_changed",{P}},
	{"get_minimum_increment",{P},D},
"GtkAdjustment"}

control[GtkComboBoxText] = {"gtk_combo_box_text",
	{GtkCellLayout,GtkCellEditable,GtkComboBox},
	{"new",{},P},
	{"new_with_entry",{},P},
	{"append",{P,S,S}},
	{"prepend",{P,S,S}},
	{"insert",{P,I,S,S}},
	{"insert_text",{P,I,S}},
	{"prepend_text",{P,S}},
	{"append_text",{P,S}},
	{"remove",{P,I}},
	{"remove_all",{P}},
	{"get_active_text",{P},S},
"GtkComboBoxText"}

control[GtkComboBox] = {"gtk_combo_box",
	{GtkCellLayout,GtkCellEditable,GtkBin},
	{"new",{},P},
	{"new_with_entry",{},P},
	{"new_with_model",{P},P},
	{"new_with_model_and_entry",{P},P},
	{"new_with_area",{P},P},
	{"new_with_area_and_entry",{P},P},
	{"get_active_iter",{P,P},B},
	{"get_active",{P},I},
	{"get_active_id",{P},S},
	{"get_add_tearoffs",{P},B},
	{"get_column_span_column",{P},I},
	{"get_entry_text_column",{P},I},
	{"get_has_entry",{P},B},
	{"get_id_column",{P},I},
	{"get_focus_on_click",{P},B},
	{"get_model",{P},P},
	{"get_popup_accessible",{P},P},
	{"get_popup_fixed_width",{P},B},
	{"get_row_separator_func",{P},P},
	{"get_row_span_column",{P},I},
	{"get_title",{P},S},
	{"get_wrap_width",{P},I},
	{"new_with_model",{P},P},
	{"new",{},P},
	{"popdown",{P}},
	{"popup",{P}},
	{"popup_for_device",{P,P}},
	{"set_active_iter",{P,P}},
	{"set_active",{P,I}},
	{"set_active_id",{P,S},B},
	{"set_add_tearoffs",{P,B}},
	{"set_column_span_column",{P,I}},
	{"set_entry_text_column",{P,I}},
	{"set_focus_on_click",{P,B}},
	{"set_model",{P,P}},
	{"set_id_column",{P,I}},
	{"set_popup_fixed_width",{P,B}},
	{"set_row_separator_func",{P,P,P,P}},
	{"set_row_span_column",{P,I}},
	{"set_title",{P,S}},
	{"set_wrap_width",{P,I}},
	{"get_button_sensitivity",{P},I},
	{"set_button_sensitivity",{P,I}},
	{"text_new",{},P},
"GtkComboBox"}

control[GtkColorButton] = {"gtk_color_button",
{GtkColorChooser,GtkActionable,GtkActivatable,GtkBin,GtkContainer,GtkWidget},
	{"get_title",{P},S},
	{"get_rgba",{P,P}},
	{"new",{},P},
	{"new_with_rgba",{P},P},
	{"set_title",{P,S}},
"GtkColorButton"}

control[GtkCheckButton] = {"gtk_check_button",
{GtkActivatable,GtkToggleButton},
	{"new_with_mnemonic",{S},P},
	{"new_with_label",{S},P},
	{"new",{},P},
"GtkCheckButton"}

control[GtkSeparatorToolItem] = {"gtk_separator_tool_item",
{GtkToolItem},
	{"new",{},P},
	{"set_draw",{P,B}},
	{"get_draw",{P},B},
"GtkSeparatorToolItem"}

control[GtkSeparatorMenuItem] = {"gtk_separator_menu_item",
{GtkActivatable,GtkMenuItem},
	{"new",{},P},
"GtkSeparatorMenuItem"}

control[GtkToolShell] = {"gtk_tool_shell",
{GInterface,GtkWidget},
	{"get_ellipsize_mode",{P},I},
	{"get_icon_size",{P},I},
	{"get_orientation",{P},I},
	{"get_relief_style",{P},I},
	{"get_style",{P},I},
	{"get_text_alignment",{P},F},
	{"get_text_orientation",{P},I},
	{"rebuild_menu",{P}},
	{"get_text_size_group",{P},P},
"GtkToolShell"}

control[GtkCalendar] = {"gtk_calendar",
{GtkWidget},
	{"clear_marks",{P}},
	{"get_date",{P,P},-routine_id("getCalendarDate")},
	{"get_day_is_marked",{P,I},B},
	{"get_display_options",{P},I},
	{"mark_day",{P,I},B},
	{"new",{},P},
	{"select_day",{P,I}},
	{"select_month",{P,I,I},B},
	{"set_display_options",{P,I}},
	{"unmark_day",{P,I},B},
	{"set_detail_func",{P,P,P,P}},
	{"set_detail_width_chars",{P,I}},
	{"get_detail_width_chars",{P},I},
	{"get_detail_height_rows",{P},I},
	{"set_detail_height_rows",{P,I}},
	{"set_date",{P,P},-routine_id("setCalendarDate")},
"GtkCalendar"}

------------------------------------------------------------------------
-- (3) Calendar convenience functions
------------------------------------------------------------------------
constant set_month = define_c_proc(GTK,"gtk_calendar_select_month",{P,I,I})
constant set_day = define_c_proc(GTK,"gtk_calendar_select_day",{P,I})
constant get_date = define_c_proc(GTK,"gtk_calendar_get_date",{P,I,I,I})

------------------------------------------------------------------------
global function setCalendarDate(atom handle, object date)
------------------------------------------------------------------------
integer yr = date[1], mo = date[2], da = date[3]
	c_proc(set_month,{handle,mo-1,yr})
	c_proc(set_day,{handle,da})
return 1
end function

------------------------------------------------------------------------
global function getCalendarDate(atom handle, object fmt)
------------------------------------------------------------------------
	atom y = allocate(8)
	atom m = allocate(8)
	atom d = allocate(8)
	if atom(fmt) then fmt = "%A, %b %d, %Y" end if 
	object result
	c_proc(get_date,{handle,y,m,d})
	result = datetime:new(peek4u(y),peek4u(m)+1,peek4u(d))
	result = locale:datetime(fmt,result)
	free(y) free(m) free(d)
	return result
end function

control[GtkAlignment] = {"gtk_alignment",
{GtkBin},
	{"get_padding",{P,I,I,I,I}},
	{"new",{F,F,F,F},P},
	{"set",{P,F,F,F,F}},
	{"set_padding",{P,I,I,I,I}},
"GtkAlignment"}

control[GtkScaleButton] = {"button",
{GtkActivatable,GtkOrientable,GtkButton},
	{"get_adjustment",{P},P},
	{"get_value",{P},D},
	{"new",{I,D,D,D,P},P},
	{"set_adjustment",{P,P}},
	{"set_icons",{P,S}},
	{"set_value_pos",{P,I}},
	{"set_value",{P,D}},
"GtkScaleButton"}

control[GtkVolumeButton] = {"gtk_volume_button",
	{GtkActivatable,GtkOrientable,GtkScaleButton},
	{"new",{},P},
"GtkVolumeButton"}

control[GtkBuilder] = {"gtk_builder",
{GObject},
	{"new",{},P},
	{"add_from_file",{P,S,P},I},
	{"add_from_string",{P,S,I,P},I},
	{"add_objects_from_file",{P,S,P,P},I},
	{"add_objects_from_string",{P,S,I,P,P},I},
	{"get_object",{P,S},P},
	{"get_objects",{P},P},
	{"connect_signals",{P,P}},
	{"connect_signals_full",{P,P,P}},
	{"set_translation_domain",{P,S}},
	{"get_translation_domain",{P},S},
	{"get_type_from_name",{P,S},P},
	{"value_from_string",{P,P,S,P,P},B},
	{"value_from_string_type",{P,P,S,P,P},B},
"GtkBuilder"}

control[GtkDrawingArea] = {"gtk_drawing_area",
{GtkWidget},
	{"new",{},P},
"GtkDrawingArea"}

control[GtkCheckMenuItem] = {"gtk_check_menu_item",
{GtkActivatable,GtkMenuItem},
	{"new_with_mnemonic",{S},P}, 
	{"new_with_label",{S},P}, 
	{"new",{},P}, 
	{"get_active",{P},B}, 
	{"set_active",{P,B}}, 
	{"toggled",{P}}, 
	{"get_inconsistent",{P},B}, 
	{"set_inconsistent",{P,B}}, 
	{"set_draw_as_radio",{P,B}}, 
	{"get_draw_as_radio",{P},B}, 
"GtkCheckMenuItem"} 

control[GtkRecentFilter] = {"gtk_recent_filter",
	{GInitiallyUnowned},
	{"new",{},P},
	{"get_name",{P},S},
	{"set_name",{P,S}},
	{"add_mime_type",{P,S}},
	{"add_pattern",{P,S}},
	{"add_pixbuf_formats",{P}},
	{"add_application",{P,S}},
	{"add_group",{P,S}},
	{"add_age",{P,I}},
	{"add_custom",{P,I,P,P,P}},
	{"get_needed",{P},I},
	{"filter",{P,P},B},
"GtkRecentFilter"}

control[GtkRecentInfo] = {"gtk_recent_info",
{GObject},
	{"get_uri",{P},S},
	{"get_display_name",{P},S},
	{"get_description",{P},S},
	{"get_mime_type",{P},S},
	{"get_added",{P},I},
	{"get_modified",{P},I},
	{"get_visited",{P},I},
	{"get_private_hint",{P},B},
	{"get_application_info",{P,S,S,I,I},B},
	{"get_applications",{P,I},A},
	{"last_application",{P},S},
	{"has_application",{P,S},B},
	{"create_app_info",{P,S,P},P},
	{"get_groups",{P,I},A},
	{"has_group",{P,S},B},
	{"get_icon",{P,I},P},
	{"get_gicon",{P},P},
	{"get_short_name",{P},S},
	{"get_url_display",{P},S},
	{"get_age",{P},I},
	{"is_local",{P},B},
	{"exists",{P},B},
	{"match",{P,P},B},
"GtkRecentInfo"}
	
control[GtkRecentManager] = {"gtk_recent_manager",
{GObject},
	{"new",{},P},
	{"get_default",{},P},
	{"add_item",{P,S},B},
	{"add_full",{P,S,P},B},
	{"remove_item",{P,S,P},B},
	{"lookup_item",{P,S,P},P},
	{"has_item",{P,S},B},
	{"move_item",{P,S,S,P},B},
	{"get_limit",{P},I},
	{"set_limit",{P,I}},
	{"get_items",{P},A},
	{"purge_items",{P,P},I},
"GtkRecentManager"}

control[GtkRecentChooser] = {"gtk_recent_chooser",
{GInterface,GObject},
	{"set_show_private",{P,B}},
	{"get_show_private",{P},B},
	{"set_show_not_found",{P,B}},
	{"get_show_not_found",{P},B},
	{"set_show_icons",{P,B}},
	{"get_show_icons",{P},B},
	{"set_select_multiple",{P,B}},
	{"get_select_multiple",{P},B},
	{"set_local_only",{P,B}},
	{"get_local_only",{P},B},
	{"set_limit",{P,I}},
	{"get_limit",{P},I},
	{"set_show_tips",{P,B}},
	{"get_show_tips",{P},B},
	{"set_sort_type",{P,I}},
	{"get_sort_type",{P},I},
	{"set_sort_func",{P,P,P,P}},
	{"get_current_item",{P},P},
	{"set_current_uri",{P,S,P},B},
	{"get_current_uri",{P},S},
	{"select_uri",{P,S,P},B},
	{"unselect_uri",{P,S}},
	{"select_all",{P}},
	{"unselect_all",{P}},
	{"get_items",{P},A},
	{"get_uris",{P,I},A},
	{"add_filter",{P,P}},
	{"remove_filter",{P,P}},
	{"list_filters",{P},A},
	{"set_filter",{P,P}},
	{"get_filter",{P},P},
"GtkRecentChooser"}

control[GtkRecentChooserWidget] = {"gtk_recent_chooser_widget",
	{GtkOrientable,GtkRecentChooser,GtkBox},
	{"new",{},P},
	{"new_for_manager",{P},P},
"GtkRecentChooserWidget"}

control[GtkRecentChooserDialog] = {"gtk_recent_chooser_dialog",
	{GtkRecentChooser,GtkDialog},
	{"new",{S,P,S,I,I},P},
  	{"new_for_manager",{S,P,P,S,I,I},P},
"GtkRecentChooserDialog"}

control[GtkRecentChooserMenu] = {"gtk_recent_chooser_menu",
	{GtkRecentChooser,GtkActivatable,GtkMenu},
	{"new",{},P},
	{"new_for_manager",{P},P},
	{"set_show_numbers",{P,B}},
	{"get_show_numbers",{P},B},
"GtkRecentChooserMenu"}

control[GtkStatusbar] = {"gtk_statusbar",
{GtkOrientable,GtkBox},
	{"get_context_id",{P,S},I},
	{"get_message_area",{P},P},
	{"new",{},P},
	{"pop",{P,I}},
	{"push",{P,I,S},I},
	{"remove",{P,I,I}},
	{"remove_all",{P,I}},
"GtkStatusbar"}

control[GtkToolBar] = {"gtk_toolbar",
{GtkOrientable,GtkToolShell,GtkContainer},
	{"get_drop_index",{P,I,I},I},
	{"get_icon_size",{P},I},
	{"get_item_index",{P,P},I},
	{"get_n_items",{P},I},
	{"get_nth_item",{P,I},P},
	{"get_orientation",{P},I},
	{"get_relief_style",{P},I},
	{"get_show_arrow",{P},B},
	{"get_style",{P},P},
	{"get_tooltips",{P},B},
	{"insert",{P,P,I}},
	{"new",{},P},
	{"set_drop_highlight_item",{P,P,I}},
	{"set_icon_size",{P,I}},
	{"set_show_arrow",{P,B}},
	{"set_style",{P,I}},
	{"set_tooltips",{P,B}},
	{"unset_icon_size",{P}},
	{"unset_style",{P}},
"GtkToolBar"}

control[GtkScale] = {"gtk_scale",
{GtkOrientable,GtkRange},
	{"new",{I,P},P},
	{"new_with_range",{I,D,D,D},P},
	{"add_mark",{P,D,I,S}},
	{"clear_marks",{P}},
	{"get_digits",{P},I},
	{"get_draw_value",{P},B},
	{"get_layout_offsets",{P,I,I}},
	{"get_layout",{P},P},
	{"get_value_pos",{P},I},
	{"get_has_origin",{P},B},
	{"set_digits",{P,I}},
	{"set_draw_value",{P,B}},
	{"set_has_origin",{P,B}},
	{"set_value_pos",{P,I}},
"GtkScale"}

control[GtkRange] = {"gtk_range",
{GtkOrientable,GtkWidget},
	{"get_value",{P},D},
	{"get_adjustment",{P},P},
	{"get_fill_level",{P},D},
	{"get_flippable",{P},B},
	{"get_inverted",{P},B},
	{"get_lower_stepper_sensitivity",{P},I},
	{"get_min_slider_size",{P},I},
	{"get_range_rect",{P,P}},
	{"get_restrict_to_fill_level",{P},B},
	{"get_round_digits",{P},I},
	{"get_slider_range",{P,I,I}},
	{"get_slider_size_fixed",{P},B},
	{"get_update_policy",{P},I},
	{"get_upper_stepper_sensitivity",{P},I},
	{"set_adjustment",{P,P}},
	{"set_fill_level",{P,D}},
	{"set_flippable",{P,B}},
	{"set_increments",{P,D,D}},
	{"set_inverted",{P,B}},
	{"set_lower_stepper_sensitivity",{P,I}},
	{"set_min_slider_size",{P,I}},
	{"set_range",{P,D,D}},
	{"set_restrict_to_fill_level",{P,B}},
	{"set_round_digits",{P,I}},
	{"set_show_fill_level",{P,B}},
	{"set_slider_size_fixed",{P,B}},
	{"set_update_policy",{P,I}},
	{"set_upper_stepper_sensitivity",{P,I}},
	{"set_value",{P,D}},
"GtkRange"}

control[GtkArrow] = {"gtk_arrow",
{GtkMisc},
	{"new",{I,I},P},
	{"set",{P,I,I}},
"GtkArrow"}

control[GtkAspectFrame] = {"gtk_aspect_frame",
{GtkFrame},
	{"new",{S,F,F,F,B},P},
	{"set",{P,F,F,F,B}},
"GtkAspectFrame"}

control[GtkAssistant] = {"gtk_assistant",
{GtkWindow},
	{"add_action_widget",{P,P}},
	{"append_page",{P,P},I},
	{"get_current_page",{P},I},
	{"get_n_pages",{P},I},
	{"get_nth_page",{P,I},P},
	{"get_page_complete",{P,P},B},
	{"get_page_title",{P,P},S},
	{"get_page_type",{P,P},I},
	{"insert_page",{P,P,I},I},
	{"new",{},P},
	{"next_page",{P}},
	{"commit",{P}},
	{"prepend_page",{P,P},I},
	{"previous_page",{P}},
	{"remove_action_widget",{P,P}},
	{"remove_page",{P,I}},
	{"set_current_page",{P,I}},
	{"set_forward_page_func",{P,P,P,P}},
	{"set_page_complete",{P,P,B}},
	{"set_page_title",{P,P,S}},
	{"set_page_type",{P,P,I}},
	{"update_buttons_state",{P}},
"GtkAssistant"}

control[GtkScrolledWindow] = {"gtk_scrolled_window",
{GtkBin},
	{"add_with_viewport",{P,P}},
	{"get_hadjustment",{P},P},
	{"get_hscrollbar",{P},P},
	{"get_placement",{P},I},
	{"get_policy",{P,I,I}},
	{"get_shadow_type",{P},I},
	{"get_vadjustment",{P},P},
	{"get_vscrollbar",{P},P},
	{"new",{P,P},P},
	{"set_hadjustment",{P,P}},
	{"set_placement",{P,I}},
	{"set_policy",{P,I,I}},
	{"set_shadow_type",{P,I}},
	{"set_vadjustment",{P,P}},
	{"unset_placement",{P}},
	{"get_min_content_height",{P},I},
	{"set_min_content_height",{P,I}},
	{"get_min_content_width",{P},I},
	{"set_min_content_width",{P,I}},
	{"get_capture_button_press",{P},B},
	{"get_kinetic_scrolling",{P},B},
	{"set_capture_button_press",{P,B}},
	{"set_kinetic_scrolling",{P,B}},
"GtkScrolledWindow"}

control[GtkScaleButton] = {"gtk_scale_button",
{GtkActivatable,GtkOrientable,GtkButton},
	{"new",{I,D,D,D,P},P},
	{"get_adjustment",{P},P},
	{"get_value",{P},D},
	{"get_popup",{P},P},
	{"get_plus_button",{P},P},
	{"get_minus_button",{P},P},
	{"new",{I,D,D,D,S},P},
	{"set_adjustment",{P,P}},
	{"set_icons",{P,A}},
	{"set_value",{P,D}},
"GtkScaleButton"}

control[GtkCellRenderer] = {"gtk_cell_renderer",
{GInitiallyUnowned},
	{"accel_new",{},P},
	{"activate",{P,P,P,S,P,P,I},B},
	{"get_alignment",{P},-routine_id("GetCellRendererAlignment")},
	{"get_fixed_size",{P,I,I}},
	{"get_padding",{P,I,I}},
	{"get_preferred_height",{P,P,I,I}},
	{"get_sensitive",{P},B},
	{"get_visible",{P},B},
	{"render",{P,P,P,P,P,P,I}},
	{"set_alignment",{P,F,F}},
	{"set_fixed_size",{P,I,I}},
	{"set_padding",{P,I,I}},
	{"set_sensitive",{P,B}},
	{"set_visible",{P,B}},
	{"start_editing",{P,P,P,S,P,P,I},P},
	{"stop_editing",{P,B}},
	{"toggle_set_activatable",{P,B}},
	{"toggle_get_activatable",{P},B},
"GtkCellRenderer"}

fn[16] = define_c_proc(GTK,"gtk_cell_renderer_get_alignment",{P,P,P})
------------------------------------------------------------------------
function GetCellRendererAlignment(atom ctl)
------------------------------------------------------------------------
	atom x = allocate(16), y = allocate(16)
	c_proc(fn,{ctl,x,y})
	return {float32_to_atom(peek({x,4})),float32_to_atom(peek({y,4}))}
end function

control[GtkCellRendererCombo] = {"gtk_cell_renderer_combo",
{GtkCellRendererText},
	{"new",{},P},
"GtkCellRendererCombo"}

control[GtkCellRendererText] = {"gtk_cell_renderer_text",
{GtkCellEditable,GtkCellRenderer},
	{"new",{},P},
	{"set_fixed_height_from_font",{P,I}},
"GtkCellRendererText"}

control[GtkCellRendererToggle] = {"gtk_cell_renderer_toggle",
{GtkCellRenderer},
	{"new",{},P},
	{"set_radio",{P,B}},
	{"get_radio",{P},B},
	{"set_active",{P,B}},
	{"get_active",{P},B},
	{"set_activatable",{P,B}},
	{"get_activatable",{P},B},
"GtkCellRendererToggle"}

control[GtkCellRendererPixbuf] = {"gtk_cell_renderer_pixbuf",
{GtkCellRenderer},
	{"new",{},P},
"GtkCellRendererPixbuf"}

control[GtkCellRendererProgress] = {"gtk_cell_renderer_progress",
{GtkCellRenderer},
	{"new",{},P},
"GtkCellRendererProgress"}

control[GtkCellRendererSpin] = {"gtk_cell_renderer_spin",
{GtkCellRendererText},
	{"new",{},P},
"GtkCellRendererSpin"}

control[GtkCellRendererSpinner] = {"gtk_cell_renderer_spinner",
{GtkCellRenderer},
	{"new",{},P},
"GtkCellRendererSpinner"}

control[GtkCellRendererAccel] = {"gtk_cell_renderer_accel",
{GtkCellRendererText},
	{"new",{},P},
"GtkCellRendererAccel"}

control[GtkTextBuffer] = {"gtk_text_buffer",
{GObject},
  {"add_mark",{P,P,P}},
  {"add_selection_clipboard",{P,P}},
  {"apply_tag",{P,P,P,P}},
  {"apply_tag_by_name",{P,S,P,P}},
  {"backspace",{P,P,B,B},B},
  {"begin_user_action",{P}},
  {"copy_clipboard",{P,P}},
  {"create_child_anchor",{P,P},P},
  {"create_mark",{P,S,P,B},P},
  {"create_tag",{P,S,S,S,I},P},
  {"cut_clipboard",{P,P,B}},
  {"delete",{P,P,P}},
  {"delete_interactive",{P,P,P,B},B},
  {"delete_mark",{P,P}},
  {"delete_mark_by_name",{P,S}},
  {"delete_selection",{P,B,B},B},
  {"end_user_action",{P}},
  {"get_bounds",{P,P,P}},
  {"get_char_count",{P},I},
  {"get_copy_target_list",{P},P},
  {"get_end_iter",{P,P}},
  {"get_has_selection",{P},B},
  {"get_insert",{P},P},
  {"get_iter_at_child_anchor",{P,P,P}},
  {"get_iter_at_line_index",{P,P,I,I}},
  {"get_iter_at_line",{P,P,I}},
  {"get_iter_at_line_offset",{P,P,I,I}},
  {"get_iter_at_mark",{P,P,P}},
  {"get_iter_at_offset",{P,P,I}},
  {"get_line_count",{P},I},
  {"get_mark",{P,S},P},
  {"get_modified",{P},B},
  {"get_paste_target_list",{P},P},
  {"get_selection_bounds",{P,P,P},B},
  {"get_slice",{P,P,P,B},S},
  {"get_start_iter",{P,P}},
  {"get_tag_table",{P},P},
  {"get_text",{P,B},-routine_id("getTextBufferText")},
  {"insert",{P,P,S,I}},
  {"insert_at_cursor",{P,S,I}},
  {"insert_child_anchor",{P,P,P}},
  {"insert_interactive",{P,P,S,I,B},B},
  {"insert_interactive_at_cursor",{P,S,I,B},B},
  {"insert_pixbuf",{P,P,P}},
  {"insert_range",{P,P,P,P}},
  {"insert_range_interactive",{P,P,P,P,B},B},
  {"insert_with_tags",{P}},
  {"insert_with_tags_by_name",{P,P,S,I,S,P}},
  {"move_mark",{P,P,P}},
  {"move_mark_by_name",{P,S,P}},
  {"new",{P},P},
  {"paste_clipboard",{P,P,P,B}},
  {"place_cursor",{P,P}},
  {"remove_all_tags",{P,P,P}},
  {"remove_selection_clipboard",{P,P}},
  {"remove_tag",{P,P,P,P}},
  {"remove_tag_by_name",{P,S,P,P}},
  {"select_range",{P,P,P}},
  {"set_modified",{P,B}},
  {"set_text",{P,P},-routine_id("setTextBufferText")},
  {"deserialize_get_can_create_tags",{P},P},
  {"deserialize",{P,P,P,P,I,I,P},B},
  {"deserialize_set_can_create_tags",{P,P,B}},
  {"get_deserialize_formats",{P,I},P},
  {"get_serialize_formats",{P,I},P},
  {"register_deserialize_format",{P,S,P,P,P},P},
  {"register_deserialize_tagset",{P,S},P},
  {"register_serialize_format",{P,S,P,P,P},P},
  {"register_serialize_tagset",{P,S},P},
  {"serialize",{P,P,P,P,P,I},I},
  {"unregister_deserialize_format",{P,P}},
  {"unregister_serialize_format",{P,P}},
"GtkTextBuffer"}

------------------------------------------------------------------------
-- (5) Text buffer functions;
------------------------------------------------------------------------
 
fn[17] = define_c_proc(GTK,"gtk_text_buffer_set_text",{P,S,I})
------------------------------------------------------------------------
global function setTextBufferText(atom buf, object txt)
------------------------------------------------------------------------
	integer len = -1
	if string(txt) then 
		len = length(txt) 
		txt = allocate_string(txt,1) 
	end if
	c_proc(fn[17],{buf,txt,len})
	return 1
end function

fn[18] = define_c_proc(GTK,"gtk_text_buffer_get_start_iter",{P,P})
fn[19] = define_c_proc(GTK,"gtk_text_buffer_get_end_iter",{P,P})
fn[20] = define_c_func(GTK,"gtk_text_buffer_get_text",{P,P,P,B},S)
------------------------------------------------------------------------
global function getTextBufferText(atom buffer, atom hidden)
------------------------------------------------------------------------
	atom start = allocate(64)
	atom fini = allocate(64)
	c_proc(fn[18],{buffer,start})
	c_proc(fn[19],{buffer,fini})
	object result = c_func(fn[20],{buffer,start,fini,hidden})
	return peek_string(result)
end function

control[GtkTextView] = {"gtk_text_view",
{GtkScrollable,GtkContainer},
  {"add_child_at_anchor",{P,P,P}},
  {"add_child_in_window",{P}},
  {"backward_display_line",{P,P},B},
  {"backward_display_line_start",{P,P},B},
  {"buffer_to_window_coords",{P,P,I,I,I,I}},
  {"forward_display_line_end",{P,P},B},
  {"forward_display_line",{P,P},B},
  {"get_accepts_tab",{P},P},
  {"get_border_window_size",{P,I},I},
  {"get_buffer",{P},P},
  {"get_cursor_visible",{P},P},
  {"get_default_attributes",{P},P},
  {"get_editable",{P},P},
  {"get_indent",{P},P},
  {"get_iter_at_location",{P,P,I,I}},
  {"get_iter_at_position",{P,P,I,I,I}},
  {"get_iter_location",{P,P,P}},
  {"get_justification",{P},P},
  {"get_left_margin",{P},P},
  {"get_line_at_y",{P,P,I,I}},
  {"get_line_yrange",{P,P,I,I}},
  {"get_overwrite",{P},P},
  {"get_pixels_above_lines",{P},P},
  {"get_pixels_below_lines",{P},P},
  {"get_pixels_inside_wrap",{P},P},
  {"get_right_margin",{P},P},
  {"get_tabs",{P},P},
  {"get_visible_rect",{P,P}},
  {"get_window",{P,P},P},
  {"get_window_type",{P,P},I},
  {"get_wrap_mode",{P},P},
  {"move_child",{P}},
  {"move_mark_onscreen",{P,P},B},
  {"move_visually",{P,P,I},B},
  {"new_with_buffer",{P},P},
  {"new",{},P},
  {"place_cursor_onscreen",{P},B},
  {"scroll_mark_onscreen",{P,P}},
  {"scroll_to_iter",{P,P,D,B,D,D}},
  {"scroll_to_mark",{P,P,D,B,D,D}},
  {"set_accepts_tab",{P}},
  {"set_border_window_size",{P,I,I}},
  {"set_buffer",{P,P}},
  {"set_cursor_visible",{P,B}},
  {"set_editable",{P,B}},
  {"set_indent",{P,I}},
  {"set_justification",{P,I}},
  {"set_left_margin",{P,I}},
  {"set_overwrite",{P,B}},
  {"set_pixels_above_lines",{P,I}},
  {"set_pixels_below_lines",{P,I}},
  {"set_pixels_inside_wrap",{P,I}},
  {"set_right_margin",{P,I}},
  {"set_tabs",{P}},
  {"set_wrap_mode",{P,I}},
  {"starts_display_line",{P,P},B},
  {"window_to_buffer_coords",{P,P,I,I,I,I}},
"GtkTextView"}

control[GtkStyleContext] = {"gtk_style_context",
{GObject},
	{"new",{},P},
	{"add_provider",{P,P,I}},
	{"add_provider_for_screen",{P,P,I}},
	{"get_direction",{P},I},
	{"get_junction_sides",{P},I},
	{"get_path",{P},S},
	{"get_property",{P,S,I,P}},
	{"get_screen",{P},P},
	{"get_state",{P},I},
	{"get_style_property",{P,S,P}},
	{"get_color",{P,I,C}},
	{"get_background_color",{P,I,C}},
	{"get_border_color",{P,I,C}},
	{"get_border",{P,I,P}},
	{"get_margin",{P,I,P}},
	{"get_font",{P,I},P},
	{"set_parent",{P,P}},
"GtkStyleContext"}

control[GtkCSSProvider] = {"gtk_css_provider",
{GObject},
	{"new",{},P},
	{"get_default",{},P},
	{"get_named",{S,S},P},
	{"load_from_data",{P,S,I,P},B},
	{"load_from_file",{P,S,P},B},
	{"load_from_path",{P,S,P},B},
	{"to_string",{P},S},
"GtkCSSProvider"}

control[PangoFontsetSimple] = {"pango_fontset_simple",
{GObject},
	{"new",{P},P},
	{"append",{P,P}},
	{"size",{P},I},
"PangoFontsetSimple"}

control[PangoFontSet] = {"pango_fontset",
{PangoFontsetSimple},
	{"get_font",{P,I},P},
	{"get_metrics",{P},P},
	{"foreach",{P,P,P}},
"PangoFontSet"}

control[PangoFontMap] = {"pango_font_map",
{PangoFontSet},
	{"create_context",{P},P},
	{"load_font",{P,P,S},P},
	{"load_fontset",{P,P,S,P},P},
	{"list_families",{P,A,I}},
	{"get_shape_engine_type",{P},S},
"PangoFontMap"}

control[PangoFontFace] = {"pango_font_face",
{PangoFontMap},
	{"get_face_name",{P},S},
	{"list_sizes",{P,A,I}},
	{"describe",{P},P},
	{"is_synthesized",{P},B},
"PangoFontFace"}

control[PangoFontFamily] = {"pango_font_family",
{PangoFontFace},
	{"get_name",{P},S},
	{"is_monospace",{P},B},
	{"list_faces",{P,A,I}},
"PangoFontFamily"}

control[PangoFont] = {"pango_font",
{PangoFontFamily},
	{"get_metrics",{P,P},P},
	{"get_font_map",{P},P},
"PangoFont"}

control[PangoFontDescription] = {"pango_font_description",
{PangoFont},
	{"new",{},P},
	{"copy",{P},P},
	{"copy_static",{P},P},
	{"hash",{P},I},
	{"equal",{P,P},B},
	{"free",{P}},
	{"set_family",{P,S}},
	{"set_family_static",{P,S}},
	{"get_family",{P},S},
	{"set_style",{P,I}},
	{"get_style",{P},I},
	{"set_variant",{P,I}},
	{"get_variant",{P},I},
	{"set_weight",{P,I}},
	{"get_weight",{P},I},
	{"set_stretch",{P,I}},
	{"get_stretch",{P},I},
	{"set_size",{P,I}},
	{"get_size",{P},I},
	{"set_absolute_size",{P,D}},
	{"get_size_is_absolute",{P},B},
	{"set_gravity",{P,I}},
	{"get_gravity",{P},I},
	{"get_set_fields",{P},I},
	{"unset_fields",{P,I}},
	{"merge",{P,P,I}},
	{"merge_static",{P,P,I}},
	{"better_match",{P,P,P},B},
	{"from_string",{S},P},
	{"to_string",{P},S},
	{"to_filename",{P},S},
"PangoFontDescription"}

control[PangoLayout] = {"pango_layout",
{GObject},
	{"new",{P},P},
	{"set_text",{P,S,I}},
	{"get_text",{P},S},
	{"set_markup",{P,S,I}},
	{"set_font_description",{P,P}},
	{"get_font_description",{P},P},
	{"set_width",{P,I}},
	{"get_width",{P},I},
	{"set_height",{P,I}},
	{"get_height",{P},I},
	{"set_wrap",{P,I}},
	{"get_wrap",{P},I},
	{"is_wrapped",{P},B},
	{"set_ellipsize",{P,I}},
	{"get_ellipsize",{P},I},
	{"is_ellipsized",{P},B},
	{"set_indent",{P,I}},
	{"get_extents",{P,P,P}},
	{"get_indent",{P},I},
	{"get_pixel_size",{P,I,I}},
	{"get_size",{P,I,I}},
	{"set_spacing",{P,I}},
	{"get_spacing",{P},I},
	{"set_justify",{P,B}},
	{"get_justify",{P},B},
	{"set_auto_dir",{P,B}},
	{"get_auto_dir",{P},B},
	{"set_alignment",{P,P}},
	{"get_alignment",{P},P},
	{"set_tabs",{P,A}},
	{"get_tabs",{P},A},
	{"set_single_paragraph_mode",{P,B}},
	{"get_single_paragraph_mode",{P},B},
"PangoLayout"}

control[GdkPangoContext] = {"gdk_pango_context",
{GObject},
	{"get",{},P},
"GdkPangoContext"}

control[PangoContext] = {"pango_context",
{GObject},
	{"new",{},P},	
	{"load_font",{P,P},P},
	{"load_fontset",{P,P,P},P},
	{"get_metrics",{P,P,P},P},
	{"list_families",{P,A,I}},
	{"set_font_description",{P,P}},
	{"get_font_description",{P},P},
	{"set_font_map",{P,P}},
	{"get_font_map",{P},P},
	{"set_base_gravity",{P,I}},
	{"get_language",{P},P},
	{"set_language",{P,P}},
	{"get_layout",{P},P},
	{"get_base_dir",{P},I},
	{"set_base_dir",{P,I}},
	{"get_base_gravity",{P},I},
	{"set_base_gravity",{P,I}},
	{"get_gravity",{P},I},
	{"get_gravity_hint",{P},I},
	{"set_gravity_hint",{P,I}},
	{"get_matrix",{P},P},
	{"set_matrix",{P,P}},
"PangoContext"}

control[GtkToolPalette] = {"gtk_tool_palette",
{GtkOrientable,GtkScrollable,GtkContainer,GtkWidget},
	{"new",{},P},
	{"get_exclusive",{P,P},B},
	{"set_exclusive",{P,P,B}},
	{"get_expand",{P,P},B},
	{"set_expand",{P,P,B}},
	{"get_group_position",{P,P},I},
	{"set_group_position",{P,P,I}},
	{"get_icon_size",{P},I},
	{"set_icon_size",{P,I}},
	{"unset_icon_size",{P}},
	{"set_style",{P,P}},
	{"get_style",{P},P},
	{"unset_style",{P}},
	{"add_drag_dest",{P,P,P,I,I}},
	{"get_drag_item",{P,P},P},
	{"get_drag_target_group",{},P},
	{"get_drag_target_item",{},P},
	{"get_drop_group",{P,I,I},P},
	{"get_drop_item",{P,I,I},P},
	{"set_drag_source",{P,P}},
"GtkToolPalette"}

control[GdkRGBA] = {"gdk_rgba",0,
	{"copy",{P},P},
	{"free",{P}},
	{"parse",{P,S},B},
	{"equal",{P,P},B},
	{"hash",{P},I},
	{"to_string",{P},S},
"GdkRGBA"}

control[GtkFinal] = {0,0,
"GtkFinal"}

-------------------
----==[ END ]==----
-------------------
26.0